function fileName = TF_snapshots( orgHistory, cellName, startTime, ...
    endTime, timeStep)
%TF_snapshots   Creates a table containing the TF state at specified times
%
%   TF_snapshots(orgHistory, cellName, startTime, endTime, timeStep)
%   creates a table in the file 'TF_numbers.txt' containing the total 
%   population of LIN31, CEH20-LIN39 complex, and NICD-LAG1 complex bound
%   and unbound to the DNA in the nucleus of the specified cell at each
%   timeStep times (default 1800s) between the specified startTime and 
%   endTime.
%
%   See the GESSA User's Manual for further description of this function. 
%  
%   GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%   Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 

    % store number of TF's each timeStep time for each organism
   % orgHistory is history for all organisms
    fileName = '';
    if (nargin < 5)
        timeStep = 1800;
    end
    NCells = length(cellName);
    NOrgs = length(orgHistory);
    snapshots = (startTime+timeStep):timeStep:endTime;
    NSnapshots = length(snapshots);
           
     try
         % store results
         fileName = 'TF_numbers.txt';
         fid = fopen(fileName, 'wt');
         fprintf(fid, 'snapshots at time:\n');
         fprintf(fid, '\t');
         fprintf(fid, '\t%3.1f', snapshots/3600);
         fprintf(fid, '\t%s', 'mean');
         fprintf(fid, '\n');   
             
         for iOrg = 1:NOrgs
             fprintf(fid, '-------------------------------------\n');
             fprintf(fid, 'organism %d\n',iOrg);
             NMoleculesLIN39 = zeros(NCells,NSnapshots);
             NMoleculesLAG2 = zeros(NCells,NSnapshots);
             NMoleculesLIP1 = zeros(NCells,NSnapshots);
    
             k = 1;
             for i = snapshots
                 NMoleculesLIN39(:,k) = getSpeciesNumber(orgHistory(iOrg,1),'nucleus_LIN31', 'TT_Specie', cellName, i) + ...
                     getSpeciesNumber(orgHistory(iOrg,1), 'nucleus_DNAlin39_LIN31', 'TT_Specie', cellName, i);
                 %+...  getSpeciesNumber(orgHistory, 'nucleus_DNAlin39_LIN31_RNAP', 'TT_LIN39Specie', cellName, i);
                 NMoleculesLAG2(:,k) = getSpeciesNumber(orgHistory(iOrg,1),'nucleus_CEH20_LIN39', 'TT_Specie', cellName, i) + ...
                     getSpeciesNumber(orgHistory(iOrg,1), 'nucleus_DNAlag2_CEH20_LIN39', 'TT_Specie', cellName, i);
                 NMoleculesLIP1(:,k) = getSpeciesNumber(orgHistory(iOrg,1),'nucleus_NICD_LAG1', 'TT_Specie', cellName, i) + ...
                     getSpeciesNumber(orgHistory(iOrg,1), 'nucleus_DNAlip1_NICD_LAG1', 'TT_Specie', cellName, i);
                 %    getSpeciesNumber(orgHistory, 'nucleus_DNAlip1_LIN12_LAG1_RNAP', 'TT_LIP1Specie', cellName, i);
                 k = k+1;
             end
             
             fprintf(fid, char('LIN31\n'));
             for iCell = 1:NCells
                 fprintf(fid,'%s\t', char(cellName(iCell)));
                 fprintf(fid, '%g\t', NMoleculesLIN39(iCell,:));
                 fprintf(fid, '\t%g\t', mean(NMoleculesLIN39(iCell,:)));
                 fprintf(fid, '\n');
             end
             fprintf(fid, char('LIN39+CEH20\n'));
             for iCell = 1:NCells
                 fprintf(fid,'%s\t', char(cellName(iCell)));
                 fprintf(fid, '%g\t', NMoleculesLAG2(iCell,:));
                 fprintf(fid, '\t%g\t', mean(NMoleculesLAG2(iCell,:)));
                 fprintf(fid, '\n');
             end
             fprintf(fid, char('LIN12+LAG1\n'));
             for iCell = 1:NCells
                 fprintf(fid,'%s\t', char(cellName(iCell)));
                 fprintf(fid, '%g\t', NMoleculesLIP1(iCell,:));
                 fprintf(fid, '\t%g\t', mean(NMoleculesLIP1(iCell,:)));
                 fprintf(fid, '\n');
             end
         end
         fclose(fid);
     catch ME
         outME = addCause( ...
             MException( ...
             'MATLAB:TF_snapshots', ...
             'Error in counting number of TFs.'), ME);
              throw(outME);
     end
end
