%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
%  classdef CellProperties
%  Defines the components and relationships in the cell.
%
%  Author:  Elana J. Fertig
%  Date:    18 November 2008
%
classdef CellProperties < handle
    
    properties (SetAccess = private, GetAccess = public)
        
        %% identity of the particular cell
        CellType  % cell type for which these properties are valid
        
        %% protein attributes
        ProteinTypes  % vector of protein types in the cell
        NProteins     % number of distinct protein types
        PPI           % NProteins x NProteins x 2 whose ith column 
                      % indicates rate of proteins that activate / repress the 
                      % ith protein.  If the last entry is 1, refers to the rate for unbound proteins
					  % if equal to 2, refers to the rate for bound proteins.
        canTransport  % NProteins vector indicating if an active protein is transported to the nucleus
        transportTime % NProteins vector with transport time for transport species 
                      % and negative for non-transport species (Simulator.transporters)
        
        %% attributes of external signals received in cell
        ExternalSignalTypes % vector of external signal types on the cell
        NExternalSignals    % number of distinct external signal types
        ExternalSignals     % NReceptors x NExternalSignals matrix indicating rate at which
                            % proteins are activated by each signal type
        isExternalNeighbor  % NReceptors logical vector which is true if the signal acts on neighbor cells
        SignalIDs;          % vector with IDs of diffused signals
        DistanceToSource    % vector corresponding to SignalIDs with distances to source of signal
        
        %% repressor attributes
        RepressorTypes % vector of repressor types on the cell
        NRepressors    % number of distinct repressor types
        Repressors     % NProteins x NRepressors x 2 array indicating which
                       % proteins are inhibited by each repressor type for unbound proteins (j=1)
					   % and bound proteins (j=2)

        %% scaffold attributes
        ScaffoldTypes % vector of scaffold types on the cell
        NScaffolds    % number of distinct scaffold types
        Scaffolds     % NProteins x NScaffolds x 2 array indicating the rate at which
                      % proteins are bound (last entry = 1) or unbound (last entry = 2) to each scaffold type
                      
        %% neighboring cell info
        NeighborCellTypes    % vector of cells neighboring this one
        NNeighborCells       % number of neighboring cells
        isNeighborCell       % NCells logical vector indicating whether each cell 
                             % neighbors this one
                             
        %% additional geometry info
        GrowthRate          % rate at which the cell moves away from the signal source
                   
        %% receptor attributes
        ReceptorTypes   % vector of receptors
        NReceptors      % number of receptors
        Receptor2Target % NProteins x NReceptors x 2 matrix indicating the
                        % proteins that are activated by the receptors
                        
        %% spontaneous activity rates
        SpontaneousActivationReceptors         % NReceptors vector with spontaneous activation rates for receptors
        SpontaneousActivationUnboundProteins   % NProteins vector with spontaneous activation rates for unbound proteins
        SpontaneousActivationBoundProteins     % NProteins x NScaffold matrix with self activation rates for bound proteins
        SpontaneousInactivationReceptors       % NReceptors vector with inactivation rates for receptors
        SpontaneousInactivationUnboundProteins % NProteins vector with inactivation rates for unbound proteins
        SpontaneousInactivationBoundProteins   % NProteins x NScaffold matrix with inactivation rates for bound proteins
       
        
                   
        %% TT part attributes
        TT_Species % vector of the SSA_Properties objects with parameters 
                   % for stochastic simulation of transcription/translation
                   % each SSA_Properties object is created from SBML file
                   % with tag TT
        
    end
    
    methods
        
        function cell = CellProperties(CellType, ...
                Proteins, PPI, ...
                ExternalSignalTypes, ExternalSignals, isNeighborSig, ...
                RepressorTypes, Repressors, ...
                ScaffoldTypes, Scaffolds, ...
                ReceptorTypes, Receptor2Target, ...
                NeighborCells, CellNames, ...
                distanceToAnchor, distanceToAnterior, distanceToPosterior, GrowthRate, ...
                canTransport)
        %% cell = CellProperties(CellType, Proteins, PPI,
        %                      ExternalSignalTypes, ExternalSignals,
        %                      isNeighborSig,
        %                      RepressorTypes, Repressors,
        %                      ScaffoldTypes, Scaffolds,
        %                      TFTypes, TF2Target,
        %                      ReceptorTypes, Receptor2Target,
        %                      NeighborCells, CellNames)
        %
        %  Constructor to instantiate CellContents based on cell components
        %  and their relationships.
        %
        %  Input Arguments:
        %  (1)  Name of the cell
        %  (2)  Vector of protein names contained in the cell (or Protein
        %       data types)
        %  (3)  Matrix of protein-protein interactions
        %  (4)  Vector of names of external signals perceived by cell
        %       receptors
        %  (5)  Matrix of receptors activated by the external signals
        %  (6)  Vector indicating if each of the external signals acts by
        %       neighbors
        %  (7)  Vector of names of repressors in the cell
        %  (8)  Matrix of proteins repressed by the repressors
        %  (9)  Vector of names of scaffolds in the cell
        %  (10) Matrix of proteins bound by the scaffolds
        %  (11) Vector of receptor types
        %  (12) Matrix of proteins activated by the receptors
        %  (13) Vector of names of neighbor cells [optional]
        %  (14) Vector of all cell names          [optional]
        %
        %  Returns:
        %  (1) An instantiation of class CellProperties
        %   
            
            if (nargin == 0)
                return;
            end
        
            if (isa(CellType,'CellProperties'))
                
                cell.CellType = CellType.CellType;
                
                cell.setCellProperties(...
                    CellType.ProteinTypes, CellType.PPI, ...
                    CellType.ExternalSignalTypes, CellType.ExternalSignals, ...
                    CellType.isExternalNeighbor, ...
                    CellType.RepressorTypes, CellType.Repressors, ...
                    CellType.ScaffoldTypes, CellType.Scaffolds, ...
                    CellType.ReceptorTypes, CellType.Receptor2Target);
                
                if (~isempty(CellType.NeighborCellTypes))
                    cell.NeighborCellTypes = CellType.NeighborCellTypes;
                    cell.NNeighborCells = CellType.NNeighborCells;
                    cell.isNeighborCell = CellType.isNeighborCell;
                end
                
                cell.TT_Species = CellType.TT_Species;
                             
                cell.setDistance(CellType.DistanceToAnchor, ...
                    CellType.DistanceToAnterior, ...
                    CellType.DistanceToPosterior, cellType.GrowthRate);
                
                cell.setCanTransport(cellType.canTransport);
                
            else
                %% identify cell
            
                cell.CellType = char(strtok(CellType));
            
                %% specify proteins
                if (isa(Proteins, 'Protein'))
                  throw(MException('MATLAB:CellProperties:cannotYetProcess', ...
                       'TODO: Make matlab construct a cell type from Protein objects, extracting from the loadOrganismXML code.'));
                else
                    if (nargin > 18)
                        cell.setCanTransport(canTransport);
                    end
                    
                    if (nargin > 17)
                        cell.setCellProperties(Proteins, PPI, ...
                            ExternalSignalTypes, ExternalSignals, isNeighborSig, ...
                            RepressorTypes, Repressors, ...
                            ScaffoldTypes, Scaffolds, ...
                            ReceptorTypes, Receptor2Target, ...
                            NeighborCells, CellNames, ...
                            distanceToAnchor, distanceToAnterior, distanceToPosterior, GrowthRate);
                    elseif (nargin > 14)
                        cell.setCellProperties(Proteins, PPI, ...
                            ExternalSignalTypes, ExternalSignals, isNeighborSig, ...
                            RepressorTypes, Repressors, ...
                            ScaffoldTypes, Scaffolds, ...
                            ReceptorTypes, Receptor2Target, ...
                            NeighborCells, CellNames, ...
                            distanceToAnchor, distanceToAnterior, distanceToPosterior, 0.);
                    elseif (nargin > 12)
                        cell.setCellProperties(Proteins, PPI, ...
                            ExternalSignalTypes, ExternalSignals, isNeighborSig, ...
                            RepressorTypes, Repressors, ...
                            ScaffoldTypes, Scaffolds, ...
                            ReceptorTypes, Receptor2Target, ...
                            NeighborCells, CellNames);
                    else
                        cell.setCellProperties(Proteins, PPI, ...
                            ExternalSignalTypes, ExternalSignals, isNeighborSig, ...
                            RepressorTypes, Repressors, ...
                            ScaffoldTypes, Scaffolds, ...
                            ReceptorTypes, Receptor2Target);
                    end
                end
            end
        end
        

        function setCanTransport(cellType, canTransport)
            cellType.canTransport = canTransport;
        end
        function setTransportTime(cellType, times)
            cellType.transportTime = times;
        end
        
        function nState = getNState(cellType, stateName, ttName)
		%% nState = cellType.getNState(stateName, ttName)
        %  Gets a the number of elements for a particular state type.
            try
                stateType = getCellPropertyByName(stateName);
                if (strcmpi(stateType, 'ActiveProtein') || ...
                        strcmpi(stateType, 'TransportProtein') )
                    stateType = 'Protein';
                end
                if (strcmpi(stateType, 'ActiveReceptor'))
                    stateType = 'Receptor';
                end
                if (strcmpi(stateType, 'TT_Specie'))
                    if (nargin < 3)
                        outME = addCause( ...
                           MException( ...
                                'MATLAB:CellProperty:getNState:invalidCall', ...
                                'The species name should be specified'));
                        throw(outME);
                    end
                    
                    % make sure working with correct one
                    
                    % find index of TT object in TT_Species
                    index = findType(ttName, {cellType.TT_Species(:).Names});
                    if (~isempty(index))
                        nState = size(cellType.TT_Species(index(1)).Names,1);
                    else 
                       outME = addCause( ...
                           MException( ...
                                'MATLAB:CellProperty:getNState:invalidSpecies', ...
                                'Could not find index for the %s species',ttName ));
                        throw(outME);
                    end
                    return;
                else
                    nState = eval(sprintf('cellType.N%ss', ...
                        stateType));
                end
            catch ME
               outME = addCause( ...
                   MException( ...
                        'MATLAB:CellState:invalidState', ...
                        'Cannot return cell state of type %s', ...
                        stateName ), ...
                    ME);
                throw(outME);
            end
        end
        
        function stateNames = getStateNames(cellType, stateName)
        %% stateNames = cellType.getStateNames(stateType)
        %  Gets the list of state names for a particular type.
        stateNames= [];
            try
                stateType = getCellPropertyByName(stateName);
                if (strcmpi(stateType, 'ActiveProtein')|| ...
                        strcmpi(stateType, 'TransportProtein') )
                    stateType = 'Protein';
                end
                if (strcmpi(stateType, 'ActiveReceptor'))
                    stateType = 'Receptor';
                end
                if (strcmpi(stateType, 'TT_Specie'))
                    for i = 1:length(cellType.TT_Species)
                        stateNames = [stateNames, {cellType.TT_Species(1,i).Names}];
                    end
                else
                    stateNames = eval(sprintf('cellType.%sTypes', ...
                        stateType));
                end

            catch ME
                outME = addCause( ...
                    MException( ...
                        'MATLAB:CellState:invalidState', ...
                        'Cannot return cell state of type %s', ...
                        stateName ), ...
                    ME);
                throw(outME);
            end
        end
        
        %% add SSA_Properties object in TT_Species property
        function addTTObject(cellType, TTObject)
           cellType.TT_Species = [cellType.TT_Species, TTObject];
        end
                
        function setNeighborCells(cell, NeighborCells, CellNames)
                    %% specify structure of neighbor cells
            cell.NeighborCellTypes = char((NeighborCells));
            cell.NNeighborCells    = int8(size(cell.NeighborCellTypes, 1));
            cell.isNeighborCell    = findTypes(...
                char(CellNames), size(CellNames,1), ...
                cell.NeighborCellTypes, cell.NNeighborCells, true);
        end
        
        function setSignalInfo(cell, ...
                signals, distances, ...
                growthRate)
            cell.SignalIDs = signals;
            cell.DistanceToSource  = distances;
            if (nargin < 4)
                growthRate = 0.;
            end
            cell.GrowthRate = growthRate;
        end
        
        function setSpontaneousRates(cell, ...
                actRecept, actUnbound, actBound, ...
                inactRecept, inactUnbound, inactBound)
            cell.SpontaneousActivationReceptors = actRecept;
            cell.SpontaneousActivationUnboundProteins = actUnbound;
            cell.SpontaneousActivationBoundProteins = actBound;
            cell.SpontaneousInactivationReceptors = inactRecept;
            cell.SpontaneousInactivationUnboundProteins = inactUnbound;
            cell.SpontaneousInactivationBoundProteins = inactBound;
        end
        
    end
    
    methods (Access = private)
        function setCellProperties(cell, Proteins, PPI, ...
                ExternalSignalTypes, ExternalSignals, isNeighborSig, ...
                RepressorTypes, Repressors, ...
                ScaffoldTypes, Scaffolds, ...
                ReceptorTypes, Receptor2Target, ...
                NeighborCells, CellNames, ...
                distanceToAnchor, distanceToAnterior, distanceToPosterior, ...
                growthRate)
            
            cell.ProteinTypes = char(Proteins);
            cell.NProteins = size(cell.ProteinTypes, 1);
            cell.PPI = PPI;
            if ( (size(cell.PPI,1) ~= cell.NProteins) || ...
                 (size(cell.PPI,2) ~= cell.NProteins))
                throw(MException(...
                    'MATLAB:CellProperties:invalidDim', ...
                    'PPI is an invalid dimension (%d x %d), should be (%d x %d)', ...
                    size(cell.PPI,1), size(cell.PPI,2), cell.NProteins, cell.NProteins));
            end
            
            %% specify external signals sensed by receptors
            cell.ExternalSignalTypes = char((ExternalSignalTypes));
            cell.NExternalSignals    = int8(size(cell.ExternalSignalTypes, 1));
            cell.ExternalSignals     = ExternalSignals;
            cell.isExternalNeighbor = isNeighborSig;
            
            %% specify repressors
            cell.RepressorTypes = char((RepressorTypes));
            cell.NRepressors    = int8(size(cell.RepressorTypes, 1));
            cell.Repressors     = Repressors;
            
            
            %% specify scaffolds
            cell.ScaffoldTypes = char((ScaffoldTypes));
            cell.NScaffolds    = int8(size(cell.ScaffoldTypes, 1));
            cell.Scaffolds     = Scaffolds;
            
            %% specify receptors
            cell.ReceptorTypes   = ReceptorTypes;
            cell.NReceptors      = int8(size(cell.ReceptorTypes,1));
            cell.Receptor2Target = Receptor2Target;
            
            if (nargin > 12)
                cell.setNeighborCells(NeighborCells, CellNames);
            end
            
            if (nargin > 14)
                cell.setDistance(distanceToAnchor, ...
                    distanceToAnterior, distanceToPosterior, ...
                    growthRate);
            end
            
        end
    end
end
