%% classdef CellState
%  Defines the concentration of cell elements at a given time.
%
%  Author:  Elana J. Fertig
%  Date:    25 November 2008
%  Version: $ID:$
%
%  Modification History:
%  25Nov08  EJF  Simplified CellContents class to represent a single time.
%  16Feb08  EJF  Changed to handle class for efficiency
%
classdef CellState < handle
    
    
    properties (SetAccess = public, GetAccess = public)
        Proteins             % Concentration of proteins
        ActiveProteins       % Concentration of active proteins
        ExternalSignals      % Concentration of external signals
        Repressors           % Concentration of repressors
        Scaffolds            % Concentration of scaffolds
        Receptors            % Concentration of receptors
        ActiveReceptors      % Concentration of active receptors
        TT_Species = {};     % Populations of species involved in TT
                             % cell vector of TT_Species size from CellProperties 
                             % each element of vector is vecor of length of
                             % SSA_Properties.Names
        NeighborSignals      % Concentration of signals in the neighbor cells
        Active = false;      % Boolean activity of the cell (i.e., live or not)
    end
    
    methods (Access = public)
        
        function outCell = CellState(...
                Proteins, ...
                ActiveProteins, ...
                ExternalSignals, ...
                Repressors, ...
                Scaffolds, ...
                Receptors, ...
                NeighborSignals, ...
                TT_Population)
        %% outCell = CellState(Prot, ActProt, Sig, Repr, Scaff, TTSpec)
        %  Construct components of the cell.
        %
        %  Input Arguments:
        %  (1) Concentration of proteins
        %  (2) Concentration of active proteins
        %  (3) Concentration of external signals
        %  (4) Concentration of repressors
        %  (5) Concentration of scaffolds
        %  (6) Population of receptors
        %  (7) Population of neighbor signals
        %  (8) Population of species in transcription/translation
        %
        %  Return:
        %  (1) Instantiation of class CellState with component
        %      concentrations.
        %
            
            if (nargin == 0)
                %initialize an empty cell
            elseif (isa(Proteins, 'CellState'))
                outCell.Proteins          = Proteins.Proteins;
                outCell.ActiveProteins    = Proteins.ActiveProteins;
                outCell.ExternalSignals   = Proteins.ExternalSignals;
                outCell.Repressors        = Proteins.Repressors;
                outCell.Scaffolds         = Proteins.Scaffolds;
                outCell.TT_Species    = Proteins.TT_Species;
                outCell.Receptors         = Proteins.Receptors;
                outCell.ActiveReceptors   = Proteins.ActiveReceptors;
                outCell.NeighborSignals   = Proteins.NeighborSignals;
                outCell.Active            = Proteins.Active;
            else
                outCell.Proteins          = convert2Double(Proteins);
                outCell.ActiveProteins    = convert2Double(ActiveProteins);
                outCell.ExternalSignals   = convert2Double(ExternalSignals);
                outCell.Repressors        = convert2Double(Repressors);
                outCell.Scaffolds         = convert2Double(Scaffolds);
                outCell.Receptors         = convert2Double(Receptors);
                outCell.ActiveReceptors   = zeros(size(outCell.Receptors)); % todo -- add this in correctly
                if (nargin > 6)
                    outCell.NeighborSignals = convert2Double(NeighborSignals);
                end
                if (nargin > 7)
                    for i = 1:length(state.TT_Population)
                       outCell.TT_Population{1,i} = convert2Double(TT_Population{1,i});
                    end
               end
                
            end
            
        end
        
        function [progeny1, progeny2] = divideCells(state)
            
            warning('CellState:invalidFunction', 'DivideCells is currently a place holder and is non-functional.\n');
            
            % initialize the cell state to match its parent state
            progeny1 = CellState(state);
            progeny2 = CellState(state);
            
            % set both progeny as active and inactivate the current state
            state.Active = false;
            progeny1.Active = true;
            progeny2.Active = true;
            
            % halve anything which evolved during the simulation
            progeny1.ActiveProteins = binornd(state.ActiveProteins, 0.5);
            progeny2.ActiveProteins = state.ActiveProteins - progeny1.ActiveProteins;
            progeny1.ExternalSignals = binornd(state.ExternalSignals, 0.5);
            progeny2.ExternalSignals = state.ExternalSignals - progeny1.ExternalSignals;
            progeny1.Repressors = binornd(state.Repressors, 0.5);
            progeny2.Repressors = state.Repressors - progeny1.Repressors;
            progeny1.ActiveReceptors = binornd(state.ActiveReceptors, 0.5);
            progeny2.ActiveReceptors = state.ActiveReceptors - progeny1.ActiveReceptors;
            progeny1.NeighborSignals = binornd(state.NeighborSignals, 0.5);
            progeny2.NeighborSignals = state.NeighborSignals - progeny1.NeighborSignals;
            
            % halve all relevant species for TT
            
            
                
        end
        
        function sumState = plus(state1, state2)
        %% sum = plus(state1, state2)
        %  Overload the addition operator for the CellState.
        %
            if (state1.Active && state2.Active)
                sumState = CellState();
                sumState.Proteins          = state1.Proteins           + ...
                    state2.Proteins; 
                sumState.ActiveProteins    = state1.ActiveProteins     + ...
                    state2.ActiveProteins;
                sumState.ExternalSignals   = state1.ExternalSignals    + ...
                    state2.ExternalSignals;
                sumState.Repressors        = state1.Repressors         + ...
                    state2.Repressors;
                sumState.Scaffolds         = state1.Scaffolds          + ...
                    state2.Scaffolds;
                sumState.Receptors         = state1.Receptors          + ...
                    state2.Receptors;
                sumState.ActiveReceptors   = state1.ActiveReceptors    + ...
                    state2.ActiveReceptors;
                sumState.NeighborSignals   = state1.NeighborSignals    + ...
                    state2.NeighborSignals;
                for i = 1:length(state1.TT_Species)
                    sumState.TT_Species{1,i} = state1.TT_Species{1,i} + ...
                        state2.TT_Species{1,i};
                end
               sumState.Active = true;
 
            elseif (state1.Active)
                sumState = CellState(state1);
                sumState.Active = true;
            elseif (state2.Active)
                sumState = CellState(state2);
                sumState.Active = true;
            else
                sumState = CellState();
                sumState.Active = false;
            end
        end
        
        function diffState = minus(state1, state2)
        %% sum = minus(state1, state2)
        %  Overload the subtraction operator for the CellState.
        %
            diffState = CellState();
            diffState.Active = false;
            if (state1.Active && state2.Active)
                diffState.Proteins         = state1.Proteins        - ...
                    state2.Proteins;
                diffState.ActiveProteins   = state1.ActiveProteins  - ...
                    state2.ActiveProteins;
                diffState.ExternalSignals  = state1.ExternalSignals - ...
                    state2.ExternalSignals;
                diffState.Repressors       = state1.Repressors      - ...
                    state2.Repressors;
                diffState.Scaffolds        = state1.Scaffolds       - ...
                    state2.Scaffolds;
                diffState.Receptors        = state1.Receptors  - ...
                    state2.Receptors;
                diffState.ActiveReceptors  = state1.ActiveReceptors  - ...
                    state2.ActiveReceptors;
                diffState.NeighborSignals  = state1.NeighborSignals  - ...
                    state2.NeighborSignals;  
                diffState.Active = true;
                
                for i = 1:length(state1.TT_Species)
                    diffState.TT_Species{1,i} = state1.TT_Species{1,i} - ...
                        state2.TT_Species{1,i};
                end
            end
            
                        
        end
        
        function isPositive = checkPositive(state)
            
            isPositive = 0;
            
            for i = 1:length(state.TT_Species)
                isPositive = all(state.TT_Species{1,i}>=0);
            end
            
            isPositive = isPositive && all(state.Proteins >= 0) && ...
                all(state.ActiveProteins >= 0) && ...
                all(state.ExternalSignals >= 0) && ...
                all(state.Repressors >= 0) && ...
                all(state.Scaffolds >= 0) && ...
                all(state.Receptors >= 0) && ...
                all(state.NeighborSignals >= 0) && ...
                all(state.ActiveReceptors >= 0);
            
%             fprintf('Proteins\n')
%             state.Proteins
%             fprintf('Active Proteins\n')
%             state.ActiveProteins
%             fprintf('External Signals\n')
%             state.ExternalSignals
%             fprintf('Repressors\n')
%             state.Repressors
%             fprintf('Receptors\n')
%             state.Receptors
%             fprintf('Scaffolds\n')
%             state.Scaffolds
%             fprintf('ActiveReceptors\n')
%             state.ActiveReceptors
%             fprintf('Neighbor Signals\n')
%             state.NeighborSignals
%             
            if (any(state.Proteins < 0))
                fprintf('Negative Protein concentration\n');
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Protein Concentration');
            end
            
            if (any(state.ActiveProteins < 0))
                
                                fprintf('Negative Active Protein Concentration\n');
                
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Protein Concentration');
            end
            
            
            if (any(state.ExternalSignals < 0))
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Signal Concentration');
            end
            
            if (any(state.Repressors < 0))
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Repressor Concentration');
            end
            
            if (any(state.Receptors < 0))
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Receptor Concentration');
            end
            
            if (any(state.Scaffolds < 0))
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Scaffold Concentration');
            end
            
            if (any(state.ActiveReceptors < 0))
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Active Receptor Concentration');
            end
            
            if (any(state.NeighborSignals < 0))
                warning('MATLAB:CellState:InvalidPopulation', ...
                    'Negative Neighbor Signal Concentration');
            end
            
            for i = 1:length(state.TT_Species)
                if (any(state.TT_Species{1,i} <0))
                    warning('MATLAB:CellState:InvalidPopulation', ...
                        'Negative TT_Species Concentration');
                end
            end
            
        end
        
        function timesState = times(state1, state2)
        %% timesState = times(state1, state2)
        %  Overload the multiplication operator for the CellState.
        %
        
            if (isa(state1, 'CellState'))
                if (~isscalar(state2) || ~isnumeric(state2))
                    throw( MException('MATLAB:CellState:invalidTimes', ...
                        'Cannot multiply CellState by %s', class(state2)));
                end
                timesState = multiplyStates(state2,CellState(state1));
            elseif (isa(state2, 'CellState'))
                if (~isscalar(state1) || ~isnumeric(state1))
                    throw( MException('MATLAB:CellState:invalidTimes', ...
                        'Cannot multiply CellState by %s', class(state1)));
                end
                timesState = multiplyStates(state1,CellState(state2));
            else
                throw( MException('MATLAB:CellState:invalidTimes', ...
                    'Cannot multiply class %s by class %s', ...
                    class(state1), class(state2)) );
            end
        end
        
        function outState = getState(cellState, stateName, index)
        %% outState = cellState.getState(stateName)
        %  Gets a particular state type.
            try
                stateType = getCellPropertyByName(stateName);
                if (nargin >2 && strcmpi(stateType, 'TT_Specie'))
                    outState = cellState.TT_Species{index};
                    return
                end

                outState = eval(sprintf('cellState.%ss', ...
                    stateType));
            catch ME
                outME = addCause( ...
                    MException( ...
                        'MATLAB:CellState:invalidState', ...
                        'Cannot return cell state of type %s', ...
                        stateName ), ...
                    ME);
                throw(outME);
            end
        end
        
        function setState(cellState, type, value, index)
        %% outState = cellState.setState(type, value, index)
        % set property according type to the value in specified index
        try
             stateType = getCellPropertyByName(type);
             if (nargin > 3)
                 % update only one value in property
                switch stateType
                    case 'Protein'
                        cellState.Proteins(index) = value;
                    case 'ActiveProtein'
                        cellState.ActiveProteins(index) = value;
                    case 'ExternalSignal'
                        cellState.ExternalSignals(index) = value;
                    case 'Repressor'
                        cellState.Repressors(index) = value;
                    case 'Receptor'
                        cellState.Receptors(index) = value;
                    case 'ActiveReceptor'
                        cellState.ActiveReceptors(index) = value;
                end
             elseif (nargin == 2)
                 % update the whole property
                switch stateType
                    case 'Protein'
                        if (length(cellState.Proteins) == length(value))
                            cellState.Proteins = value;
                        end
                    case 'ActiveProtein'
                       if (length(cellState.ActiveProteins) == length(value))
                         cellState.ActiveProteins = value;
                        end
                    case 'ExternalSignal'
                       if (length(cellState.ExternalSignals) == length(value))
                         cellState.ExternalSignals = value;
                       end
                   case 'Repressor'
                        if (length(cellState.Repressors) == length(value))
                        cellState.Repressors = value;
                        end
                   case 'Receptor'
                       if (length(cellState.Receptors) == length(value))
                         cellState.Receptors = value;
                       end
                   case 'ActiveReceptor'
                        if (length(cellState.ActiveReceptors) == length(value))
                            cellState.ActiveReceptors = value;
                        end
               end
            end
         catch ME
                outME = addCause( ...
                    MException( ...
                        'MATLAB:CellState.setState:invalidState', ...
                        'Cannot set type %s', ...
                        type ), ...
                    ME);
                throw(outME);
         end
          
        end
        
        
        function addTTPopulation(cellState, population)
                cellState.TT_Species = [cellState.TT_Species, population];
        end
    end
    
    
    methods (Access = private)
        function timesState = multiplyStates(scalar,timesState)
                
            if (timesState.Active)
                timesState.Proteins          = scalar.*timesState.Proteins;
                timesState.ActiveProteins    = scalar.*timesState.ActiveProteins;
                timesState.ExternalSignals   = scalar.*timesState.ExternalSignals;
                timesState.Repressors        = scalar.*timesState.Repressors;
                timesState.Scaffolds         = scalar.*timesState.Scaffolds;
                timesState.Receptors         = scalar.*timesState.Receptors;
                timesState.ActiveReceptors   = scalar.*timesState.ActiveReceptors;
                timesState.NeighborSignals   = scalar.*timesState.NeighborSignals; 
                
                for i = 1: length(timesState.TT_Species)
                   timesState.TT_Species{1,i} = scalar.*timesState.TT_Species{1,i};
                 
                end
            end
        end
    end
    
end
