%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% classdef Cells
%  Defines the elements within the cell.
%
%  Author:  Elana J. Fertig
%  Date:    25 November 2008
%  Version: $ID:$
%
%  Modification History:
%  25Nov08  EJF  Unified components of the cell within the original 
%                Organism and CellContents classes.
%
classdef Cells < handle
    
    properties (SetAccess = private, GetAccess = public)
        Name  % String describing the name of the cell
        Type  % CellProperties defining the type of the cell
    end
    
    properties (SetAccess = private, GetAccess = private)
        canActivate
    end
    
    properties (SetAccess = public, GetAccess = public)
        State % Concentration of components in the cell
    end
    
    methods (Access = public)
        

        function newCell = Cells(Name, Type)
        %% newCell = Cells(Name, Type)
        %  Construct a cell based of a specified type.
        %
        %  Input Arguments:
        %  (1) String identifying the cell being constructed
        %  (2) Instantiation of CellProperties describing the cell type
        %
        %  Return:
        %  (1) Instantiation of class Cells with zero initial concentration
        %
            if (nargin > 0)
                newCell.Name = Name;
                newCell.Type = Type;
            
                newCell.State = CellState(...
                    zeros(Type.NProteins, 1), ...
                    zeros(Type.NProteins, 1), ...
                    zeros(Type.NExternalSignals, 1), ...
                    zeros(Type.NRepressors, 1), ...
                    zeros(Type.NScaffolds, 1), ...
                    zeros(Type.NExternalSignals,1));
                
                newCell.setInactive();
                newCell.canActivate = true;
            end
        end
        
        function setState(thisCell, State)
        %% setState(thisCell, State)
        %  Set the state of this instantiation of the cell.
        %
        %  Input Arguments:
        %  (1) Instantiation of CellState defining the concentration of
        %      components in the cell.
        %
        %  Return:
        %  (None)
        %
            try
                % check the dimensions of the input state
                if (thisCell.State.Active)
                    isValidDim(thisCell.Type, State);
                end
            
                % set the state of the cell to be the input state
                thisCell.State = CellState(State);
            catch ME
                warning(...
                    'MATLAB:PBN:Cells:invalidDimensions', ...
                    'Value of %s cell is not being updated.', ...
                    thisCell.Name);
                warning(getReport(ME));
            end
            
        end
        
        function [stateOfType, stateIdx] = getState(thisCell, stateType, stateName)
            switch nargin
                case 1
                    stateOfType = thisCell.State;
                    stateIdx    = -999;
                case 2
                    stateOfType = thisCell.State.getState(stateType);
                    stateIdx    = 1:length(stateOfType);
                case 3
                    tempState = thisCell.State.getState(stateType);
                    stateList = findType(char(stateName), ...
                        thisCell.Type.getStateNames(stateType));
                    stateOfType = tempState(stateList);
                    stateIdx  = stateList;
            end
        end
        
        function setActive(thisCell)
            if (thisCell.canActivate)
                thisCell.State.Active = true;
            end
            thisCell.canActivate = false; % ensures that a cell may only be activated one time in its life
        end
        
        function setInactive(thisCell)
            thisCell.State.Active = false;
        end
        
        function isActive = Active(thisCell)
            isActive = thisCell.State.Active;
        end
        
    end
    
end
