% DiffusionUpdater < SystemUpdater
%
%  SystemUpdater class that updates the organism's state by
%  propogating external signals with diffusion
% 
%  GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%  Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%  This program is free software: you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation, either version 3 of the License, or
%  (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program.  If not, see <http://www.gnu.org/licenses/>.

classdef DiffusionUpdater < SystemUpdater
    
    %% Internal variables to the TT updater.
    properties (GetAccess = private, SetAccess = private)
        
         
        nDiffuseSignal        % number of signals which can be diffused
        diffuseSignal         % char array of signal IDs 
        diffuseSpecies        % char array of species to be diffused
        diffusionType         % char array of diffusion type (e.g. 3D)
        diffuseSpeed          % array of diffusion speeds for each signal
        sourceGrowthRate      % array of rates at which each signal grows
        sourceMagnitude       % array of maximum magnitudes for each signal
        sourceInitialTime     % array of times at which signals first appears
        sourceEndTime         % array of times at which signal stops
        
        % for optimization puporse
        sigInds         % nCell x nSignal matrix of indecies of signal IDs 
                        % from CellProperties.SignalIDs in diffuseSignal
                        % property for each cell
                        
        specInds        % nCell x nSignal matrix of indecies of species 
                        % from diffuseSpecies property in
                        % CellProperties.ExternalSignalTypes for each cell
      
    end
    
    methods
        
        %% constructor
        function dUpdater = DiffusionUpdater()
        end
        
        %% initializer (inherited from SystemUpdater)
        function [state, time] = initializer(dUpdater, organism, ...
                initFile, initTime)

            
            %% initialize the diffusion properties from file
            readInitFile(dUpdater, initFile);
            
            
            % for optimization puporse fill in sigInds and specInds
            % matrixes
            maxNSig = 0;
            for iCell=1:organism.NCells
                if (maxNSig < organism.OrgCells(iCell).Type.NExternalSignals)
                    maxNSig = organism.OrgCells(iCell).Type.NExternalSignals;
                end
            end
            dUpdater.sigInds = zeros(organism.NCells,maxNSig);
            dUpdater.specInds = zeros(organism.NCells,maxNSig);
            
            for iCell = 1:organism.NCells
            
                % iterate signals for the cell
                s = size(organism.OrgCells(iCell).Type.SignalIDs);
                for iSig = 1:s(1)
                    
                    % find the signal id in dUpdater.diffuseSignal = sigInd
                    sigIndex= findType(organism.OrgCells(iCell).Type.SignalIDs(iSig,:),...
                        dUpdater.diffuseSignal);
                    if (~isempty(sigIndex))
                        dUpdater.sigInds(iCell,iSig) = sigIndex;
                    else
                       warning('DiffusionUpdater:initializer', ...
                            'Cannot find signal %s\n', ...
                            org.OrgCells(iCell).Type.SignalIDs(iSig,:));
                        continue;
                  
                    end
                    % get species for this signal dUpdater.diffuseSpecies(sigInd)
                    % find this species in CellProperties.ExternalSignalTypes = specInd
                    dUpdater.specInds(iCell,iSig) = findType(dUpdater.diffuseSpecies(sigIndex,:),...
                        organism.OrgCells(iCell).Type.ExternalSignalTypes);
                end % iSig
            end % iCell
            
            % update the initial state
            state = organism;
            if (nargin <= 3)
                time  = 0.;
            else
                time = initTime;
            end
            
        end
        
        %% updateState (inherited from SystemUpdater)
        function [delState, newDelT] = update(dUpdater, org, ...
                currentTime, delT)
            
            % Performs diffusion update for all signals:
            % NOTE: This version assumes a single source for each species!
            
            % initialize the array of output states and timestamps
            newState = repmat(CellState, org.NCells, 1);
            delState = repmat(CellState, org.NCells, 1);
            
            
            for iCell = 1:org.NCells
                
                newState(iCell,1) = CellState(org.OrgCells(iCell,1).State);

                              
                % diffuse each signal in the cell
                s = size(org.OrgCells(iCell).Type.SignalIDs);                
                for iSig=1:s(1)
                    
                    % determine which signal is being diffused
                    sigInd = dUpdater.sigInds(iCell,iSig);
                    specInd = dUpdater.specInds(iCell,iSig);
                    if (sigInd ==0 && specInd == 0)
                        % do not diffuse signal which doesn't map to any
                        % species
                        continue;
                    end
                    
                    % switch by type of diffusion
                    switch strtrim(dUpdater.diffusionType(sigInd,:))
                        case '3DPoint'
                            % if this signal is not instantaneous
                            % (sustained) and next time step is less
                            % than time of signal end
                            if (dUpdater.sourceInitialTime(sigInd) <...
                                    dUpdater.sourceEndTime(sigInd) &&...
                                    (currentTime+delT) < dUpdater.sourceEndTime(sigInd))
                                % update population of the species in
                                % CellState.ExternalSignals
                                newState(iCell,1).ExternalSignals(specInd) = ...
                                    org.OrgCells(iCell).State.ExternalSignals(specInd) + ...
                                    dUpdater.diffuseSustain3D(...
                                    dUpdater.sourceInitialTime(sigInd), ...
                                    currentTime, delT, ...
                                    dUpdater.sourceMagnitude(sigInd), ...
                                    dUpdater.sourceGrowthRate(sigInd), ...
                                    dUpdater.diffuseSpeed(sigInd), ...
                                    org.OrgCells(iCell).Type.DistanceToSource(specInd));
                            elseif (dUpdater.sourceInitialTime(sigInd) == ...
                                    dUpdater.sourceEndTime(sigInd))
                                % evolve an instantaneous signal 
                                newState(iCell,1).ExternalSignals(specInd) = ...
                                    dUpdater.diffuse3D(currentTime, ...
                                    dUpdater.sourceInitialTime(sigInd), ...
                                    delT, ...
                                    dUpdater.sourceMagnitude(sigInd), ...
                                    org.OrgCells(iCell).Type.DistanceToSource(specInd));
                            end

                                
                        otherwise
                            warning('DiffusionUpdater:update', ...
                                'Cannot diffuse signal with %s\n', ...
                                dUpdater.diffusionType(sigInd));
                            newState(iCell,1).ExternalSignals(specInd) = ...
                                org.OrgCells(iCell).State.ExternalSignals(specInd);
                    end % switch
                end % for iSig
            end % for iCell
            % get the difference between first state and final state
            for iCell = 1:org.NCells
                
            	delState(iCell,1) = newState(iCell,1) - org.OrgCells(iCell).State;
            end
           % update the time that has elapsed
            newDelT = delT;
        end
    end
   
    methods (Access = private)
        function outSig = diffuse3D(dUpdater, currentTime, initTime,...
                delT, maximumSignal, distance)

            
            timeElapsed = currentTime - initTime + delT;
            
            if (timeElapsed > 0)

               outSig = floor(...
                   diffusion(maximumSignal, diffusionSpeed, timeElapsed, ...
                             distance, 3, false));    
            else
                outSig = 0.;
            end

        end
        
        % diffusion from a sustained point source in dim dimensions
        % with rate of change of source phi(t) is given by 
        % C(r,t) = 1/(4piD)^3/2 \int_{0}^{t} f(t')dt'
        % f(t)   = \phi(t')exp(-r^2/4D(t-t')}/(t-t')^{dim/2}
        % (The mathematics of diffusion by Crank)
        function outSig = diffuseSustain3D(dUpdater, ...
                initTime, currentTime, delT, maximumSignal, ...
                growthRate, diffusionSpeed, distance)
            

            timeElapsed = currentTime - initTime + delT;
            
            timeStart = currentTime - initTime;
            
            %newTime = currentTime + delT;
            
            if (timeElapsed > 0 && timeStart > 0)
               outSig = ...
                   diff_integral(maximumSignal, diffusionSpeed, ...
                   timeStart, timeElapsed, distance, 3);

                outSig = max(outSig, 0.);
                
            else
                outSig = 0.;
            end
            
        end
        
        % read file with initial information about diffusion
        % NOTE now it's read xml file only
        
        function readInitFile( dUpdater, file )
            
            switch lower(file.type)
                case 'xml'
                    % read XML file
                    tree = xmlread(file.name);
                    % get signals
                    ListOfSignals = tree.getElementsByTagName('signal');
                    % initiate number of sinals
                    NSignals = ListOfSignals.getLength;
                    dUpdater.nDiffuseSignal = NSignals;
                    dUpdater.diffuseSignal = {};
                    dUpdater.diffuseSpecies = {};
                    dUpdater.diffusionType = {};
                    dUpdater.diffuseSpeed = [];
                    dUpdater.sourceGrowthRate = [];
                    dUpdater.sourceMagnitude  = [];
                    dUpdater.sourceInitialTime = [];
                    dUpdater.sourceEndTime = [];
                    % iterate signals
                    for i = 1:NSignals
                        % get one element
                        SignalItem = ListOfSignals.item(i-1);
                        % get their attributes
                        allAttributes = SignalItem.getAttributes;
                        % get attributes
                        name = allAttributes.getNamedItem('id');
                        species = allAttributes.getNamedItem('ligand');
                        type = allAttributes.getNamedItem('type');
                        speed = allAttributes.getNamedItem('speed');
                        sourceGrowthRate = allAttributes.getNamedItem('sourceGrowthRate');
                        sourceMagnitude = allAttributes.getNamedItem('sourceMagnitude');
                        sourceInitialTime = allAttributes.getNamedItem('startTime');
                        sourceEndTime = allAttributes.getNamedItem('endTime');
                        
                        % get values of attributes
                        if (~isempty(name))
                            dUpdater.diffuseSignal = [dUpdater.diffuseSignal,...
                                char(name.getNodeValue)];
                        else
                            fprint('DiffusionUpdater:readInitFile: There is no signal name in element %d', i);
                            continue;
                        end
                        if (~isempty(species))
                            % check that there is no such species
                            % in diffuseSpecies list, if there is, throw
                            % out an error
                            sp = char(species.getNodeValue);
                            index = findType(sp,dUpdater.diffuseSpecies);
                            if (isempty(index))
                                dUpdater.diffuseSpecies = ...
                                    [dUpdater.diffuseSpecies, sp];
                            else
                               throw(MException( ...
                                        'MATLAB:DiffusionUpdater:readInitFile:invalidCall', ...
                                        'DiffusionUpdater currently diffuses species from a single source. Will allow multiple sources in future releases.'));
                            end
                        end
                        if (~isempty(type))
                            dUpdater.diffusionType = [dUpdater.diffusionType,...
                                char(type.getNodeValue)];
                        end
                        if (~isempty(speed))
                            dUpdater.diffuseSpeed = [dUpdater.diffuseSpeed, ...
                                str2double(char(speed.getNodeValue))];
                        end
                        if (~isempty(sourceGrowthRate))
                            dUpdater.sourceGrowthRate = [dUpdater.sourceGrowthRate, ...
                                str2double(char(sourceGrowthRate.getNodeValue))];
                        end
                        if (~isempty(sourceMagnitude))
                            dUpdater.sourceMagnitude = [dUpdater.sourceMagnitude, ...
                                str2double(char(sourceMagnitude.getNodeValue))];
                        end
                        if (~isempty(sourceInitialTime))
                            dUpdater.sourceInitialTime = [dUpdater.sourceInitialTime, ...
                                str2double(char(sourceInitialTime.getNodeValue))];
                        end
                        if (~isempty(sourceEndTime))
                            dUpdater.sourceEndTime = [dUpdater.sourceEndTime, ...
                                str2double(char(sourceEndTime.getNodeValue))];
                        end
                    end% for i
                    dUpdater.diffuseSignal = char(dUpdater.diffuseSignal');
                    dUpdater.diffuseSpecies = char(dUpdater.diffuseSpecies');
                    dUpdater.diffusionType = char(dUpdater.diffusionType');
                    dUpdater.sourceGrowthRate = dUpdater.sourceGrowthRate';
                    dUpdater.sourceMagnitude = dUpdater.sourceMagnitude';
                    dUpdater.sourceInitialTime = dUpdater.sourceInitialTime';
                    dUpdater.sourceEndTime = dUpdater.sourceEndTime';
                    
            end % switch
            
        end
        

    end
    

end
