%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.


function loadOrganismXML(org, fileName )

    % read xml file
    try
        XML_contents = XML_Reader(fileName);
    catch ME
        fprintf('loadTTXML:invalid file. Invalid input file: %s \n', file);
        outME = addCause( ...
            MException( ...
            'MATLAB:loadTTXML:invalid file', ...
            'Invalid input file: %s', fileName), ...
             ME);
       throw(outME);  
    end      
     
    % specify properties of organism in subject to value of tag
    switch char(XML_contents.Tag)
        %% specify TT properties
        case {'TT'}       
             specifyTTProperties(org, XML_contents);
        %% specify network properties    
        case 'network'
            specifyPPIProperties(org, XML_contents);
        case 'cells_info' 
           setCellInfo(org, XML_contents);            
                
    end % switch  
       
end

%% Specifies cell information
function setCellInfo(org, XML_contents)
    NCells = length(XML_contents.CellList); % number of cells
    
    setCellTimes = false;
    DivisionTimes = zeros(NCells,1);
    StartTimes = zeros(NCells,1);
    
    % get names of cell names from organism and XML
    NamesFromOrg = {};
    NamesFromXML = {};
    if (NCells == org.NCells)
        for i = 1:NCells
            cell = XML_contents.CellList(i);
            NamesFromXML = [NamesFromXML, cell.Name];
            cell = org.OrgCells(i);
            NamesFromOrg = [NamesFromOrg, cell.Name];
        end
    else
        throw(MException(...
            'loadOrganismXML:setCellInfo: number of cells in organism and in XML. Check cells in XML files.'));
    end
    % check if names from organism and xml are the same
    if (checkCellNames(NamesFromOrg, NamesFromXML))
        % specify neighboring cell and progenies info
        Neighbors = zeros(NCells);
        Progenies = zeros(NCells);
        for i = 1:NCells
            name = NamesFromOrg(i);
            cell = CellInfo.empty;
            % find this name in CellList
            for j = 1:NCells
                cell = XML_contents.CellList(j);
                if (strcmp(name, cell.Name))
                    break;
                end
            end % for j
            if (isa(cell, 'CellInfo'))
                % names of rigth and left neighbors of cell
                index_right = getIndexByName(org.OrgCells, cell.RightNeighbor);
                index_left = getIndexByName(org.OrgCells, cell.LeftNeighbor);
                if (index_right>0)
                    Neighbors(i, index_right) = 1;
                    Neighbors(index_right,i) = 1;
                end %if
                if (index_left>0)
                    Neighbors(i, index_left) = 1;
                    Neighbors(index_left,i) = 1;
                end %if

                % if this cell is progeny of other cell
                index_progeny = getIndexByName(org.OrgCells, cell.ProgenyOf);
                if (index_progeny > 0)
                    Progenies(i,index_progeny) = 1;
                end

                % set distances for this cell                
                org.OrgCells(i,1).Type.setSignalInfo(...
                    cell.SignalTypes,...
                    cell.DistanceToSource,...
                    str2double(cell.GrowthRate));

                % find the times of cell division
                setCellTimes = true;
                DivisionTimes(i,1) = str2double(cell.DivisionTime);
                StartTimes(i,1) = str2double(cell.StartTime);
                
                
            end % 
        end %for i
        
        if (setCellTimes)
            org.setCellTimes(DivisionTimes, StartTimes);
        end
        
        % set CellNetwork property of Organism
        org.setCellNetwork(Neighbors);
        % set NeighborCellTypes, NNeighborCells, isNeighborCell properties
        % of OrgCells (CellProperties)
        NamesFromOrg = NamesFromOrg';
        for iCell = 1:NCells
            org.OrgCells(iCell, 1).Type.setNeighborCells(...
                NamesFromOrg(logical(Neighbors(iCell,:)),:), NamesFromOrg);
        end
        % set the progeny of each cell
        % NCells x NCells matrix whose i j entry is true if
        % cell i is a progeny of cell j, and false otherwise
         org.setProgeny(Progenies);
        
    end % if
end % setCellInfo

%% Specifies PPI properties
function specifyPPIProperties(org, XML_contents)
% sort proteins by their types
[InternalProteins, ExternalSignals, Repressors, ...
    Scaffolds, Receptors, NeighborSignals, Transported] = ...
    sortProteins(XML_contents.ProteinsList);

% get their names
ProteinNames = getNames(InternalProteins); % row vector
ExternalSignalNames = getNames(ExternalSignals);  % row vector
RepressorNames = getNames(Repressors); % row vector
ScaffoldNames = getNames(Scaffolds); % row vector
ReceptorNames = getNames(Receptors);

% NProteins x NProteins x 2 matrix with PPI(:,:,1) containing the
% rates for unbound proteins and PPI(:,:,2) the rates for bound proteins
% (negative rates refer to repression)
[PPI,SpontActUnbound, SpontInacUnbound] = ...
    getPPI(InternalProteins, XML_contents.ReactionsList);

% create NReceptors x NExternalSignals matrix indicated which
% proteins are activated by each signal type
ExternalSignalsPI = getExternalSignalsPI(Receptors, ExternalSignals, ...
    XML_contents.ReactionsList);
%NExternalSignals logical vector which is true if the signal acts on neighbor cells
isNeighborSignal = ...
    getNeighborSignals(ExternalSignals,NeighborSignals);

% create matrix of repressor-protein interactions
% indicate which proteins are repressed by repressors
% NProteins x NRepressors x 2 matrix with rates for eachrepressor type
% for Repressors(:,:,1) referring to unbound and (:,:,2) to bound 
RepressorsPI = getRepressorsPI(InternalProteins, Repressors,...
    XML_contents.ReactionsList);

% NProteins x NScaffolds x 2 array with Scaffolds(:,:,1)
% containing binding rates and Scaffolds(:,:,2) containing unbinding rates
% NProteins x NScaffold matrix with self activation rates for bound proteins
% NProteins x NScaffold matrix with inactivation rates for bound proteins
[ScaffoldedProteins,SpontActBound, SpontInacBound] = ...
    getScaffoldedProteins(InternalProteins, Scaffolds,...
    XML_contents.ReactionsList);

% NReceptors vector with spontaneous activation rates for receptors
% NReceptors vector with inactivation rates for receptors
[SpontActReceptors, SpontInacReceptors] = ...
    getSpontReceptors(Receptors, XML_contents.ReactionsList);

% NProteins x NReceptors x 2 matrix indicating the
% proteins that are activated by the receptors
% (:,:,1) for unbound and (:,:,2) for bound 
Receptors2Target = getReceptorMatrix(InternalProteins, Receptors,...
    XML_contents.ReactionsList);

% create NProteins vector indicating if an active protein is transported to
% the nucleus
canTransport = getTransportProteins(InternalProteins, Transported);
% vector with transport times
transTimes = getTransportTimes(char(ProteinNames), XML_contents.ReactionsList);

% specify parameters for cell properties
NCells = length(XML_contents.CellNames);
% get active and inactive protiens concentration
[ProteinConcAc, ProteinConcIn] = getAmounts(InternalProteins); % column vector
[ExtSignalsConcAc, ExtSignalsConcIn] = getAmounts(ExternalSignals); % column vector
[RepressorsConcAc, RepressorsConcIn] = getAmounts(Repressors);% column vector
[ReceptorsConcAc, ReceptorsConcIn] = getAmounts(Receptors); % column vector
[ScaffoldsConcAc, ScaffoldsConcIn] = getAmounts(Scaffolds);% column vector
[NeighborSignalsConc, NeighborSignalsConcIn]= getAmounts(NeighborSignals); % column vector

% if organism cells are not defined
if (isempty(org.OrgCells))
    % if cell names were read from file
    if (~isempty(XML_contents.CellNames))
        % create vector of cells
        orgCells = repmat(Cells, NCells, 1);
        CellNames = {};
        for i = 1:NCells
            % create CellProperties
            type = 'default';
            CellType = CellProperties(type, ...
                ProteinNames, PPI, ...
                ExternalSignalNames, ExternalSignalsPI, ...
                isNeighborSignal, ...
                RepressorNames, RepressorsPI, ...
                ScaffoldNames, ScaffoldedProteins,...
                ReceptorNames, Receptors2Target);
            
            CellType.setCanTransport(canTransport);
            CellType.setTransportTime(transTimes);
            
            CellType.setSpontaneousRates(SpontActReceptors, SpontActUnbound, ...
                SpontActBound, SpontInacReceptors,...
                SpontInacUnbound, SpontInacBound);
            
            CellNames = [CellNames, char(XML_contents.CellNames(i))];
            
            CellName = char(XML_contents.CellNames(i));
            % add cell into organism
            orgCells(i, 1) = Cells(CellName, CellType);
            
            
            % initiate protein concentrations
            orgCells(i).setState(CellState(...
                (ProteinConcIn+ProteinConcAc), ...
                ProteinConcAc, ...
                ExtSignalsConcIn, ...
                RepressorsConcIn, ...
                ScaffoldsConcIn, ...
                ReceptorsConcIn));
            
            
        end
      
        % initialise organism cells
        org.setCellTypes(CellType);
        org.setOrgCells(orgCells);

    end
end
end % specifyPPIProperties

%% Specifies TT properties by tag
function specifyTTProperties(org, XML_contents)
   
    NSpecies = length(XML_contents.ProteinsList); % number of species
    NReactions = length(XML_contents.ReactionsList); % number of reactions
    % get species names
    SpeciesNames = getNames(XML_contents.ProteinsList, NSpecies); % column vector
    % get species concentration (column vector)
    [SpeciesAmountAc,SpeciesAmount] = getAmounts(XML_contents.ProteinsList, NSpecies); % column vector
    % get reaction rates (row vector)
    ReactionRates = getReactionRates(XML_contents.ReactionsList, NReactions);
    % get D matrix - NSpecies x NReactions matrix
    % -1 if species is reactant
    % 1 if species is product
    D = getD(XML_contents.ReactionsList, SpeciesNames, NReactions, NSpecies);
    % get transport time
    for i = 1:org.NCells
        if (isa(org.OrgCells(i).Type,'CellProperties'))
            % create SSA_Properties object
            tt_obj = SSA_Properties;
            tt_obj.Names = SpeciesNames;
            tt_obj.Population = SpeciesAmount;
            tt_obj.ReactionRates = ReactionRates;
            tt_obj.StoichMatrix = D;
            tt_obj.V = convert2Double(XML_contents.Volume);
            org.OrgCells(i,1).Type.addTTObject(tt_obj);
            % convert transport rate into time (s)
            %% TODO Create matrix in CellState class and think how to fill it
            org.OrgCells(i,1).State.addTTPopulation(SpeciesAmount);
            
         end % if
    end % for i

end % function specifyTTProperties

%% returns index in the list of cells
function index = getIndexByName(cellList, name)
    index = 0;
    for i = 1:length(cellList)
        if (strcmp(cellList(i).Name, name))
            index = i;
            break;
        end % if
    end % for i
end % getIndexByName

%% Compares cell names from organism and xml
function result = checkCellNames(fromOrg, fromXML)
    result = false;
    for i = 1:length(fromOrg)
        check = false;
        for j = 1:length(fromOrg)
            if (strcmp(fromOrg(i), fromXML(j)))
                check = true;
                break;
            end % if
        end % for j
        if (~check)
            fprint('loadOrganismXML:checkCellNames: couldn"t find name ', ...
                fromOrg(i));
             throw(MException('loadOrganismXML:checkCellNames: couldn"t find name ', ...
                fromOrg(i)));
        end % if        
    end %for i 
    result = true;
end% checkCellNames

%% Get names from array of Protein objects
function ProteinNames = getNames(proteins, NProt)
    if (nargin <2)
        NProt = length(proteins);
    end % if
    ProteinNames = {};       % output array of internal protein names       
    for i = 1:NProt
        % get a protein form the list
      protein = proteins(i); 
      % check if protein is object of Protein class
      if (isa(protein,'Protein'))
        prot_name = char(protein.name);
        ProteinNames = [ProteinNames, prot_name];         
      end % if
    end % for i
    ProteinNames = ProteinNames'; % create column vector
end % function getNames

%% Get amount from array of Protein objects
function [amountActive, amountInactive] = getAmounts(proteins, NProt)
    
    if (nargin <2)
        NProt = length(proteins);
    end % if
    amountActive = zeros(NProt, 1); % output array of internal active protein amount  
    amountInactive = zeros(NProt, 1); % output array of internal inactive protein amount  
    for i = 1:NProt
        % get a protein form the list
      protein = proteins(i); 
      % check if protein is object of Protein class
      if (isa(protein,'Protein'))
        prot_amount = convert2Double(protein.amountActive);
        amountActive(i, 1) = prot_amount;
        prot_amount = convert2Double(protein.amountInactive);
        amountInactive(i, 1) = prot_amount;
      end % if
    end % for i     
end % function getAmounts

%% Get reaction rates from array of Reaction objects
function Rates = getReactionRates(reactions, NReactions)
    Rates = zeros(1, NReactions); % row vector of rates
    
    for i = 1:NReactions
        % get a reaction
        reaction = reactions(i);
        % check if reaction is object of Reaction class
        if (isa(reaction, 'Reaction'))
            rate = convert2Double(char(reaction.Parameter));
            Rates(1, i) = rate;
        end % IF        
    end % FOR i
end % getReactionRates

%% Get rates of transport reactions
function transportTimes = getTransportTimes(InternalProteins, reactions)

    NReactions = length(reactions);
    NInternalProteins = length(InternalProteins);
    
    transportTimes = -1*ones(NInternalProteins,1);
    
    for i = 1:NReactions
        % get a reaction
        r = reactions(i);
        % check if reaction is object of Reaction class
        if (isa(r, 'Reaction') && strcmp(r.Type,'transport'))
            % find index of transported protein in the protein list
            ind = findType(char(r.Reactants),InternalProteins);
            transportTimes(ind,1) = convert2Double(char(r.Parameter));
        end % IF        
    end % FOR i
end % getTransportTimes

%% create D matrix 
% NSpecies x NReactions matrix
% -1 if species is reactant
% 1 if species is product

function D = getD(reactions, species, NReactions, NSpecies)
    
    D = zeros(NSpecies, NReactions);
    % iterate reactions
    for i=1:NReactions
        % get reaction
       r = reactions(i);
       if isa(r,'Reaction')  % if it's object of Reaction type 
          % if 'reactants' property is not empty 
          if (~isempty(r.Reactants))  
           % interate reactants
           for k = 1:length(r.Reactants)
               % get name of reactant
               reactant = char(r.Reactants(k));
               % search this name into species array
               for j=1:NSpecies                   
                   p = species(j);
                   if (strcmp(p,reactant))
                    D(j,i) = -1;
                    break;
                   end
               end % for j
           end % k
         end % if 
         % if 'products' property is not empty
        if (~isempty(r.Products))
            % interate products
           for k = 1:length(r.Products)
               % get name of reactant
               product = char(r.Products(k));
               % search this name into species array
               for j=1:NSpecies                   
                   p = species(j);
                   if (strcmp(p,product))
                    D(j,i) = 1;
                    break;
                   end
               end % for j
           end % k
        end % if
       end % if
    end % for i
end % getD

%% Sort proteins by their types
function [InternalProteins, ExternalSignals, Repressors, Scaffolds,...
     Receptors, NeighborSignals, Transported] = ...
    sortProteins(proteins)
%  Sort list of objects of Protein type by their type
%
%  Input Arguments:
%  (1) array of objects of Protein type
%
%  Return:
%  (1) array of proteins with type 'activator', 'repressor' 
%  (2) array of proteins with type 'external_signal', 'neighbor_signal'
%  (3) array of proteins with type 'repressor'
%  (4) array of proteins with type 'scaffold'
%  (5) array of proteins with type 'receptor'
%  (6) array of protiens with type 'neighbor_signal'
%  (7) array of proteins with type 'TF'

    NProt = length(proteins);
    InternalProteins = Protein.empty;% output array of internal protein 
    ExternalSignals = Protein.empty; % output array of external signal
    Repressors = Protein.empty;      % output array of repressors    
    Scaffolds = Protein.empty;       % output array of scaffolds
    Receptors = Protein.empty;       % output array of receptors 
    NeighborSignals = Protein.empty; % output array of neighbor signals 
    Transported = Protein.empty;     % output array of transported proteins
     
    for i = 1:NProt
        % get a protein form the list
      protein = proteins(i); 
      if (isa(protein,'Protein'))
          prot_type = char(protein.type);
          switch prot_type
              case 'scaffold'          
                Scaffolds = [Scaffolds, protein];
              case 'TF'
                InternalProteins = [InternalProteins, protein];
                Transported = [Transported, protein];
              case 'external_signal'
                ExternalSignals = [ExternalSignals, protein];
              case 'repressor'
                InternalProteins = [InternalProteins, protein]; 
                Repressors =[Repressors, protein];
              case 'activator'
                InternalProteins = [InternalProteins, protein]; 
              case 'neighbor_signal'
                NeighborSignals = [NeighborSignals, protein];
                ExternalSignals = [ExternalSignals, protein];
              case 'receptor'
                Receptors = [Receptors, protein];
              case 'receptor_TF'
                InternalProteins = [InternalProteins, protein];
                Receptors = [Receptors, protein];
                Transported = [Transported, protein];
          end % switch
      end % if
    end % for i
end % function sortProteins
%% create PPI
   % PPI is NProteins x NProteins x 2 matrix of protein-protein interaction 
   % ith row protein interact with jth column protein
   % positive value is rate of activation
   % negtive value is rate of repressesion
   % PPI(:,:,1) containing the rates for unbound proteins and 
   % PPI(:,:,2) the rates for bound proteins
   % NProteins vector with spontaneous activation rates for unbound proteins
   % NProteins vector with inactivation rates for unbound proteins

function [PPI, SpontActUnbound, SpontInacUnbound] = getPPI(proteins,reactions)
 NProteins = length(proteins);
 NReactions = length(reactions);
   PPI = zeros(NProteins,NProteins,2);
   SpontActUnbound = zeros(NProteins,1);
   SpontInacUnbound = zeros(NProteins,1);
   for i=1:NReactions
       r = reactions(i);
       if(isa(r,'Reaction'))
           switch r.Type
               % if it's reaction of activation unbound proteins
               case 'activation'
                   index1 = getProteinIndex(proteins, r.Reactants(1));
                   index2 = getProteinIndex(proteins, r.Products(1));
                   if(index1~=-1 && index2 ~= -1)
                        PPI(index1,index2,1)=convert2Double(char(r.Parameter));
                   end
               % if it's reaction of repression unbound proteins
               case 'repression'
                   index1 = getProteinIndex(proteins, r.Reactants(1));
                   index2 = getProteinIndex(proteins, r.Products(1));
                   if(index1~=-1 && index2 ~= -1)
                        PPI(index1,index2,1)=-convert2Double(char(r.Parameter));
                   end
               case 'activation_bound'
                   index1 = getProteinIndex(proteins, r.Reactants(1));
                   index2 = getProteinIndex(proteins, r.Products(1));
                   if(index1~=-1 && index2 ~= -1)
                        PPI(index1,index2,2)=convert2Double(char(r.Parameter));
                   end
               % if it's reaction of repression unbound proteins
               case 'repression_bound'
                   index1 = getProteinIndex(proteins, r.Reactants(1));
                   index2 = getProteinIndex(proteins, r.Products(1));
                   if(index1~=-1 && index2 ~= -1)
                        PPI(index1,index2,2)=-convert2Double(char(r.Parameter));
                   end
                case 'spontaneous_unbound_activation'
                   index1 = getProteinIndex(proteins, r.Reactants(1));
                   if(index1~=-1)
                        SpontActUnbound(index1,1)=convert2Double(char(r.Parameter));
                   end
               % if it's reaction of repression unbound proteins
               case 'spontaneous_unbound_inactivation'
                   index1 = getProteinIndex(proteins, r.Reactants(1));
                   if(index1~=-1)
                        SpontInacUnbound(index1,1)=convert2Double(char(r.Parameter));
                   end
          end
      end
   end % for i
   
end % getPPI

%% returns index of protein in protein vector
function index = getProteinIndex(proteins, proteinName)
    index = -1;
    for i = 1:length(proteins)
        if (isa(proteins(i),'Protein') && strcmp(proteins(i).name,proteinName))
            index = i;
            return;
        end
    end
end

%% create matrix external signal-protein interactions
function ExternalSignalsPI = getExternalSignalsPI(...
    Receptors, ExternalSignals, reactions) 
    NProteins = length(Receptors);
    NExternalSignals = length(ExternalSignals);
    ExternalSignalsPI = zeros(NProteins, NExternalSignals);
    NReactions = length(reactions);
    for i=1:NReactions
       r = reactions(i);
       if(isa(r,'Reaction')&&(strcmp(r.Type,'ES_activation')))
            index1 = getProteinIndex(Receptors, r.Products(1));
            index2 = getProteinIndex(ExternalSignals, r.Reactants(1));
            if(index1~=-1 && index2 ~= -1)
               ExternalSignalsPI(index1,index2)=convert2Double(char(r.Parameter));
            end
       end
   end % for i
end % getExternalSignalsPPI

%% returns NExternalSignals logical vector 
% which is true if the signal acts on neighbor cells            
function isNeighborSignals = getNeighborSignals(...
    ExternalSignals, NeighborSignal) 
    NExternalSignals = length(ExternalSignals);
    NNeighborSignals = length(NeighborSignal);
    isNeighborSignals = zeros(NExternalSignals, NNeighborSignals);
    for i=1:NNeighborSignals
       p_i = NeighborSignal(i);
       if (isa(p_i,'Protein') && ... % if it's object of Protein type               
           ~isempty(p_i.name))  % if 'name' property of protein is not empty
            index = getProteinIndex(ExternalSignals, p_i.name);
            if (index ~=-1)
                isNeighborSignals(index,i) = 1;
            end
       end % if 
    end % for i
end % getExternalNeighborSignals
%% create NProteins x NReceptors x 2 matrix 
% indicating the proteins that are activated by the receptors
% (:,:,1) for unbound and (:,:,2) for bound 
function Receptors2Target = getReceptorMatrix(InternalProteins, Receptors,...
    reactions)

    NProteins = length(InternalProteins);
    NReceptors = length(Receptors);
    Receptors2Target = zeros(NProteins, NReceptors,2);
    NReactions = length(reactions);
    for i=1:NReactions
       r = reactions(i);
       if(isa(r,'Reaction')&&(strcmp(r.Type,'activation_receptor')))
            index1 = getProteinIndex(InternalProteins, r.Products(1));
            index2 = getProteinIndex(Receptors, r.Reactants(1));
            if(index1~=-1 && index2 ~= -1)
               Receptors2Target(index1,index2,1)=convert2Double(char(r.Parameter));
            end
       end
        if(isa(r,'Reaction')&&(strcmp(r.Type,'activation_receptor_bound')))
            index1 = getProteinIndex(InternalProteins, r.Products(1));
            index2 = getProteinIndex(Receptors, r.Reactants(1));
            if(index1~=-1 && index2 ~= -1)
               Receptors2Target(index1,index2,2)=convert2Double(char(r.Parameter));
            end
        end
    end % for i
end % getReceptorMatrix

%% create NProteins vector indicating the transport proteins
function canTransport = getTransportProteins(InternalProteins, Transported)
    NInternalProteins = length(InternalProteins);
    NTF = length(Transported);
    canTransport = zeros(NInternalProteins,1);
    for i=1:NTF
       p_i = Transported(i);
       if (isa(p_i,'Protein') && ... % if it's object of Protein type               
           ~isempty(p_i.name))  % if 'name' property of protein is not empty
           % get name of  protein
           index = getProteinIndex(InternalProteins, p_i.name);
           if (index ~=-1)
            canTransport(index,1) = 1;
           end
       end % if 
    end % for i
end

%% create matrix repressor-protein interactions
% NProteins x NRepressors x 2 matrix with rates for eachrepressor type
% for Repressors(:,:,1) referring to unbound and (:,:,2) to bound
function RepressorsPI = getRepressorsPI(InternalProteins, Repressors, reactions)
    NProteins = length(InternalProteins);
    NRepressors = length(Repressors);
    RepressorsPI = zeros(NProteins, NRepressors,2);
    NReactions = length(reactions);
    for i=1:NReactions
       r = reactions(i);
       if(isa(r,'Reaction')&&(strcmp(r.Type,'repression')))
            index1 = getProteinIndex(InternalProteins, r.Products(1));
            index2 = getProteinIndex(Repressors, r.Reactants(1));
            if(index1~=-1 && index2 ~= -1)
               RepressorsPI(index1,index2,1)=convert2Double(char(r.Parameter));
            end
       end
        if(isa(r,'Reaction')&&(strcmp(r.Type,'repression_bound')))
            index1 = getProteinIndex(InternalProteins, r.Products(1));
            index2 = getProteinIndex(Repressors, r.Reactants(1));
            if(index1~=-1 && index2 ~= -1)
               RepressorsPI(index1,index2,2)=convert2Double(char(r.Parameter));
            end
        end
    end % for i

end % getRepressorsPI

%% create matrix which indicates what proteins are bound in scaffold
% NProteins x NScaffolds x 2 array with ScaffoldedProteins(:,:,1) containing binding
% rates and ScaffoldedProteins(:,:,2) containing unbinding rates
% NProteins x NScaffold matrix with self activation rates for bound proteins
% NProteins x NScaffold matrix with inactivation rates for bound proteins
function [ScaffoldedProteins,SpontActBound, SpontInacBound] = ...
    getScaffoldedProteins(InternalProteins, Scaffolds, reactions)
    
    NProteins = length(InternalProteins);
    NScaffolds = length(Scaffolds);
    ScaffoldedProteins = zeros(NProteins, NScaffolds, 2);
    SpontActBound = zeros(NProteins, NScaffolds);
    SpontInacBound = zeros(NProteins, NScaffolds);
    NReactions = length(reactions);
    % fill in ScaffoldedProteins
    for i=1:NReactions
       r = reactions(i);
       if(isa(r,'Reaction'))
           switch r.Type
               % if it's reaction of activation unbound proteins
               case 'scaffold_binding'
                   index2 = getProteinIndex(Scaffolds, r.Products(1));
                   for j=1:length(r.Reactants) 
                    index1 = getProteinIndex(InternalProteins, r.Reactants(j));
                    if(index1~=-1 && index2 ~= -1)
                       ScaffoldedProteins(index1,index2,1)=convert2Double(char(r.Parameter));
                    end
                   end
               case 'scaffold_unbinding'
                index2 = getProteinIndex(Scaffolds, r.Reactants(1));
                for j=1:length(r.Products)
                    index1 = getProteinIndex(InternalProteins, r.Products(j));
                    if(index1~=-1 && index2 ~= -1)
                       ScaffoldedProteins(index1,index2,2)=convert2Double(char(r.Parameter));
                    end
                end
           end% switch
       end% if
    end % for i
    % fill in SpontActBound and SpontInacBound
    for i=1:NReactions
       r = reactions(i);
       if(isa(r,'Reaction'))
           switch r.Type
               case 'spontaneous_bound_activation'
                   % index in Proteins vector
                   index1 = getProteinIndex(InternalProteins, r.Reactants(1));
                   % index of scaffold where this protein binds
                   index2 = -1;
                   for j = 1:NScaffolds
                       if (ScaffoldedProteins(index1,j,1)>0)
                           index2 = j;
                           break;
                       end
                   end
                   if(index1~=-1)
                        SpontActBound(index1,index2)=convert2Double(char(r.Parameter));
                   end
               % if it's reaction of repression unbound proteins
               case 'spontaneous_bound_inactivation'
                   index1 = getProteinIndex(InternalProteins, r.Reactants(1));
                   % index of scaffold where this protein binds
                   index2 = -1;
                   for j = 1:NScaffolds
                       if (ScaffoldedProteins(index1,j,1)>0)
                           index2 = j;
                           break;
                       end
                   end
                   if(index1~=-1)
                        SpontInacBound(index1,index2)=convert2Double(char(r.Parameter));
                   end
           end
       end
     end % for i

end % getScaffoldedProteins

%% retunrs vectors with spontaneous activation and inactivation rates of
%% receptors
% NReceptors vector with spontaneous activation rates for receptors
% NReceptors vector with inactivation rates for receptors
function [SpontActReceptors, SpontInacReceptors] = ...
    getSpontReceptors(Receptors, reactions)
    NReceptors = length(Receptors);
    SpontActReceptors = zeros(NReceptors,1);
    SpontInacReceptors = zeros(NReceptors,1);
    NReactions = length(reactions);
    for i=1:NReactions
       r = reactions(i);
       if(isa(r,'Reaction'))
           switch r.Type
               case 'spontaneous_receptor_activation'
                   index1 = getProteinIndex(Receptors, r.Reactants(1));
                   if(index1~=-1)
                        SpontActReceptors(index1,1)=convert2Double(char(r.Parameter));
                   end
               % if it's reaction of repression unbound proteins
               case 'spontaneous_receptor_inactivation'
                   index1 = getProteinIndex(Receptors, r.Reactants(1));
                   if(index1~=-1)
                        SpontInacReceptors(index1,1)=convert2Double(char(r.Parameter));
                   end

           end % switch
       end % if(isa(r,'Reaction'))
    end % for i
end


