%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% classdef Organism
%  Contains the cell properties and states.
%
%  Dependencies:
%  (1) CellProperties
%  (2) CellContents
%
%  Author:  Elana J. Fertig
%  Date:    18 November 2008
%
%  Modification History:
%  25Nov08  EJF  Modified for simplicity, so that external routines perform
%                the majority of modification and initialization of the
%                organism (note, organism only contains state at one time).
%
classdef Organism < handle
    
    properties (SetAccess = private, GetAccess = public)
       
        %% cell attributes
        NCellTypes  % Number of distinct cell types
        CellTypes   % Array of CellProperty classes describing the cell
                    % types available within the organism
        
        %% cells
        NCells      % Number of distinct cells in the organism
        OrgCells    % Array of Cells classes describing cells of type
                    % CellTypes within the organism                    
        CellNetwork % NCells x NCells matrix that describes the topology
                    % of the network of cells
                    
        % information about progeny
        Progeny       % NCells x NCells matrix whose ij entry is 1 if cell i is a progeny of cell j and 0 otherwise
        CellStartTime % NCells vector indicating when each cell starts
        DivisionTime  % NCells vector the indicates the division time of each cell

    end
    
    methods
        
        function orgout = Organism()
        %% org = Organism()
        %  Default constructor for the class Organism.
        %
        %  Input Arguments:
        %  (none)
        %
        %  Return:
        %  (1) An instantiation of class Organism
        %
        end
        
        function setCellTypes(org, CellTypes)
            if (~isa(CellTypes, 'CellProperties'))
                throw(...
                    MException(...
                    'MATLAB:Organism:invalidType', ...
                    'Cannot set cell types of organism as type %s.', ...
                    class(CellTypes)));
            end
            
            org.NCellTypes = size(CellTypes, 1);
            org.CellTypes  = CellTypes;
            
        end
        
        function setProgeny(org, Progeny)
            if (size(Progeny,1)==org.NCells && size(Progeny,2)==org.NCells)
                org.Progeny = logical(Progeny);
            end
        end
        
        function setCellTimes(org, divisionTimes, cellStartTime)
            
            if (length(divisionTimes)~= org.NCells)
                
                throw(MException(...
                    'MATLAB:Organism:invalidType', ...
                    'Must specify division times for all %d cells, only specified %d.', ...
                    org.NCells, length(divisionTimes)));
            end
            org.DivisionTime = divisionTimes;
            
            if (length(cellStartTime)~= org.NCells)
                throw(MException(...
                    'MATLAB:Organism:invalidType', ...
                    'Must specify starting times for all %d cells, only specified %d.', ...
                    org.NCells, length(cellStartTime)));
            end
            
            org.CellStartTime = cellStartTime;
        end
            
        function setOrgCells(org, orgCells, CellNetwork)

            if (~isa(orgCells, 'Cells'))
                throw(...
                    MException(...
                    'MATLAB:Organism:invalidType', ...
                    'Cannot set cells of organism as type %s.', ...
                    class(orgCells)));
            end
            
            org.NCells   = uint32(size(orgCells, 1));
            org.OrgCells = orgCells;
            
            if (isempty(org.Progeny))
                org.Progeny = false(org.NCells);
            end
            
            if (nargin == 2)
                if (isempty(org.CellNetwork))
                    org.initiateEmptyCellNetwork();
                elseif (numel(org.CellNetwork) ~= org.NCells * org.NCells)
                    org.initiateEmptyCellNetwork();
                end
            elseif (nargin > 2)
                org.setCellNetwork(CellNetwork);
            end
        end
        
        function setCellNetwork(org, CellNetwork)
            if (numel(CellNetwork) ~= org.NCells*org.NCells)
                if (org.NCells ~= 0)
                    warning('MATLAB:Organism:invalidDim', ...
                        'Cell network has %d entries but requires %d. %s.', ...
                        numel(CellNetwork), ...
                        org.NCells*org.NCells, ...
                        'Initializing as zero');
                    org.initiateEmptyCellNetwork;
                end
            end
            
            org.CellNetwork = int8(CellNetwork);
        end
        
        function cellStates = getCellStates(org)
            cellStates = repmat(CellState,org.NCells,1);
            for iCell=1:org.NCells
                cellStates(iCell) = CellState(org.OrgCells(iCell,1).State);
            end
        end
        
        function setCellStates(org, CellStates)
            if (size(CellStates, 1) ~= org.NCells)
                throw(...
                    MException(...
                    'MATLAB:Organism:invalidDim', ...
                    'States of only %d of %d cells specified.  %s', ...
                    size(CellStates, 1), ...
                    org.NCells, ...
                    'Cannot update the state.'));
            end
            
            for iCell = 1:org.NCells
                org.OrgCells(iCell,1).setState(CellStates(iCell, 1));
            end
        end
        
        function phenotype = getPhenotype(org)
            % todo - come up with phenotype definition
        end
        
    end
    
    methods (Access = private)
        function initiateEmptyCellNetwork(org)
            if (org.NCells > 0)
                org.CellNetwork = zeros(org.NCells, 'int8');
            end
        end
    end
    
end
