%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.


classdef OrganismHistory < handle
    
    properties (GetAccess = public, SetAccess = private)
        timeTracker     % tracks the time
        currentTime     % marks the current time
        orgCellTracker  % array of cell states for each cell and time
        cellNames       % name of cells
        cellTypes       % types of cells
        startActiveTime % first active time for each cell
        endActiveTime   % time of inactivity for each cell
        isActive        % indicates if the cell is active at the current time
    end
    
    methods (Access = public)
        function orgHist = OrganismHistory(org, initTime)
            
            if (nargin > 0)
                
                if (isa(org, 'Organism'))

                    orgHist.cellNames = cell(org.NCells, 1);
                    orgHist.cellTypes = repmat(CellProperties, org.NCells, 1);
                    for iCell = 1:org.NCells
                        orgHist.cellNames(iCell, 1) = ...
                            cellstr(org.OrgCells(iCell).Name);
                        orgHist.cellTypes(iCell, 1) = ...
                            org.OrgCells(iCell).Type;
                    end
                    
                    orgHist.cellNames = char(orgHist.cellNames);

                    orgHist.currentTime = 1;

                    % set the initial time of the simulation
                    if (nargin < 2)
                        orgHist.timeTracker(orgHist.currentTime) = 0.;
                    else
                        orgHist.timeTracker(orgHist.currentTime) = ...
                            double(initTime);
                    end

                    % initialize the state tracker of the organism
                    orgHist.orgCellTracker = CellState.empty;
                    for iCell = 1:org.NCells
                        orgHist.orgCellTracker(iCell,orgHist.currentTime) = ...
                            org.OrgCells(iCell,1).State;
                    end
                    
                    % initialize the arrays for active times of each cell
                    orgHist.startActiveTime = zeros((org.NCells),1);
                    orgHist.endActiveTime   = zeros((org.NCells),1);
                    orgHist.isActive        = false((org.NCells),1);
                    
                elseif (isa(org,'OrganismHistory'))
                    orgHist.timeTracker     = org.timeTracker;
                    orgHist.orgCellTracker  = org.orgCellTracker;
                    orgHist.cellNames       = org.cellNames;
                    orgHist.cellTypes       = org.cellTypes;
                    orgHist.startActiveTime = org.startActiveTime;
                    orgHist.endActiveTime   = org.endActiveTime;
                else
                    thows( MException(...
                        'MATLAB:OrganismHistory:invalidInit', ...
                        'Class %s cannot instantiate a history tracker', ...
                        class(org)) );
                end
            

                
            end
            
        end
        
        function updateHist(orgHist, org, newTime)
            if (newTime < orgHist.timeTracker(orgHist.currentTime))
                throw ( MException(...
                    'MATLAB:OrganismHistory:invalidTime', ...
                    'Cannot update state from time %f to time %f', ...
                    orgHist.timeTracker(orgHist.currentTime), newTime) );
            end
            
            switch class(org)
                case 'Organism'
                    for iCell = 1:org.NCells
                        if (org.OrgCells(iCell,1).State.Active)
                            if (~orgHist.isActive(iCell,1))
                                orgHist.isActive(iCell,1) = true;
                                orgHist.startActiveTime(iCell,1) = newTime;
                            end
                            orgHist.orgCellTracker(iCell,...
                                orgHist.currentTime+1) = ...
                                CellState(org.OrgCells(iCell,1).State);
                            orgHist.endActiveTime(iCell,1) = newTime;
                        else
                            orgHist.isActive(iCell,1) = false;
                        end
                    end
                case 'CellState'
                    for iCell = 1:length(org)
                        if (org(iCell).Active)
                            if (~orgHist.isActive(iCell,1))
                                orgHist.isActive(iCell,1) = true;
                                orgHist.startActiveTime(iCell,1) = newTime;
                            end
                      
                            orgHist.orgCellTracker(iCell, ...
                                orgHist.currentTime+1) = ...
                                CellState(org(iCell));
                            orgHist.endActiveTime(iCell,1) = newTime;
                        else
                            orgHist.isActive(iCell,1) = false;
                        end
                    end
                otherwise
                    thows( MException(...
                        'MATLAB:OrganismHistory:invalidInit', ...
                        'Class %s cannot update a history tracker', ...
                        class(org)) );
            end
            
            orgHist.currentTime = orgHist.currentTime + 1;
            orgHist.timeTracker(orgHist.currentTime) = newTime;
            
        end
        
        % [outState, time] = orgHist.getState(cell, type, name)
        %  Obtain the time history for specific cell(s) and state types.
        %
        %  Input Arguments:
        %  (1) Name of cell(s) 
        %  (2) Name of state type (optional)
        %  (3) Name of type within state (option; requires argument 2)
        %
        %  Return:
        %  (1) Array with desired state history
        %  (2) Simulation time of the state (optional)
        %
        %  TODO:
        %  (1) For now, specific states and types can only be returned for 
        %      a single cell.  This requirement should be relaxed.
        %
        function [outState, timeState] = getState(orgHist, cellName, ...
                stateType, stateName)
            
             stateType = getCellPropertyByName(stateType);
            % TT must specify a stateName
            if (nargin > 2)
                if (nargin <4 && strcmpi(stateType, 'TT_Specie'))
                 warning('MATLAB:OrganismHistory:getState:invalidCell', ...
                    'A species name should be specified for TT species');
                 return;
                end
               
            end
            
            % specify the output simulation time
            timeState = orgHist.timeTracker;
            
            % find the cell(s) in the array for which to return the state
            try
                cellList = findType(char(cellName), orgHist.cellNames);
            catch ME
                warning('MATLAB:OrganismHistory:invalidCell', ...
                    'Cannot return requested cells %s', ...
                    cellName);
                warning(getReport(ME));
                return;
            end
            
            % return the state of the desired cell(s)
            if (nargin == 2)
                outState = orgHist.orgCellTracker(cellList, :);
                return;
            end

            % if more than one cell is specified, return the entire state
            if (numel(cellList) > 1)
                warning('MATLAB:OrganismHistory:invalidDim', ...
                    '%s. %s.', ...
                    'Cannot return specific states for multiple cells', ...
                    'Returning complete state');
                outState = orgHist.orgCellTracker(cellList, :);
                return;
            end
            
            % find the state type to return
            %try
                %tempState = zeros(...
                %    length(orgHist.timeTracker), ...
                %    orgHist.cellTypes(cellList).getNState(stateType));
                %if (isequal(strtok(getCellPropertyByName(stateType)), ...
                %    'NeighborSignal'))
                %    tempState = zeros(...
                %        length(orgHist.timeTracker), ...
                %        1);
                %end
                
                
                
                initTime = find(orgHist.timeTracker >= ...
                    orgHist.startActiveTime(cellList),1,'first');
                endTime = find(orgHist.timeTracker < ...
                    orgHist.endActiveTime(cellList),1,'last');
                
                timeState = zeros(length(initTime+1:endTime-1),1);
                
                % if TT, getNState(stateType, stateName)
                stateType = getCellPropertyByName(stateType);
               if (strcmpi(stateType, 'TT_Specie'))
                  tempState = zeros(...
                    length(initTime+1:endTime-1), ...
                    floor(orgHist.cellTypes(cellList).getNState(stateType, stateName)));
               else
                tempState = zeros(...
                    length(initTime+1:endTime-1), ...
                    floor(orgHist.cellTypes(cellList).getNState(stateType)));
               end
               idx = 1;
              % get index of TT object in TT_Species vector
              if (strcmpi(stateType, 'TT_Specie'))
                   % get index of TT object in TT_Species vector
                   index = findType(stateName, {orgHist.cellTypes(cellList).TT_Species(:).Names});
               end
               for iTime = initTime+1:endTime-1
                    %if (mod(iTime,100)==0)
                    %    orgHist.timeTracker(iTime)
                    %end
                    if (orgHist.timeTracker(iTime) > orgHist.startActiveTime(cellList) && ...
                            orgHist.timeTracker(iTime) < orgHist.endActiveTime(cellList))
                      if (strcmpi(stateType, 'TT_Specie')&& ~isempty(index))
                        tempState(idx, :) = ...
                            orgHist.orgCellTracker(cellList, iTime).getState(stateType, index(1));
                      else
                        tempState(idx, :) = ...
                            orgHist.orgCellTracker(cellList, iTime).getState(stateType);
                      end
                        timeState(idx,1) = orgHist.timeTracker(iTime);
                    %else
                    %    tempState(idx, :) = ...
                    %        -999.*ones(1,size(tempState,2));
                    %end
                        idx = idx + 1;
                    end
                end
                
                tempState = tempState(1:idx-1,:);
                timeState = timeState(1:idx-1,1);

            %catch ME
            %    warning('MATLAB:OrganismHistory:invalidType', ...
            %        '%s %s. %s.', ...
            %        'Cannot return type', ...
            %        stateType, ...
            %        'Returning complete state');
            %    warning(getReport(ME));
            %    outState = orgHist.orgCellTracker(cellList, :);
            %    return;
            %end
            
            % return all states of that type if desired
            if (nargin == 3 || isequal(strtok(getCellPropertyByName(stateType)), ...
                    'NeighborSignal'))
                outState = tempState;
                return;
            end
            
            % find the specific states of that type to return
            try
                stateList = findType(char(stateName), ...
                    orgHist.cellTypes(cellList).getStateNames(stateType));
            catch ME
                warning('MATLAB:OrganismHistory:invalidState', ...
                    '%s %s. %s %s.', ...
                    'Cannot return state of', ...
                    stateName, ...
                    'Returning complete state for type', ...
                    stateType);
                warning(getReport(ME));
                outState = tempState;
                return;
            end
            
            % return all states for the specific state type
            if (~isempty(stateList))
                if (~strcmpi(stateType, 'TT_Specie'))
                    outState = tempState(:, stateList);
                else
                    outState = tempState(:, stateList(2,1));
                end
            end
                    
        end
        
        function sumHist = sumSetTimes(hist1, hist2, timeList)
           
           
            % add together multiple histories at a specified set of times
            % for plotting - allows simulations to be at multiple times
            % (and ultimately can be modified to allow cells to be at
            % multiple times)
            if (nargin == 1)
                sumHist = hist1;
                return;
            elseif (nargin == 2)
                if (~isValidAdd(hist1, hist2))
                    throw( MException('MATLAB:OrganismHistory:invalidAdd', ...
                        'Cannot add histories - histories are not compatable.'));
                end
                sumHist = plus(hist1, hist2);
                return;
            elseif (nargin >=2)
                if (~isValidAdd(hist1, hist2))
                    throw( MException('MATLAB:OrganismHistory:invalidAdd', ...
                        'Cannot add histories - histories are not compatable.'));
                end
                
                % find the dimensions of the history
                NCells = size(hist1.orgCellTracker,1);
                NTime = length(timeList);
                
                % initialize the new organism history
                sumHist = OrganismHistory();                
                sumHist.cellNames   = hist1.cellNames;
                sumHist.cellTypes   = hist1.cellTypes;
                sumHist.timeTracker = timeList;
                sumHist.orgCellTracker = repmat(CellState, NCells, NTime);
                
                % find the indices in each of the histories that surround
                % the times in the specified list of times to perform
                % interpolation
                NTime1 = length(hist1.timeTracker);
                timeIdx1 = Time.findTimeIndices(timeList, hist1.timeTracker);
                NTime2 = length(hist2.timeTracker);
                timeIdx2 = Time.findTimeIndices(timeList, hist2.timeTracker);
                
                sumHist.startActiveTime = min(hist1.startActiveTime,...
                    hist2.startActiveTime);
                sumHist.endActiveTime   = max(hist1.endActiveTime, ...
                    hist2.endActiveTime);
                
                % add together the states at the specified times
                for iTime = 1:NTime
                    for iCell = 1:NCells
                        
                        % find the values for the first cell we're
                        % referring to
                        if (timeIdx1(iTime) == 0)
                            cell1 = hist1.orgCellTracker(iCell, 1);
                        elseif (timeIdx1(iTime) >= NTime1)
                            cell1 = hist1.orgCellTracker(iCell, NTime1);
                        else
                            % interpolate between adjacent
                            if (timeIdx1(iTime) == floor(timeIdx1(iTime)))
                                cell1 = hist1.orgCellTracker(iCell,timeIdx1(iTime));
                            else
                                w2 = timeIdx1(iTime) - floor(timeIdx1(iTime));
                                w1 = 1. - w2;
                                cell1 = w1.*hist1.orgCellTracker(iCell,floor(timeIdx1(iTime))) + ...
                                    w2.*hist1.orgCellTracker(iCell,ceil(timeIdx1(iTime)));
                            end
                        end
                        
                                                
                        % find the values for the second cell we're
                        % referring to
                        if (timeIdx2(iTime) == 0)
                            cell2 = hist2.orgCellTracker(iCell, 1);
                        elseif (timeIdx2(iTime) >= NTime2)
                            cell2 = hist2.orgCellTracker(iCell, NTime2);
                        else
                            % interpolate between adjacent
                            if (timeIdx2(iTime) == floor(timeIdx2(iTime)))
                                cell2 = hist2.orgCellTracker(iCell,timeIdx2(iTime));
                            else
                                w2 = timeIdx2(iTime) - floor(timeIdx2(iTime));
                                w1 = 1. - w2;
                                cell2 = w1.*hist2.orgCellTracker(iCell,floor(timeIdx2(iTime))) + ...
                                    w2.*hist2.orgCellTracker(iCell,ceil(timeIdx2(iTime)));
                            end
                        end
                        
                        sumHist.orgCellTracker(iCell, iTime) = cell1 + cell2;
                    end
                end
                
                return;
            end
        end
        
        function validAdd = isValidAdd(hist1, hist2)
            validAdd = true;
            
            if (~isa(hist1,'OrganismHistory') || ~isa(hist2,'OrganismHistory'))
                validAdd = false;
                return;
            end
            
            if (~isequal(hist1.cellNames,hist2.cellNames))
                validAdd = false;
                return;
            end
            
            if (~isequal(hist1.cellTypes,hist2.cellTypes))
                validAdd = false;
                return;
            end
            
            
        end
        
        
        % overload the addition operator (average history of multiple
        % organisms)
        function sumHist = plus(hist1, hist2)
            
            
            if (~isValidAdd(hist1, hist2))
                throw( MException('MATLAB:OrganismHistory:invalidAdd', ...
                    'Cannot add histories - histories are not compatable.'));
            end
            
            [mergedTimeList, indexList] = Time.mergeTimeLists(...
                hist1.timeTracker, hist2.timeTracker, true);
            
            if (numel(mergedTimeList)==0)
                throw( MException('MATLAB:OrganismHistory:invalidTime', ...
                    'Cannot add histories - cell times do not match. hist1: (%f, %f); hist2: (%f, %f).', ....
                    hist1.timeTracker(1), hist1.timeTracker(end), ...
                    hist2.timeTracker(1), hist2.timeTracker(end)));
            end
            
            sumHist = OrganismHistory();
            sumHist.cellNames   = hist1.cellNames;
            sumHist.cellTypes   = hist1.cellTypes;
            sumHist.timeTracker = mergedTimeList;
            
            sumHist.startActiveTime = min(hist1.startActiveTime,...
                hist2.startActiveTime);
            sumHist.endActiveTime   = max(hist1.endActiveTime, ...
                hist2.endActiveTime);
            
            NTime1 = length(hist1.timeTracker);
            NTime2 = length(hist2.timeTracker);
            
            sumHist.orgCellTracker = repmat(CellState, size(hist1.orgCellTracker,1), length(mergedTimeList));
            
            for iTime = 1:length(sumHist.timeTracker)
                for iCell = 1:size(hist1.orgCellTracker,1)
                    
                    % find the state for the first organism
                    if (indexList(iTime,1) == 0)
                    	cell1 = hist1.orgCellTracker(iCell, 1);
                    elseif (indexList(iTime,1) >= NTime1)
                    	cell1 = hist1.orgCellTracker(iCell, NTime1);
                    else
                        % interpolate between adjacent
                        if (indexList(iTime,1) == floor(indexList(iTime,1)))
                        	cell1 = hist1.orgCellTracker(iCell,indexList(iTime,1));
                        else
                        	w2 = indexList(iTime,1) - floor(indexList(iTime,1));
                            w1 = 1. - w2;
                            cell1 = w1.*hist1.orgCellTracker(iCell,floor(indexList(iTime,1))) + ...
                                    w2.*hist1.orgCellTracker(iCell,ceil(indexList(iTime,1)));
                        end
                    end
                    
                    % find the state of the second organism
                    if (indexList(iTime,2) == 0)
                    	cell2 = hist2.orgCellTracker(iCell, 1);
                    elseif (indexList(iTime,2) >= NTime2)
                    	cell2 = hist2.orgCellTracker(iCell, NTime2);
                    else
                        % interpolate between adjacent
                        if (indexList(iTime,2) == floor(indexList(iTime,2)))
                        	cell2 = hist2.orgCellTracker(iCell,indexList(iTime,2));
                        else
                        	w2 = indexList(iTime,2) - floor(indexList(iTime,2));
                            w1 = 1. - w2;
                            cell2= w1.*hist2.orgCellTracker(iCell,floor(indexList(iTime,2)) )+ ...
                                    w2.*hist2.orgCellTracker(iCell,ceil(indexList(iTime,2)));
                        end
                    end
                    
                    sumHist.orgCellTracker(iCell, iTime) = cell1 + cell2;
                end
            end
            
        end
        
        % overload the multiplication operator
        function timesHist = times(hist1, hist2)
            
            timesHist = OrganismHistory();
            
            if (isa(hist1, 'OrganismHistory'))
                if (~isscalar(hist2) || ~isnumeric(hist2))
                    throw( MException('MATLAB:OrganismHistory:invalidTimes', ...
                        'Cannot multiply an organism history by object of type %s', ...
                        class(hist2)) );
                end
                
                timesHist.cellNames   = hist1.cellNames;
                timesHist.cellTypes   = hist1.cellTypes;
                timesHist.timeTracker = hist1.timeTracker;
            
                timesHist.startActiveTime = hist1.startActiveTime;
                timesHist.endActiveTime   = hist1.endActiveTime;
                
                timesHist.orgCellTracker = hist1.orgCellTracker;
                
                NCells = size(hist1.orgCellTracker,1);
                NTime  = size(hist1.orgCellTracker,2);
                
                for iTime = 1:NTime
                    for iCell = 1:NCells
                        timesHist.orgCellTracker(iCell, iTime) = ...
                            hist2.*timesHist.orgCellTracker(iCell, iTime);
                    end
                end
                
            elseif (isa(hist2, 'OrganismHistory'))
                if (~isscalar(hist1) || ~isnumeric(hist1))
                    throw( MException('MATLAB:OrganismHistory:invalidTimes', ...
                        'Cannot multiply an organism history by object of type %s', ...
                        class(hist1)) );
                end
                
                timesHist.cellNames   = hist2.cellNames;
                timesHist.cellTypes   = hist2.cellTypes;
                timesHist.timeTracker = hist2.timeTracker;
                
                timesHist.startActiveTime = hist2.startActiveTime;
                timesHist.endActiveTime   = hist2.endActiveTime;
            
                timesHist.orgCellTracker = hist2.orgCellTracker;
                
                NCells = size(hist2.orgCellTracker,1);
                NTime  = size(hist2.orgCellTracker,2);
             
                for iTime = 1:NTime
                    for iCell = 1:NCells
                        timesHist.orgCellTracker(iCell, iTime) = ...
                            hist1.*timesHist.orgCellTracker(iCell, iTime);
                    end
                end
                
            else
                throw( MException('MATLAB:OrganismHistory:invalidTimes', ...
                    'Cannot multiply class %s by class %s', ...
                    class(hist1), class(hist2)) );
            end
            
        end
        
        
    end
    
end
