%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% PPIUpdater < SystemUpdater
%   
%  SystemUpdater class that updates the organism's state from activity in
%  the PPI network using PPBN algorithm.
classdef PPIUpdater < SystemUpdater
    
    %% Internal variables to the PPI updater.
    properties (GetAccess = public, SetAccess = private)
        scaffoldState % Data structure for proteins bound to scaffolds
        

    end
    
    %% Update steps for the PPI network.
    methods (Access = public)
        
        %% Constructor
        function ppiUpdater = PPIUpdater()
        end
        
        %% initializer (inherited from SystemUpdater)
        %  Initialize the current state of variables interal to the PPI
        %  updater and any state variables in organism affected by the PPI
        %  network or values in the specified input file.
        function [firstState, firstTime] = initializer(ppiUpdater, ...
                organism, initFile, initTime)
            if (nargin == 2)
                ppiUpdater.scaffoldState = initializePPIUpdaterFromFile(...
                    organism);
            else
                ppiUpdater.scaffoldState = initializePPIUpdaterFromFile(...
                    organism, initFile);
            end
            firstState = organism;
            if (nargin <= 3)
                firstTime = 0.;
            else
                firstTime = initTime;
            end
            

        end
   
        %% update (inherited from SystemUpdater)
        %  Apply the rules from the PPI network to update the state of
        %  cells in the organism.  Note: the initial state of the organism
        %  is stored in the values at timeStep and delT is the length of
        %  the update step from timeStep - 1 to timeStep.
        function [delState, newDelT] = update(ppiUpdater, org, ...
                previousTime, delT)
           
            throw(MException('PPIUpdater:invalidUpdate', 'Must use updateTransport'));
        
        end
            
        function [delState, newDelT] = updateTransport(ppiUpdater, org, ...
                previousTime, delT, transporters)
            
            
            
            % set the new time step
            newDelT = delT;
                       
            % initialize the array of output states
            newScaffoldState = zeros(size(ppiUpdater.scaffoldState));
            newState = repmat(CellState, org.NCells, 1);
            delState = repmat(CellState, org.NCells, 1);
            
            % initialize placeholders for signal distribution
            MaxReceptors  = 0.;
            MaxProteins   = 0.;
            MaxNReceptors = 0.;
            MaxNProteins  = 0.;
            for iCell = 1:org.NCells
                MaxReceptors  = max(MaxReceptors, ...
                    org.OrgCells(iCell,1).Type.NReceptors);
                MaxProteins   = max(MaxProteins, ...
                    org.OrgCells(iCell,1).Type.NProteins);
                MaxNReceptors = max(MaxNReceptors, ...
                    max(org.OrgCells(iCell,1).State.Receptors));
                MaxNProteins  = max(MaxNProteins, ...
                    max(org.OrgCells(iCell,1).State.Proteins));
            end
            
            % used to keep track of the total number of activating and
            % repressing signals
            ReceptorSignals     = zeros(MaxReceptors, org.NCells);
            ActivatorSignals    = zeros(MaxProteins,  org.NCells);
            RepressorSignals    = zeros(MaxProteins,  org.NCells);
            ScaffoldActivator   = zeros(size(ppiUpdater.scaffoldState));
            ScaffoldRepressor   = zeros(size(ppiUpdater.scaffoldState));
             
          
            % update the scaffold state
            for iCell = 1:org.NCells
                % no update to the state of inactive proteins
                if (~org.OrgCells(iCell,1).Active)
                    continue;
                end
                
                % distribute available proteins among scaffolds
                newScaffoldState(:, :, :, iCell) = ppiUpdater.randomScaffolds(...
                	ppiUpdater.scaffoldState(:, :, :, iCell), ...
                    org.OrgCells(iCell,1).Type, ... 
                    org.OrgCells(iCell,1).State, delT); 
            end

            
            % determine the number of bound and inactive bound proteins of
            % each type in each cell
            NumBound = reshape( sum(sum( ...
                (newScaffoldState>=0), 1), 2), ...
                size(ppiUpdater.scaffoldState,3), ...
                size(ppiUpdater.scaffoldState,4));
            
            NumActiveBound = reshape( sum(sum( ...
                (newScaffoldState>0), 1), 2), ...
                size(ppiUpdater.scaffoldState,3), ...
                size(ppiUpdater.scaffoldState,4));

            
            % distribute signals and find out the result of each reaction 
            % were it to occur
            for iCell = 1:org.NCells
                
                % set the new cell state to match the old value
                newState(iCell,1) = CellState(org.OrgCells(iCell,1).State);
               
                
                % no update to the state of inactive proteins
                if (~org.OrgCells(iCell,1).Active)
                    continue;
                end
                
                % distribute external signals from iCell and try to react
                [delReceptorSignals, delRepressorSignals, ...
                    delScaffoldActivate, delScaffoldRepress] = ...
                    ppiUpdater.externalSignalReactions(org, ...
                    MaxReceptors, MaxProteins, ...
                    MaxNReceptors, MaxNProteins, newScaffoldState, ...
                    iCell, delT);
                
                ReceptorSignals = ReceptorSignals + delReceptorSignals;
                
                RepressorSignals = RepressorSignals + delRepressorSignals;
                
                ScaffoldActivator = ScaffoldActivator + ...
                    delScaffoldActivate;
                ScaffoldRepressor = ScaffoldRepressor + ...
                    delScaffoldRepress;

                % distribute signals from receptors to targets and react
                [delActivatorSignals, delRepressorSignals, ...
                    delScaffoldActivator, delScaffoldRepressor] = ...
                    ppiUpdater.receptorReactions(org, iCell,...
                    newScaffoldState(:,:,:,iCell), ...
                    MaxProteins, MaxNProteins, delT);
               
                
                ActivatorSignals(:,iCell) = ActivatorSignals(:,iCell) + ...
                    delActivatorSignals;
                RepressorSignals(:,iCell) = RepressorSignals(:,iCell) + ...
                    delRepressorSignals;
                ScaffoldActivator(:,:,:,iCell) = ScaffoldActivator(:,:,:,iCell) + ...
                    delScaffoldActivator;
                ScaffoldRepressor(:,:,:,iCell) = ScaffoldRepressor(:,:,:,iCell) + ...
                    delScaffoldRepressor;               
                

                
                % distribute proteins among targets and react 
                [delActivatorSignals, delRepressorSignals, ...
                    delScaffoldActSig, delScaffoldRepSig] = ...
                    ppiUpdater.proteinReactions(org, ...
                    newScaffoldState(:,:,:,iCell), MaxProteins, MaxNProteins, ...
                    iCell, delT);
                ActivatorSignals(:,iCell) = ActivatorSignals(:,iCell) + delActivatorSignals;
                RepressorSignals(:,iCell) = RepressorSignals(:,iCell) + delRepressorSignals;
                
                ScaffoldActivator(:,:,:,iCell)   = ScaffoldActivator(:,:,:,iCell)   + delScaffoldActSig;

                ScaffoldRepressor(:,:,:,iCell)   = ScaffoldRepressor(:,:,:,iCell)   + delScaffoldRepSig;
                               
                     
            end
            

            
            % evolve the state
            for iCell = 1:org.NCells
                
                % no update to the state of inactive proteins
                if (~org.OrgCells(iCell,1).Active)
                    delState(iCell,1) = newState(iCell,1) - ...
                        org.OrgCells(iCell,1).State;
                    continue;
                end
                
                % evolve the receptors based on the reactions and self
                % activation and deactivation
                
                % revise this to account for reactions already done
                % and to incorporate self activation / deactivation
                newState(iCell,1).ActiveReceptors = ...
                    ppiUpdater.evolveReceptors(org.OrgCells(iCell,1), ...
                    ReceptorSignals(:,iCell), delT);
                
                
                % update the unbound protein states
                
                % revise this to account for reactions already done
                % and to incorporate self activation / deactivation
                [activeUnbound, numTransported] = ppiUpdater.evolveUnboundProteins(...
                    org.OrgCells(iCell,1), ...
                    NumBound(:,iCell), NumActiveBound(:,iCell), ...
                    ActivatorSignals(:,iCell), RepressorSignals(:,iCell), ...
                    previousTime, ...
                    org.OrgCells(iCell,1).Type.canTransport, ...
                    org.OrgCells(iCell,1).Type.ProteinTypes, ...
                    transporters(iCell), delT);
                
                % update the bound protein states

                % revise this to account for reactions already done
                % and to incorporate self activation / deactivation
                newScaffoldState(:, :, :, iCell) = ...
                    ppiUpdater.evolveScaffoldProteins(...
                    org.OrgCells(iCell,1).Type, ...
                    newScaffoldState(:, :, :, iCell), ...
                    ScaffoldActivator(:,:,:,iCell), ...
                    ScaffoldRepressor(:,:,:,iCell), delT);
                
                
                % update the total number of active proteins
                newState(iCell,1).ActiveProteins = ...
                    activeUnbound + ...
                    reshape(sum(sum(newScaffoldState(:,:,:,iCell)>0,1),2), ...
                    size(activeUnbound));
                
                % figure out which of the active proteins are transported 
                % FOR NOW DOES NOT CORRECTLY HANDLE SCAFFOLDS -- MUST BE
                % FIXED LATER
                for iProt = 1:org.OrgCells(iCell,1).Type.NProteins
                    if (~org.OrgCells(iCell,1).Type.canTransport(iProt))
                        continue;
                    end

                    newState(iCell,1).Proteins(iProt) = ...
                        newState(iCell,1).Proteins(iProt) - numTransported(iProt);
                    
                end
                
                % compute the change in the state
                delState(iCell,1) = newState(iCell,1) - ...
                    org.OrgCells(iCell,1).State;
                
                ppiUpdater.scaffoldState(:, :, :, iCell) = ...
                    newScaffoldState(:, :, :, iCell);
            end
            
            
            
        end
        
        
        
        
    end
    
	methods (Access = private)
        
        function currentScaffoldState = randomScaffolds(ppiUpdater, ...
                scaffoldState, type, state, delT)
            
                        
            if (~isa(type, 'CellProperties'))
                throw( MException('MATLAB:RandomScaffolds:invalidClass', ...
                    '   %s is an invalid class to specify cell types for update', ...
                    class(type)));
            end
            
            if (~isa(state, 'CellState'))
                throw( MException('MATLAB:RandomScaffolds:invalidClass', ...
                    '%s is an invalid class to specify cell states for update', ...
                    class(state)));
            end
            
            % initialize the current scaffold state
            currentScaffoldState = -ones(size(scaffoldState));
            
            % do not update if there are not any scaffolds
            if (state.Scaffolds == 0)
                return;
            end
            
            % update each protein individually
            for iProt = 1:type.NProteins
                
                % find the scaffolds to which the protein can bind
                scaffoldTypes = find(type.Scaffolds(iProt,:,1)>0);
                
                % do not update proteins that do not bind to any scaffold
                if (all(type.Scaffolds(iProt,:,1)<=0))
                    continue;
                end
                
                % count the number of proteins that have been bound to the scaffold
                nActiveBound = ...          % number of active proteins bound
                    sum(sum(scaffoldState(:,:,iProt)==1));
                nBound = nActiveBound + ... % total number of proteins bound
                    sum(sum(scaffoldState(:,:,iProt)==0));
                
                % try to bound any unbound proteins to the scaffold
                if (nBound < state.Proteins(iProt))
                    
                    unboundScaffoldList = [];
                    
                    % determine the list of available scaffolds
                    for iScaff = scaffoldTypes
                        if (any(scaffoldState(1:state.Scaffolds(iScaff), ...
                                iScaff, iProt) == -1))
                            unboundScaff = find(...
                                scaffoldState(1:state.Scaffolds(iScaff),...
                                iScaff, iProt) == -1);
                            scaffListItem = cat(2,...
                                unboundScaff, iScaff.*ones(size(unboundScaff)));
                            unboundScaffoldList = cat(1,...
                                unboundScaffoldList, scaffListItem);
                        end
                    end
                    
                    % bind proteins to availaListble slots on scaffolds (if there are any)
                    if (numel(unboundScaffoldList) ~= 0)
                        % number of proteins available to bind
                        nProtAvail = state.Proteins(iProt) - nBound;
                        nActiveAvail = max(0, ...
                            state.ActiveProteins(iProt) - nActiveBound);
                        
                        % list of protein types available to bind
                        proteinList = zeros(nProtAvail,1);
                        proteinList(1:nActiveAvail) = ones(nActiveAvail,1);
                        % if there are more scaffolds than available proteins, force
                        % proteins states to -1's to ensure that we don't get
                        % overzealous binding
                        if (nProtAvail < size(unboundScaffoldList,1))
                            proteinList(nProtAvail+1:size(unboundScaffoldList,1)) = ...
                                -ones(size(nProtAvail+1:size(unboundScaffoldList,1)));
                        end
                        
                        % select scaffolds to which to bind available proteins
                        scaffoldPermute = randperm(size(unboundScaffoldList,1));
                        permuteUnboundScaffold = unboundScaffoldList(scaffoldPermute,1);
                        permuteScaffoldType = unboundScaffoldList(scaffoldPermute,2);
                        permuteProteins = ...
                            proteinList(randperm(length(proteinList)));
                        
                        pBind = expcdf(delT, type.Scaffolds(iProt,iScaff,1));
                        if (isnan(pBind))
                            pBind = 0.;
                        end

                        
                        for iUnbound = 1:size(unboundScaffoldList,1)
                            % accept each proposed bound protein with pBind
                            if ((permuteProteins(iUnbound) ~= -1) && ...
                                    (rand(1) < pBind))
                                iScaff = permuteScaffoldType(iUnbound);
                                nScaff = permuteUnboundScaffold(iUnbound);
                                currentScaffoldState(nScaff, iScaff, iProt) = ...
                                    permuteProteins(iUnbound);
                            end
                        end
                    end % if there are available slots on the scaffold
                    
                end % if there are available proteins
                
                % randomly unbind proteins with a fixed probability
                for iScaff = scaffoldTypes
                    
                    pUnbind = expcdf(delT, type.Scaffolds(iProt,iScaff,2));
                    if (isnan(pUnbind))
                        pUnbind = 0.;
                    end
                    
                    for jScaff = 1:state.Scaffolds(iScaff)
                        if (scaffoldState(jScaff, iScaff, iProt) ~= -1)
                            if (rand(1) < pUnbind)
                                currentScaffoldState(jScaff, iScaff, iProt) = -1;
                            else
                                currentScaffoldState(jScaff, iScaff, iProt) = ...
                                    scaffoldState(jScaff, iScaff, iProt);
                            end
                        end
                    end
                end
                
            end % loop over each protein
            
        end
        
        function [delReceptorSignals, delRepressorSignals, ...
                delScaffoldActivate, delScaffoldRepress] = ...
                externalSignalReactions(ppiUpdater, org, ...
                MaxReceptors, MaxProteins, ...
                MaxNReceptors, MaxNProteins, newScaffoldState, ...
                iCell, delT)
            
            % get the paramaters
            NExternalSignals = org.OrgCells(iCell,1).Type.NExternalSignals;
            NRepressors = org.OrgCells(iCell,1).Type.NRepressors;
            
            % initialize the output signals
            delReceptorSignals = zeros(MaxReceptors, ...
                org.NCells);
            delRepressorSignals = zeros(MaxProteins, ...
                org.NCells);
            delScaffoldActivate = zeros(size(newScaffoldState));
            delScaffoldRepress  = zeros(size(newScaffoldState));
            
            % do not distribute signal for inactive cell
            if (~org.OrgCells(iCell,1).Active)
                return;
            end
            
            % distribute the external signals
            for iSig = 1:NExternalSignals
                
                ExtSigConc = floor(org.OrgCells(iCell,1).State.ExternalSignals(iSig,1));

                
                
                if (ExtSigConc <= 0)
                    continue;
                end
                
                % handle neighbor signals
                if (org.OrgCells(iCell,1).Type.isExternalNeighbor(iSig,1))
                    
                    % find the amount of signals per neighbor cells
                    if (~any(org.OrgCells(iCell,1).Type.isNeighborCell))
                        continue;
                    end
                    
                    %probPerCell = double(org.OrgCells(iCell,1).Type.isNeighborCell) ./ ...
                    %    sum(org.OrgCells(iCell,1).Type.isNeighborCell);
                    %signalPerCell = mnrnd(floor(ExtSigConc), probPerCell);
                    
                    signalPerCell = PPIUpdater.distributeSignals(...
                        floor(ExtSigConc), ...
                        org.OrgCells(iCell,1).Type.isNeighborCell);
                    
                    % handle distribution among each cell
                    for jCell = 1:org.NCells
                        
                        if (~org.OrgCells(iCell,1).Type.isNeighborCell(jCell,1) || ...
                                signalPerCell(jCell) <= 0)
                            continue;
                        end
                        
                        % find the corresponding signal in the neighbor
                        % cell
                        try
                            sigIdx = findType(org.OrgCells(iCell,1).Type.ExternalSignalTypes(iSig,:), ...
                                org.OrgCells(jCell,1).Type.ExternalSignalTypes);
                        catch ME
                            % since nothing was found, don't do an update
                            continue;
                        end
                        
                        % since nothing was found, don't do an update
                        if (isempty(sigIdx))
                            continue;
                        end
                        
                        % distribute neighbor signals across the cell
                        probPerRecept = ...
                            (org.OrgCells(jCell,1).Type.ExternalSignals(:,sigIdx(1))>0).*...
                            org.OrgCells(jCell,1).State.Receptors;
                        if (sum(probPerRecept)==0)
                            continue;
                        end
                        %probPerRecept = probPerRecept./ sum(probPerRecept);
                        %receptorSignals = mnrnd(signalPerCell(jCell), ...
                        %    probPerRecept);
                        
                        receptorSignals = PPIUpdater.distributeSignals(...
                            signalPerCell(jCell), probPerRecept);
                        
                        probLigandBind = expcdf(delT, ...
                            org.OrgCells(jCell,1).Type.ExternalSignals(:, sigIdx(1)));
                        probLigandBind(isnan(probLigandBind)) = 0;
                        
                        for jRecept = 1:length(receptorSignals)
                        
                            delReceptorSignals(jRecept,jCell) = ...
                                delReceptorSignals(jRecept,jCell) + ...
                                binornd(receptorSignals(jRecept), ...
                                probLigandBind(jRecept));
                        end

                        
                    end
                    
                    % handle direct signals
                else
                    
                    
                    % distribute signals within this cell
                    probPerRecept = ...
                        (org.OrgCells(iCell,1).Type.ExternalSignals(:,iSig)>0).*...
                        org.OrgCells(iCell,1).State.Receptors;
                    
                    
                    if (sum(probPerRecept)<=0)
                        continue;
                    end
                    %probPerRecept = probPerRecept./ sum(probPerRecept);                  
                    %receptorSignals = mnrnd(floor(ExtSigConc), ...
                    %    probPerRecept);
                    receptorSignals = PPIUpdater.distributeSignals(...
                        floor(ExtSigConc), probPerRecept);
                    
                    probLigandBind = expcdf(delT, ...
                        org.OrgCells(iCell,1).Type.ExternalSignals(:, iSig));
                    
                    probLigandBind(isnan(probLigandBind)) = 0;
                    
                    for jRecept = 1:length(receptorSignals)
                    
                        delReceptorSignals(jRecept,iCell) = ...
                            delReceptorSignals(jRecept,iCell) + ...
                            binornd(receptorSignals(jRecept), ...
                            probLigandBind(jRecept));
                    end
                   
                    
                end
                
            end
            
            
            % distribute the repressors
            for iRep = 1:NRepressors
                
                % figure out which repressor goes to each protein
                RepressorConc = floor(org.OrgCells(iCell,1).State.Repressors(iRep,1));
                
                if (RepressorConc <= 0 || ...
                        ~any(org.OrgCells(iCell,1).Type.Repressors(:,iRep,1)>0) || ...
                        ~any(org.OrgCells(iCell,1).Type.Repressors(:,iRep,2)>0))
                    continue;
                end
                
                probPerProt = ...
                    (org.OrgCells(iCell,1).Type.Repressors(:,iRep,1)>0 | ...
                     org.OrgCells(iCell,1).Type.Repressors(:,iRep,2)>0).*...
                    org.OrgCells(iCell,1).State.Proteins;
                if (sum(probPerProt) == 0)
                    continue;
                end
                %probPerProt = probPerProt ./ sum(probPerProt);
                %Repressors = mnrnd(RepressorConc, probPerProt)';
                
                Repressors = PPIUpdater.distributeSignals(...
                    RepressorConc, probPerProt)';
                
                filledScaffolds = sum(sum(...
                    newScaffoldState(:,:,:,iCell)>=0,3),1);
                
                for iProt = 1:org.OrgCells(iCell,1).Type.NProteins
                    
                    if (Repressors(iProt,1) <= 0)
                        continue;
                    end
                    
                    if (floor(org.OrgCells(iCell,1).State.Proteins(iProt,1)) <= 0)
                        continue;
                    end
                    
                    numBound = sum(sum(newScaffoldState(:,:,iProt,iCell)>=0));
                    
                    numUnbound = max(0,floor(org.OrgCells(iCell,1).State.Proteins(iProt,1)) - ...
                        numBound);
                    
                    if (any(org.OrgCells(iCell,1).Type.Scaffolds(iProt,:,1)>0))
                        numScaffolds = ...
                            (org.OrgCells(iCell,1).Type.Scaffolds(iProt,:,1)>0).* ...
                            filledScaffolds;
                            %org.OrgCells(iCell,1).State.Scaffolds';
                        
                        % find which repressors go to scaffolds and which to

                        
                        numToScaffolds = PPIUpdater.distributeSignals(...
                            Repressors(iProt,1), ...
                            [sum(numScaffolds) numUnbound]);
                        %binornd(Repressors(iProt,1), ...
                        %    probScaffold);
                       
                        
                        numToScaffolds = numToScaffolds(1);
                        
                         
                        
                        if (numToScaffolds > 0)
                            Repressors(iProt,1) = Repressors(iProt,1) - ...
                                numToScaffolds;
                            
                            % find which scaffold each repressor goes to
                            %toEachScaffoldRep = mnrnd(numToScaffolds, ...
                            %    numScaffolds./sum(numScaffolds));
                            
                            toEachScaffoldRep = PPIUpdater.distributeSignals(...
                                numToScaffolds, numScaffolds);
                            
                            % distribute among copies of the scaffold
                            for iScaff = 1:org.OrgCells(iCell,1).Type.NScaffolds
                                if (toEachScaffoldRep(iScaff) <= 0)
                                    continue;
                                end
                                
                                
                                probRepPhosphBindScaff = expcdf(delT, ...
                                    org.OrgCells(iCell,1).Type.Repressors(iProt,iRep,2));
                                if (isnan(probRepPhosphBindScaff))
                                    probRepPhosphBindScaff = 0.;
                                end
                                
                                toScaffCopies = PPIUpdater.distributeSignals(...
                                    toEachScaffoldRep(iScaff), ...
                                    newScaffoldState(:,iScaff,iProt,iCell)>=0);
                                    %ones(1,org.OrgCells(iCell,1).State.Scaffolds(iScaff,1)));
                                
                                toScaffCopies = binornd(toScaffCopies, ...
                                    probRepPhosphBindScaff);
                                
                                delScaffoldRepress(:,iScaff,iProt,iCell) = ...
                                    delScaffoldRepress(:,iScaff,iProt,iCell) + ...
                                    toScaffCopies';
                                
%                                 for jSig = 1:toEachScaffoldRep(iScaff)
%                                     jScaffCopy = randi(...
%                                         org.OrgCells(iCell,1).State.Scaffolds(iScaff,1));
%                                     
%                                     if (rand(1) < probRepPhosphBindScaff)
%                                         delScaffoldRepress(jScaffCopy, iScaff, iProt, iCell) = ...
%                                             delScaffoldRepress(jScaffCopy, iScaff, iProt, iCell) + 1;
%                                     end
%                                     
%                                 end
                            end
                            
                        end
                        
                    end
                    
                    if (Repressors(iProt,1) <= 0)
                        continue;
                    end

                    
                    if (numUnbound <= 0)
                        continue;
                    end
                    
                    probRepPhosphBind = expcdf(delT, ...
                        org.OrgCells(iCell,1).Type.Repressors(iProt,iRep,1));
                    if (isnan(probRepPhosphBind))
                        probRepPhosphBind = 0.;
                    end
                    
                    
                    % react the distributed repressors
                    delRepressorSignals(iProt, iCell) = ...
                        delRepressorSignals(iProt, iCell) + ...
                        binornd(Repressors(iProt,1), probRepPhosphBind);
                    
                    
                end
                
            end
            
        end
        
        
        function [delActivatorSignals, delRepressorSignals, ...
                delScaffoldActivator, delScaffoldRepressor] = ...
                receptorReactions(ppiUpdater, org, iCell,...
                newScaffoldState, MaxProteins, MaxNProteins, delT)
            
            % initialize output variables
            delActivatorSignals = zeros(MaxProteins,1);
            delRepressorSignals = zeros(MaxProteins,1);
            delScaffoldActivator = zeros(size(newScaffoldState));
            delScaffoldRepressor = zeros(size(newScaffoldState));
            
            % only update the state for active cells
            if (~org.OrgCells(iCell,1).Active)
                return;
            end
            
            % get relevant parameters
            NReceptors = org.OrgCells(iCell).Type.NReceptors;
            NProteins = org.OrgCells(iCell).Type.NProteins;
            
            
            % perform the update for all the receptors
            for iRecept = 1:NReceptors
                if (org.OrgCells(iCell,1).State.ActiveReceptors(iRecept,1) <= 0)
                    continue;
                end
                
                % distribute receptor signal among targets
                probPerProt = (org.OrgCells(iCell,1).Type.Receptor2Target(:,iRecept,1) >0 | ...
                    org.OrgCells(iCell,1).Type.Receptor2Target(:,iRecept,2)>0).*...
                    org.OrgCells(iCell,1).State.Proteins;
               
                
                if (sum(probPerProt)==0)
                	continue;
                end
                %probPerProt = probPerProt ./ sum(probPerProt);
                %ReceptActSig = mnrnd(org.OrgCells(iCell,1).State.ActiveReceptors(iRecept,1), ...
                %    probPerProt)';
                
                ReceptActSig = PPIUpdater.distributeSignals(...
                    org.OrgCells(iCell,1).State.ActiveReceptors(iRecept,1), ...
                    probPerProt)';
                
                
                filledScaffolds = sum(sum(...
                    newScaffoldState(:,:,:)>=0,3),1);
                
                
                for iProt = 1:NProteins
                    
                    if (org.OrgCells(iCell,1).State.Proteins(iProt) <= 0)
                        continue;
                    end
                    
                    if (~org.OrgCells(iCell,1).Type.Receptor2Target(iProt,iRecept,1) > 0 && ...
                        ~org.OrgCells(iCell,1).Type.Receptor2Target(iProt,iRecept,2) > 0)
                        continue;
                    end
                    
                                        
                    numBound = sum(sum(newScaffoldState(:,:,iProt)>=0));
                    
                    numUnbound = max(0,floor(...
                        org.OrgCells(iCell,1).State.Proteins(iProt,1)) - ...
                        numBound);
                   
                    
                    numScaffolds = ...
                        (org.OrgCells(iCell,1).Type.Scaffolds(iProt,:,1)>0)...
                        .*filledScaffolds;
                        %.* org.OrgCells(iCell,1).State.Scaffolds';
                    
                    
                    if (sum(numScaffolds) > 0)
                        
                        %probScaffold = sum(numScaffolds) ./ ...
                        %    (sum(numScaffolds) + ...
                        %    org.OrgCells(iCell,1).State.Proteins(iProt,1));
                 
                        numToScaffolds = PPIUpdater.distributeSignals(...
                            ReceptActSig(iProt,1), ...
                            [sum(numScaffolds), numUnbound]);
                        
                        numToScaffolds = numToScaffolds(1);
                        
                        %binornd(ReceptActSig(iProt,1), probScaffold);
                        %toEachScaffoldSig = mnrnd(numToScaffolds, ...
                        %    numScaffolds./sum(numScaffolds));
                        
                        toEachScaffoldSig = PPIUpdater.distributeSignals(...
                            numToScaffolds, numScaffolds);
                        
                        ReceptActSig(iProt,1) = ReceptActSig(iProt,1) -  ...
                            numToScaffolds;

                        for iScaff = 1:org.OrgCells(iCell,1).Type.NScaffolds
                            if (toEachScaffoldSig(iScaff) <= 0)
                                continue;
                            end
                            
                            probPhosphBindScaff = expcdf(delT, ...
                                org.OrgCells(iCell,1).Type.Receptor2Target(iProt,iRecept,2));
                            if (isnan(probPhosphBindScaff))
                                probPhosphBindScaff = 0.;
                            end
                            
                            signalsToScaffCopy = PPIUpdater.distributeSignals(...
                                toEachScaffoldSig(iScaff), ...
                                newScaffoldState(:,iScaff,iProt)>=0);
                                %ones(1,org.OrgCells(iCell,1).State.Scaffolds(iScaff,1)));
                            
                            delScaffoldActivator(:, iScaff, iProt) = ...
                                delScaffoldActivator(:, iScaff, iProt) + ...
                                binornd(signalsToScaffCopy, probPhosphBindScaff)';
                            

                        end
                        
                    end
                    
                    % figure out which signal goes to which protein and
                    % react
                   
                    
                    if (ReceptActSig(iProt,1) <= 0)
                        continue;
                    end
                    


                    
                    if (numUnbound <= 0)
                        continue;
                    end
                    
                    probPhosphBind = expcdf(delT, ...
                        org.OrgCells(iCell,1).Type.Receptor2Target(iProt,iRecept,1));
                    if (isnan(probPhosphBind))
                        probPhosphBind = 0.;
                    end
                                        
                    % distribute repressors among individual copies of
                    % proteins and scaffolds
                    delActivatorSignals(iProt) = ...
                        delActivatorSignals(iProt) + ...
                        binornd(ReceptActSig(iProt,1), ...
                        probPhosphBind);
                    
                end
            end
            
        end
        
        function [delActivatorSignals, delRepressorSignals, ...
                delScaffoldActSig, delScaffoldRepSig] = ...
                proteinReactions(ppiUpdater, org, ...
                newScaffoldState, MaxProteins, MaxNProteins, ...
                iCell, delT)
            
            
            % TODO - continue fixing up code from here to deal with
            % distribution 
            
            % initialize outputs
            delActivatorSignals = zeros(MaxProteins,1);
            delRepressorSignals = zeros(MaxProteins,1);
            delScaffoldActSig   = zeros(size(newScaffoldState));
            delScaffoldRepSig   = zeros(size(newScaffoldState));
            
            % only update the state for active cells
            if (~org.OrgCells(iCell,1).Active)
                return;
            end
          
            % figure out scaffold information
            NProteins = org.OrgCells(iCell,1).Type.NProteins;           
            numActiveBound = reshape(sum(sum(newScaffoldState>0,1),2), NProteins, 1);
            
            numBound = reshape(sum(sum(newScaffoldState>=0,1),2), NProteins, 1);                
            numUnbound = max(0,floor(org.OrgCells(iCell,1).State.Proteins) - ...
                    numBound);
            numActiveUnbound = max(0, floor(org.OrgCells(iCell,1).State.ActiveProteins) - ...
                numActiveBound);
            
            
            filledScaffolds = sum(sum(...
                newScaffoldState(:,:,:)>=0,3),1);
                
 
            % distribute signals for each protein
            for iProt = 1:NProteins
                
                if (org.OrgCells(iCell,1).State.ActiveProteins(iProt) <= 0)
                    continue;
                end
                
                % find the proteins activated and repressed downstream
                protAct = logical(org.OrgCells(iCell,1).Type.PPI(iProt,:,1)>0 | ...
                    org.OrgCells(iCell,1).Type.PPI(iProt,:,2)>0)';
                protRep = logical(org.OrgCells(iCell,1).Type.PPI(iProt,:,1)<0 | ...
                    org.OrgCells(iCell,1).Type.PPI(iProt,:,2)<0)';
                
                % find candidate downstream prots bound to scaffolds
                % find the scaffolds to which the protein can bind
                scaffList = false(org.OrgCells(iCell,1).Type.NScaffolds,1);
                for iScaff = 1:org.OrgCells(iCell,1).Type.NScaffolds
                    if (any((org.OrgCells(iCell,1).Type.Scaffolds(:,iScaff,1) >0) & ...
                            protRep) || (any(org.OrgCells(iCell,1).Type.Scaffolds(:,iScaff,1) > 0 & ...
                            protAct)))
                        scaffList(iScaff,1) = true;
                    end
                end
                
                
                % figure out which go downstream in scaffolds
                nBoundToBound = 0;
                if (org.OrgCells(iCell,1).Type.NScaffolds > 0)
                    if (numActiveBound(iProt) > 0)
                        for iScaff = 1:org.OrgCells(iCell,1).Type.NScaffolds
                            if (~scaffList(iScaff,1))
                                continue;
                            end

                            for jScaff = 1:org.OrgCells(iCell,1).State.Scaffolds(iScaff)

                                % see if there is a phosphate for this scaffold
                                if (newScaffoldState(jScaff, iScaff, iProt) > 0)

                                    % find all available slots for this
                                    % phosphate within the scaffold
                                    slots = org.OrgCells(iCell,1).Type.Scaffolds(:,iScaff,1) > 0 & ...
                                        (protRep | protAct);

                                    % if there are any, figure out where it
                                    % goes and distribute
                                    if (sum(slots) > 0)
                                        
                                        
                                        nBoundToBound = nBoundToBound + 1;
                                        slotDist = PPIUpdater.distributeSignals(...
                                            newScaffoldState(jScaff, iScaff, iProt), ...
                                            slots);

                                        for jProt = 1:NProteins
                                            if (slotDist(jProt) <= 0)
                                                continue;
                                            end
                                            
                                            probBindScaff = expcdf(delT, ...
                                                abs(org.OrgCells(iCell,1).Type.PPI(iProt,jProt,2)));
                                            
                                            if (isnan(probBindScaff))
                                                probBindScaff = 0.;
                                            end
                                            
                                            if (protAct(jProt))
                                                delScaffoldActSig(jScaff,iScaff,jProt) = ...
                                                    delScaffoldActSig(jScaff,iScaff,jProt) + ...
                                                    binornd(slotDist(jProt), probBindScaff);
                                            end
                                            
                                            if (protRep(jProt))
                                                delScaffoldRepSig(jScaff,iScaff,jProt) = ...
                                                    delScaffoldRepSig(jScaff,iScaff,jProt) + ...
                                                    binornd(slotDist(jProt), probBindScaff);
                                            end
                                            
                                                    
        
                                        end

                                    end
                                end
                            end
                        end
                    end
                end
                
               
                % Add any phopshpates which were not received by scaffolds
                % to the general pool
                numActiveUnbound(iProt) = numActiveUnbound(iProt) + ...
                    numActiveBound(iProt) - nBoundToBound;
                              

                
                if (numActiveUnbound(iProt) <= 0)
                    continue;
                end
                
                % figure out downstream candidates
                if (any((org.OrgCells(iCell,1).Type.Scaffolds(iProt,:,1)>0)))
                    % if this protein could bind to a scaffold, it may not
                    % activate any proteins in a scaffold
                    
                    % TODO - check this for bugs!!!
                    scaffConc = scaffList.*...
                        filledScaffolds.*...
                        (~(org.OrgCells(iCell,1).Type.Scaffolds(iProt,:,1)>0));
                        %org.OrgCells(iCell,1).State.Scaffolds.*...
                        %(~(org.OrgCells(iCell,1).Type.Scaffolds(iProt,:,1)>0));
                else
                    scaffConc = scaffList.*...
                        filledScaffolds;
                %org.OrgCells(iCell,1).State.Scaffolds;
                end
                protActConc = protAct.*numUnbound;
                protRepConc = protRep.*numUnbound;
                
                
                NDownstream = (sum(protActConc)) + ...
                    (sum(protRepConc)) + ...
                    (sum(scaffConc));
                
                if (NDownstream <= 0)
                    continue;
                end
                
                
                pAct = sum(protActConc); 
                pRep = sum(protRepConc); 
                pScaff = sum(scaffConc); 
                NProtDist = PPIUpdater.distributeSignals(...
                    numActiveUnbound(iProt), [pAct, pRep, pScaff]);
                NProtAct = NProtDist(1);
                NProtRep = NProtDist(2);
                NScaff = NProtDist(3);
                
                
                if (sum(protActConc) > 0 && NProtAct > 0)
                    %probPerProt = protActConc ./ sum(protActConc);
                    %activatorSignals = mnrnd(NProtAct, probPerProt)';
                    
                    activatorSignals = PPIUpdater.distributeSignals(...
                        NProtAct, protActConc)';
                    
                else
                    activatorSignals = zeros(size(protActConc));
                end
                
                if (sum(protRepConc) > 0 && NProtRep > 0)
                    %probPerProt = protRepConc ./ sum(protRepConc);
                    %repressorSignals = mnrnd(NProtRep, probPerProt)';
                    
                    repressorSignals = PPIUpdater.distributeSignals(...
                        NProtRep, protRepConc)';
                else
                    repressorSignals = zeros(size(protRepConc));
                end
                
                
                for jProt = 1:NProteins
                    
                    if (numUnbound(jProt) <= 0)
                        continue;
                    end
                    
                    probBind = expcdf(delT, ...
                        abs(org.OrgCells(iCell,1).Type.PPI(iProt,jProt,1)));
                    
                    if (isnan(probBind))
                        probBind = 0.;
                    end
                            
                    
                    if (activatorSignals(jProt) > 0) 
                        delActivatorSignals(jProt,1) = ...
                            delActivatorSignals(jProt,1) + ...
                            binornd(activatorSignals(jProt,1), ...
                            probBind);
                    end
                    
                    if (repressorSignals(jProt) > 0)
                        delRepressorSignals(jProt,1) = ...
                            delRepressorSignals(jProt,1) + ...
                            binornd(repressorSignals(jProt,1), ...
                            probBind);
                    end
                    
                end
                
                
                % Distribute the remaining signals among scaffolds
                if (sum(scaffConc) > 0 && NScaff > 0)
                    %probPerScaff = scaffConc ./ sum(scaffConc);
                    %scaffIdx = mnrnd(NScaff, probPerScaff);
                    
                    scaffIdx = PPIUpdater.distributeSignals(NScaff, scaffConc);
                    
                    for iScaff = 1:org.OrgCells(iCell,1).Type.NScaffolds
                        if (scaffIdx(iScaff) > 0)
                            
                            signalsToScaff = PPIUpdater.distributeSignals(...
                                scaffIdx(iScaff), ...
                                sum(newScaffoldState(:,iScaff,:)>=0,3));
                                %ones(org.OrgCells(iCell,1).State.Scaffolds(iScaff),1));
                            
                            for jScaff = 1:org.OrgCells(iCell,1).State.Scaffolds(iScaff)
                            
                                scaffProtScaff = org.OrgCells(iCell,1).Type.Scaffolds(:,iScaff) > 0 & ...
                                    (protRep | protAct) & ... 
                                    reshape(sum(newScaffoldState(:,iScaff,:)>=0,1), ...
                                    org.OrgCells(iCell,1).Type.NProteins,1);
                                
                                if (signalsToScaff(1,jScaff) <= 0)
                                    continue;
                                end
                                
                                slotDist = PPIUpdater.distributeSignals(...
                                    signalsToScaff(1,jScaff), ...
                                    scaffProtScaff);
                                
                                %slotDist = mnrnd(1, scaffProtScaff./sum(scaffProtScaff));
                                
                                numAct = reshape((slotDist'.*protAct),1,1,org.OrgCells(iCell,1).Type.NProteins);
                                numRep = reshape((slotDist'.*protRep),1,1,org.OrgCells(iCell,1).Type.NProteins);
                                
                                for kProt = 1:org.OrgCells(iCell,1).Type.NProteins
                                    
                                    probScaffBind = expcdf(delT, ...
                                        abs(org.OrgCells(iCell,1).Type.PPI(iProt,kProt,2))); 
                                    if (isnan(probScaffBind))
                                        probScaffBind = 0.;
                                    end
                                    
                                    if (numAct(kProt) > 0)
                                        delScaffoldActSig(jScaff,iScaff,kProt) = ...
                                            delScaffoldActSig(jScaff,iScaff,kProt) + ...
                                            binornd(numAct(kProt), probScaffBind);
                                    end
                                    
                                    if (numRep(kProt) > 0)
                                        delScaffoldRepSig(jScaff,iScaff,kProt) = ...
                                            delScaffoldRepSig(jScaff,iScaff,kProt) + ...
                                            binornd(numRep(kProt), probScaffBind);
                                    end
                                   
                                end
                            end
                        end
                    end
                end
                
                
            end
            
            
        end
        
        function ActiveReceptors = evolveReceptors(ppiUpdater, orgCell, ...
                ReceptorSignals, delT)
            
            
            NReceptors = orgCell.Type.NReceptors;
            
            ActiveReceptors = zeros(NReceptors, 1);
            
            % activate each receptor
            for iReceptor = 1:NReceptors
                
                if (orgCell.State.Receptors(iReceptor,1) <= 0)
                    continue;
                end
                
                probReceptSpontInact = expcdf(delT, ...
                    orgCell.Type.SpontaneousInactivationReceptors(iReceptor));
                if (isnan(probReceptSpontInact))
                    probReceptSpontInact = 0.;
                end
                
                numActive = orgCell.State.ActiveReceptors(iReceptor) - ...
                   binornd(orgCell.State.ActiveReceptors(iReceptor), ...
                    probReceptSpontInact);

                probToInactive = 1 - numActive / ...
                    orgCell.State.Receptors(iReceptor);
                
                % determine the number of signals which reach inactive
                % proteins (relative to active)
                signalsToInactive = binornd(ReceptorSignals(iReceptor), ...
                    probToInactive);
                
                % spontaneous activation
                spontActReceptProb = expcdf(delT, ...
                    orgCell.Type.SpontaneousActivationReceptors(iReceptor));
                if (isnan(spontActReceptProb))
                    spontActReceptProb = 0.;
                end

                 signalsToInactive = signalsToInactive + binornd(...
                     orgCell.State.Receptors(iReceptor) - numActive, ...
                     spontActReceptProb);
                
                % update the state of inactive receptors
                ActiveReceptors(iReceptor) = numActive + ...
                    min(signalsToInactive, ...
                    orgCell.State.Receptors(iReceptor) - numActive);

                
            end
            
        end
        
        
        function newScaffoldState = evolveScaffoldProteins(ppiUpdater,...
                cellType, newScaffoldState, scaffoldActivator, ...
                scaffoldRepressor, delT)
            
          
            for k = 1:size(newScaffoldState,3)
                for j = 1:size(newScaffoldState,2)
                    
                    if (all(newScaffoldState(:,j,k) < 0))
                        continue;
                    end
                    
                    probScaffSpontInact = expcdf(delT, ...
                        cellType.SpontaneousInactivationBoundProteins(k,j));
                    if (isnan(probScaffSpontInact))
                        probScaffSpontInact = 0.;
                    end
                    spontActProtScaffProb = expcdf(delT, ...
                        cellType.SpontaneousActivationBoundProteins(k,j));
                    if (isnan(spontActProtScaffProb))
                        spontActProtScaffProb = 0.;
                    end
                        
                    
                    for i=1:size(newScaffoldState,1)
                        if (newScaffoldState(i,j,k) < 0)
                            continue;
                        end
                        
                        if (newScaffoldState(i,j,k) == 1)
                            newScaffoldState(i,j,k) = int8(rand(1) > ...
                                probScaffSpontInact);
                        end
                        
                        if (newScaffoldState(i,j,k) < 1 || ...
                                scaffoldRepressor(i,j,k) > 0)
                        
                            newState = scaffoldActivator(i,j,k) + ...
                                (rand(1) < spontActProtScaffProb) - ...
                                scaffoldRepressor(i,j,k);
                        
                            if (newState > 0) 
                                newScaffoldState(i,j,k) = 1;
                            else
                                newScaffoldState(i,j,k) = 0;
                            end
                        end
                        
                    end
                end
            end
            
        end
        
        
        function [activeUnbound, numTransported] = evolveUnboundProteins(...
                ppiUpdater, orgCell, NumBound, NumActiveBound, ...
                ActivatorSignals, RepressorSignals, previousTime, ...
                canTransport, protNames, transporters, delT)
           
            
            
            NProteins = orgCell.Type.NProteins;
            UnboundProteins = orgCell.State.Proteins - NumBound;
            ActiveProteins = orgCell.State.ActiveProteins - NumActiveBound;
            numTransported = zeros(size(ActiveProteins));
            activeUnbound = zeros(NProteins,1);
            
            for iProt = 1:NProteins
                
                names = {};
                actTimes = [];
                transTimes = [];
                if (canTransport(iProt))
                    [names, actTimes, transTimes] = transporters.getTransportList(...
                        protNames(iProt,:), previousTime, 1);
                end
  
                
                numTransport = length(names);
                
                if (UnboundProteins(iProt) + numTransport == 0)
                    continue;
                end
                
                probUnboundSpotInact = expcdf(delT, ...
                    orgCell.Type.SpontaneousInactivationUnboundProteins(iProt));
                if (isnan(probUnboundSpotInact))
                    probUnboundSpotInact = 0.;
                end
                spontActProtProb = expcdf(delT, ...
                    orgCell.Type.SpontaneousActivationUnboundProteins(iProt));
                if (isnan(spontActProtProb))
                    spontActProtProb = 0.;
                end
                
                % perform spontaneous deactivation on the transported
                % proteins
                keepTransportedProt = rand(numTransport,1) >  ...
                    probUnboundSpotInact;
                
                activeProt = binornd(ActiveProteins(iProt,1), ...
                    1-probUnboundSpotInact);
                inActiveProt = UnboundProteins(iProt,1) - activeProt;
                
                % figure out which category each thing goes to
                probToClass = [activeProt inActiveProt ones(1,numTransport)];% ./ ...
                    %(UnboundProteins(iProt) + numTransport);
                
                %actSig = mnrnd(ActivatorSignals(iProt,1), probToClass);
                %repSig = mnrnd(RepressorSignals(iProt,1), probToClass);
                
                actSig = PPIUpdater.distributeSignals(...
                    ActivatorSignals(iProt,1), probToClass);
                repSig = PPIUpdater.distributeSignals(...
                    RepressorSignals(iProt,1), probToClass);
                
                % find which meet inactive proteins
                nSigBlock = repSig(2);
                nSig = actSig(2) + ...
                    binornd(inActiveProt, spontActProtProb);
                
                nSigBlockAct = repSig(1);
                nSigAct = actSig(1) + binornd(activeProt, spontActProtProb);
                
                % figure out who's active in the end
                if (UnboundProteins(iProt) >= 0)
                    
                    activeRemaining = activeProt;
                    if (nSigBlockAct > 0)
                        nActiveBlocked = max(nSigBlockAct - ...
                            binornd(nSigAct, ...
                            min(nSigBlockAct, activeProt) / activeProt), 0);
                        activeRemaining = max(activeProt - nActiveBlocked, 0);
                    end
                        
                    
                    activeUnbound(iProt) = ...
                        activeRemaining + ...
                        max(min(inActiveProt, nSig - nSigBlock), 0);
                else
                    activeUnbound(iProt) = 0;
                end
                
                if (canTransport(iProt))
                    
                    % add all active proteins to the transported state
                    for jTransProt = 1:activeUnbound(iProt)
                        transporters.addTransportProtein(...
                            {protNames(iProt,:)}, previousTime);
                        numTransported(iProt) = numTransported(iProt) + 1;
                    end
                    activeUnbound(iProt) = activeUnbound(iProt) - ...
                        numTransported(iProt);
                    
                    % remove any transported proteins which do not stay
                    % active
                    for jTransProt = 1:numTransport
                        if (keepTransportedProt(jTransProt))
                            continue;
                        end
                        protIdx = 2+jTransProt;
                        nSigBlock = repSig(protIdx);
                        nSig = actSig(protIdx) + ...
                            double(rand(1) < spontActProtProb);
                        
                        keepTransportedProt(jTransProt) = nSig > nSigBlock;
                        if (~keepTransportedProt(jTransProt))
                            numRemoved = transporters.inactivateProt(...
                                previousTime, names(jTransProt), ...
                                actTimes(jTransProt), ...
                                transTimes(jTransProt));
                            numTransported(iProt) = numTransported(iProt) - ...
                                numRemoved;
                        end
                        
                    end
                    
                end


                
            end
            

        end

        
    end

    methods (Access = public, Static = true) 
        
        function signalCopies = distributeSignals(nSignals, numInSlots)
            % distribute signals preferentially among each copy and only
            % then allow for multiple signals to be distributed
            
            
            signalCopies = zeros(1,length(numInSlots));
            
            if (nSignals <= 0) 
                return;
            end
            
            nSigToEach = floor(nSignals / sum(numInSlots));
            
            if (nSigToEach > 0) 
                signalCopies = signalCopies + nSigToEach.*...
                    reshape(numInSlots,1,length(numInSlots));
            end
            
            nSignalsRemaining = nSignals - nSigToEach*sum(numInSlots);
            
            distributePermutation = randperm(sum(numInSlots));
            
            if (nSignalsRemaining > 0)
                
                            
                distributePermutation = ...
                    distributePermutation(1:nSignalsRemaining);
                
                lBound = 0;
                for iSig = 1:length(numInSlots)
                    rBound = lBound + numInSlots(iSig);
                    signalCopies(1,iSig) = signalCopies(1,iSig) + ...
                        sum(distributePermutation <= rBound & ...
                            distributePermutation > lBound);
                    lBound = rBound;
                end

            end
        end
        
    end
    
    
    
end
