%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.


classdef Simulator < SystemUpdater
    
    properties (GetAccess = public, SetAccess = private)
        currentTime
        currentState
        
        transporters
    end
    
    properties (GetAccess = protected, SetAccess = protected)
        ppiUpdate
        ttUpdate
        diffusionUpdate
 
        canActivate
        
        useUpdaters
    end
    
    properties (GetAccess = private, SetAccess = private, Constant = true)
        nUpdaters = 3;
        
        % index for each of the updaters
        ppiUpdaterIdx       = 1;
        ttUpdaterIdx        = 2;
        diffusionUpdaterIdx = 3;
    end
    
    methods (Access = public)
        
        function sim = Simulator(useUpdaters)

            sim.currentTime     = double.empty();
            sim.currentState    = Organism.empty();
            sim.ppiUpdate       = PPIUpdater();
            sim.ttUpdate        = TTUpdater(); 
            sim.diffusionUpdate = DiffusionUpdater();
            if (nargin==1)
                if (length(useUpdaters)~= Simulator.nUpdaters)
                    warning('MATLAB:Simulator:invalidDim', ...
                        'Case for %d of %d updaters specified.  %s.', ...
                        length(useUpdaters), ...
                        Simulator.nUpdaters, ...
                        'All updaters are set as true');
                else
                    sim.useUpdaters = useUpdaters;
                    if (sum(useUpdaters)==0)
                        warning('MATLAB:Simulator:noSimulators', ...
                            'All updaters are set to false.  %s.', ...
                            'The simulator will only update the time');
                    end
                    return;
                end
            end
            sim.useUpdaters = true(Simulator.nUpdaters,1);
        end
        
        function [firstState, firstTime] = initializer(sim, organism, ...
                initFiles, initTime)
            if (nargin < 3)
                initTime = 0.;
            end
            
            
            try
                sim.update(organism, initTime);
                if (sim.useUpdaters(Simulator.ppiUpdaterIdx))
                    sim.ppiUpdate.initializer(sim.currentState, ...
                        initFiles(Simulator.ppiUpdaterIdx), initTime);
                end
                if (sim.useUpdaters(Simulator.ttUpdaterIdx))
                    sim.ttUpdate.initializer(sim.currentState, ...
                        initFiles(Simulator.ttUpdaterIdx), initTime);
                end
                if (sim.useUpdaters(Simulator.diffusionUpdaterIdx))
                    sim.diffusionUpdate.initializer(sim.currentState, ...
                        initFiles(Simulator.diffusionUpdaterIdx), initTime);
                end
            catch ME
                outME = addCause( ...
                    MException( ...
                        'MATLAB:Simulator:invalidInitialization', ...
                        'Cannot initialize the simulator from the input files.'), ...
                    ME);
                throw(outME);
            end
            
            sim.canActivate = true(organism.NCells,1);
            
            firstState = sim.currentState;
            firstTime  = sim.currentTime;
            
            sim.transporters = [];
            
            for iCell = 1:organism.NCells
                sim.transporters = [sim.transporters, ...
                    TransportProtein(organism.OrgCells(iCell,1).Type)];
            end
            
        end
        
        function orgHistory = simulateAllTime(sim, simTimeSteps, finalTime)
            orgHistory = OrganismHistory(sim.currentState, sim.currentTime);

			% clean out the transporters
			for iCell = 1:sim.currentState.NCells
				sim.transporters(iCell).clearTransporter();
			end
            
            % perform basic sanity checks
            if (finalTime <= sim.currentTime)
                return;
            end
            if (~any(sim.useUpdaters))
                warning('MATLAB:Simulator:NoUpdate', ...
                    'No simulators specified.  Not performing a simulation.');
            end
            
            % check that the number of timesteps is the same as the number
            % of updaters
            if (length(simTimeSteps) == 1)
               timeStep = simTimeSteps.*ones(Simulator.nUpdaters,1);
            elseif (length(simTimeSteps) == Simulator.nUpdaters)
                timeStep = simTimeSteps;
            else
                warning('MATLAB:Simulator:NoUpdate', ...
                    'Only specified timesteps for %d of the %d updaters', ...
                    length(simTimeSteps), sim.nUpdaters);
                return;
            end
            
            ppiNextTime = sim.currentTime;
            ttNextTime = sim.currentTime;
            diffusionNextTime = sim.currentTime;
            
            % perform the update
            while ( sim.currentTime < finalTime )
                

                [divide, anyActive] = sim.divideCells();
                
                % do not continue a simulation w/o any active cells
                if ( ~divide && ~anyActive && ~any(sim.canActivate) )
                    break; 
                end 
                
                if (~anyActive)
%                    fprintf('No cells active; time: %f\n', sim.currentTime);
                    delT = min(timeStep(logical(timeStep>0)));
                    newState = sim.currentState.getCellStates();
                    sim.update(newState, sim.currentTime, delT);
                else
                
                    % perform cell division, if desired (returns true if cells
                    % divide, false otherwise)
                    if (divide)

                        % perform cell division
%                        fprintf('Dividing the cells at time %f\n', ...
%                            sim.currentTime);

                        % reset the next times so that all relevant simulators
                        % are run on the new cells
                        ppiNextTime       = sim.currentTime;
                        ttNextTime        = sim.currentTime;
                        diffusionNextTime = sim.currentTime;
                    else


                        % perform the updates
                        if (sim.useUpdaters(Simulator.ppiUpdaterIdx)  && ...
                                ppiNextTime <= sim.currentTime)
                            [delPPIState, ppiDelT] = sim.ppiUpdate.updateTransport(...
                                sim.currentState, sim.currentTime, ...
                                timeStep(Simulator.ppiUpdaterIdx), ...
                                sim.transporters);
                            ppiNextTime = sim.currentTime + ppiDelT;
                        end

                        if (sim.useUpdaters(Simulator.ttUpdaterIdx) && ...
                                ttNextTime <= sim.currentTime)
                             [delTTState, ttDelT] = sim.ttUpdate.updateTransport(...
                                 sim.currentState, sim.currentTime, ...
                                 timeStep(Simulator.ttUpdaterIdx), ...
                                 sim.transporters);

                            ttNextTime = sim.currentTime + ttDelT;
                        end

                        if (sim.useUpdaters(Simulator.diffusionUpdaterIdx) && ...
                                diffusionNextTime <= sim.currentTime)
                            [delDiffusionState, diffusionDelT] = sim.diffusionUpdate.update(...
                                sim.currentState, sim.currentTime, ...
                                timeStep(Simulator.diffusionUpdaterIdx));
                            diffusionNextTime = sim.currentTime + diffusionDelT;
                        end
                        

                        % find the next time at which to update the state
                        nextTime = max(max(ppiNextTime, ttNextTime), diffusionNextTime);
                        if (ppiNextTime > sim.currentTime && ppiNextTime < nextTime)
                            nextTime = ppiNextTime; 
                        end
                        if (ttNextTime > sim.currentTime && ttNextTime < nextTime)
                            nextTime = ttNextTime;
                        end
                        if (diffusionNextTime > sim.currentTime && diffusionNextTime < nextTime)
                            nextTime = diffusionNextTime;
                        end

                        if (nextTime == sim.currentTime) 
                            warning('MATLAB:Simulator:noUpdate', ...
                                'None of the updaters were active at time %f', ...
                                sim.currentTime);
                        end

%                        fprintf('time: %f\n', nextTime);

                        % find the next state
                        newState = sim.currentState.getCellStates();
                        for iCell = 1:sim.currentState.NCells
                            
                            if (~newState(iCell,1).Active)
                                continue;
                            end
                            
                            if (ppiNextTime == nextTime && ...
                                    sim.useUpdaters(Simulator.ppiUpdaterIdx))
                                newState(iCell,1) = newState(iCell,1) + ...
                                    delPPIState(iCell,1);
                                
                                if (iCell == 1)
%                                    fprintf('\tppi');
                                end
                                if (~newState(iCell,1).checkPositive())
                                    throw(MException('MATLAB:Simulator:invalidUpdate', ...
                                        'PPI causes negative cell population in cell %d', ...
                                        iCell));
                                end
                            end
                            if (ttNextTime == nextTime && ...
                                    sim.useUpdaters(Simulator.ttUpdaterIdx))
                                newState(iCell,1) = newState(iCell,1) + ...
                                    delTTState(iCell,1);
                                if (iCell == 1)
%                                    fprintf('\ttt');
                                end
                                if (~newState(iCell,1).checkPositive())
                                    throw(MException('MATLAB:Simulator:invalidUpdate', ...
                                        'TT causes negative cell population in cell %d', ...
                                        iCell));
                                end
                            end
                            if (diffusionNextTime == nextTime && sim.useUpdaters(Simulator.diffusionUpdaterIdx))
                                newState(iCell,1) = newState(iCell,1) + ...
                                    delDiffusionState(iCell,1);
                                if (iCell == 1)
%                                    fprintf('\tdiffusion');
                                end
                                if (~newState(iCell,1).checkPositive())
                                    throw(MException('MATLAB:Simulator:invalidUpdate', ...
                                        'Diffusion causes negative cell population in cell %d', ...
                                        iCell));
                                end
                            end
                        end            
                        
                        
                        % update the state in the simulator
                        %fprintf('calling updater from simulator\n');
                        sim.update(newState, sim.currentTime, ...
                            nextTime - sim.currentTime);
                        
                        clear newState;
                    end

                end
                    
                % update the organism history to reflect the next value
                orgHistory.updateHist(sim.currentState, sim.currentTime);
                    
            end
            
        end
        
        function [divide, anyActive] = divideCells(sim)
            
            
            divide = false;
            anyActive = false;
            
            newState = sim.currentState.getCellStates();
            
            divisionTime = 0.;
            
            for iCell = 1:sim.currentState.NCells
                
                if (sim.currentState.OrgCells(iCell,1).State.Active)
                    anyActive = true;
%                    fprintf('%s\t', sim.currentState.OrgCells(iCell,1).Name);
                end
                
                
                % activate any inactive cells which should be activated
                if (~sim.currentState.OrgCells(iCell,1).State.Active && ...
                        sim.currentTime >= sim.currentState.CellStartTime(iCell,1) && ...
                        sim.canActivate(iCell,1))
                    newState(iCell,1).Active = true;
                    sim.canActivate(iCell,1) = false;
                    
                    % though not strictly a division, needed to update the
                    % state
                    divide = true;
                end
                
                % divide any cells
                if (sim.currentState.OrgCells(iCell,1).State.Active && ...
                        sim.currentTime >= sim.currentState.DivisionTime(iCell,1))
                    divide = true;
                    
                    
                    newState(iCell,1).Active = false;
                    
                    % move over any proteins from the transporter
                    % TODO - debug
                    [numTransProt, protNames] = sim.transporters(iCell).clearTransporter(sim.currentState.OrgCells(iCell,1).Type.ProteinTypes);
                    newState(iCell,1).Proteins = newState(iCell,1).Proteins + numTransProt';
                    newState(iCell,1).ActiveProteins = newState(iCell,1).ActiveProteins + numTransProt';
                    
                    if (~any(sim.currentState.Progeny(:,iCell)))
                        continue;
                    end
                    
                    for jCell = 1:sim.currentState.NCells
                        if (sim.currentState.Progeny(jCell,iCell) && ...
                                sim.canActivate(jCell,1))

                            newState(jCell,1) = CellState(newState(iCell,1));
                            
                            divisionTime = max(divisionTime, ...
                                sim.currentState.CellStartTime(jCell,1) - ...
                                sim.currentState.DivisionTime(iCell,1));
                            newState(jCell,1).Active = true;
                            sim.canActivate(jCell,1) = false;
                        end
                    end
                end
                
                
                %sim.currentState.OrgCells(iCell,1).State
                
                
            end
            
            if (divide)
                sim.update(newState, sim.currentTime, divisionTime);
                
            end
            
            
        end
        
        
        function [newState, newTime] = scheduler(sim, timeStep)
            
            % update the time but not the state if no updaters are used
            if (~any(sim.useUpdaters))
                newState = sim.update(...
                    sim.currentState, sim.currentTime, timeStep);
                newTime = sim.currentTime;
                return;
            end
            
            % get the current state
            newState = sim.currentState.getCellStates();
            
            % if desired, perform the ppi update
            ppiDelT = 0.;
            if (sim.useUpdaters(Simulator.ppiUpdaterIdx))
                [ppiState, ppiDelT] = sim.ppiUpdate.update(...
                    sim.currentState, sim.currentTime, timeStep);
                if (ppiDelT ~= 0)
                    for iCell = 1:sim.currentState.NCells
                        newState(iCell,1) = newState(iCell,1) + ...
                            ppiState(iCell,1);
                        if (~newState(iCell,1).checkPositive())
                            warning('MATLAB:Simulator:invalidState', ...
                                'PPIUpdater causes cell %d to have a negative species population', ...
                                iCell);
                        end
                    end
                end
            end
            
            % if desired, perform the tt update
            ttDelT = 0.;
            if (sim.useUpdaters(Simulator.ttUpdaterIdx))
                [ttState, ttDelT] = sim.ttUpdate.update(...
                    sim.currentState, sim.currentTime, timeStep);
                if (ttDelT ~= 0)
                    for iCell = 1:sim.currentState.NCells
                        newState(iCell,1) = newState(iCell,1) + ...
                            ttState(iCell,1);
                        if (~newState(iCell,1).checkPositive())
                            warning('MATLAB:Simulator:invalidState', ...
                                'TTUpdater causes cell %d to have a negative species population', ...
                                iCell);
                        end
                    end
                end
            end
            
            % if desired, perform the diffusion update
            diffusionDelT = 0.;
            if (sim.useUpdaters(Simulator.diffusionUpdaterIdx))
                [diffusionState, diffusionDelT] = sim.diffusionUpdate.update(...
                    sim.currentState, sim.currentTime, timeStep);
                if (diffusionDelT ~= 0)
                    for iCell = 1:sim.currentState.NCells
                        newState(iCell,1) = newState(iCell,1) + ...
                            diffusionState(iCell,1);
                        if (~newState(iCell,1).checkPositive())
                            warning('MATLAB:Simulator:invalidState', ...
                                'DiffusionUpdater causes cell %d to have a negative species population', ...
                                iCell);
                        end                                    
                    end
                end
            end
            
            % update the state from all three update steps
            trueDelT = max(max(ppiDelT, ttDelT), diffusionDelT);
            [newState] = sim.update(newState, sim.currentTime, trueDelT);
            newTime = sim.currentTime;
        end      
        
        function [newState, newDelT] = update(sim, state, previousTime, ...
                delT)
            
            
            newState = state;
            
            switch class(newState)
                case 'Organism'
                    sim.currentState = newState;
                case 'CellState'
                    sim.currentState.setCellStates(newState);
                otherwise
                    throw( MException ( ...
                        'MATLAB:Simulator:invalidUpdate', ...
                        '%s cannot update the current state.', ...
                        class(newState) ) );
            end
            
            if (nargin == 3)
                newDelT = 0.;
                sim.currentTime = previousTime;
            elseif (nargin > 3)
                        
                newDelT  = delT;

                if (newDelT < 0.)
                    throw( MException ( ...
                        'MATLAB:Simulator:invalidTime', ...
                        'Attempting to update in negative time step: %f.', ...
                        newDelT) );
                end

                if (sim.currentTime ~= previousTime)
                    throw( MException ( ...
                        'MATLAB:Simulator:invalidTime', ...
                        'At time %f attempting to update invalid time %f.', ...
                        sim.currentTime, previousTime) );
                end

                sim.currentTime = previousTime + newDelT;
            else
                throw( MException('MATLAB:Simulator:invalidUpdate', ...
                    '%d arguments cannot update the state.', ...
                    nargin) );
            end
            
        end
            
    end
    
end
