%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% TTUpdater < SystemUpdater
%
%  SystemUpdater class that updates the organism's state from transcription
%  and translational processes.
classdef TTUpdater < SystemUpdater
    
    %% Internal variables to the TT updater.
    properties (GetAccess = private, SetAccess = private)
        % indexes for getting and updating number of proteins
        
        simulatorsThreshold = 100000; % number of proteins when we should 
                                    % switch stochastic simulator on ode
        mapping = []; % vector with a cell of the following type:
                      % {'species_name', species_type, property_index1,species_index1,
                      % property_index2, species_index2, isTF}, where
                      % species_type is type of species (protein, repressor, and so on) for network only
                      % property_index1 and property_index2 are 0 for network 
                      % or index in TT_Species;
                      % species_index1 and species_index2 are indices in
                      % corresponding vectors;
                      % isTF indicates if this species is TF and it should
                      % be transported (for network only)
                      % For example, {'LIP-1',0,'repressor,5,3,10,0} means that LIP-1
                      % repressor has index 5 in network and index 10 in
                      % third TT object and it's not transported
                                    
    end
    
    methods
        
        %% constructor
        function ttUpdater = TTUpdater()
        end
        
        %% initializer (inherited from SystemUpdater)
        function [state, time] = initializer(ttUpdater, organism, ...
                initFile, initTime)
            state = organism;
            try
                % read initFile
               switch lower(initFile.type)
                case 'txt'
                    % load file
                    data = importdata(initFile.name);
                    % find header lines
                    networkLine = 0;
                    ttLine = 0;
                    for i = 1:length(data)
                       switch lower(strtok(char(data{i})))
                           case lower('[networkToTT]')
                               networkLine = i;                               
                           case lower('[TTtoNetwork]')
                               ttLine = i;
                       end
                    end
                    % create mapping for lines from file
                    if (networkLine < ttLine)
                        % mapping from network to TT
                        for i = networkLine+1:ttLine-1
                            % split line by tab
                            line = regexp(data{i},'\t','split');
                            % find index in network
                            name_type = regexp(line{1},',','split');
                            names = organism.CellTypes.getStateNames(name_type(2));
                            netInd = findType(name_type(1), names);
                            state = organism.OrgCells(1).State.getState(name_type(2));
                            % find indices in TT_objects
                            ttInd = ttUpdater.findIndices(line(2),organism.CellTypes.TT_Species);
                            
                            if (~isempty(netInd) && ~isempty(ttInd))
                                s = size(ttInd);
                                for ind = 1:s(1)
                                    ttUpdater.mapping = [ttUpdater.mapping;...
                                        {strtrim(name_type(1)),...  % species name
                                        name_type(2), ...type of species 
                                        0, ... indicates that it's network
                                        netInd, ... index of this species in network
                                        ttInd(ind,1), ... index of TT object
                                        ttInd(ind,2), ... index of this species in TT object
                                        organism.CellTypes.canTransport(netInd)}]; % indicates if this species is a TF
                                end
                            end
                        end
                        % mapping from TT to network
                        for i = ttLine+1:length(data)
                           % split line by tab
                            line = regexp(data{i},'\t','split');
                            % find index in network
                            name_type = regexp(line{2},',','split');
                            names = organism.CellTypes.getStateNames(name_type(2));
                            netInd = findType(name_type(1), names);
                            % find indices in TT_objects
                            ttInd = ttUpdater.findIndices(line(1),organism.CellTypes.TT_Species);
                            
                            if (~isempty(netInd) && ~isempty(ttInd))
                                s = size(ttInd);
                                for ind = 1:s(1)
                                    ttUpdater.mapping = [ttUpdater.mapping;...
                                        {strtrim(strtrim(name_type(1))),...  % species name
                                        name_type(2), ...type of species 
                                        0, ... indicates that it's network
                                        netInd, ... index of this species in network
                                        ttInd(ind,1), ... index of TT object
                                        ttInd(ind,2), ... index of this species in TT object
                                        0}]; % indicates if this species is a TF
                                end
                            end
                            
                        end
                    else
                        % mapping from TT to network
                        for i = ttLine+1:networkLine-1
                            
                        end
                        % mapping from network to TT
                        for i = networkLine+1:length(data)
                            
                        end
                    end
                        
               end
               
                % fill in the mapping vector
                for i = 1: length(organism.CellTypes.TT_Species)
                    % get SSA_Properties object
                    tt_obj1 = organism.CellTypes.TT_Species(i);
                    % check if this object has SSA_Properties type
                    if (isa(tt_obj1, 'SSA_Properties'))
                        % iterate names and find matches
                        for j = 1:length(tt_obj1.Names)
                       % search in other TT objects
                            for k = i+1: length(organism.CellTypes.TT_Species)
                                % get another SSA_Properties object
                                tt_obj2 = organism.CellTypes.TT_Species(k);
                                % find index
                                ind = findType(tt_obj1.Names(j),...
                                        char(tt_obj2.Names));
                                    if (~isempty(ind))
                                       ttUpdater.mapping = [ttUpdater.mapping;...
                                            {tt_obj1.Names(j),...  % species name
                                            [],... empty reference
                                            i, ... index of the first TT object
                                            j, ... index of species
                                            k, ... index of the second TT object
                                            ind,...index of species
                                            0}]; % not transported
                                         
                                    end
                                
                             end
                        end
                    end
                end
            catch ME
                outME = addCause( ...
                    MException( ...
                        'MATLAB:TTUpdater:invalidInitialization', ...
                        'Cannot initialize the TTUpdater.'), ...
                    ME);
                throw(outME);
            end
         
            if (nargin <= 3)
                time  = 0.;
            else
                time = initTime;
            end
        end
        %% find indices in the list of SSA_Properties objects
        % and return matrix where each line is index of SSA_Properties object 
        % in the list and index species in the SSA_Properties object
        function ttInd = findIndices(ttUpdater, species, ttObjs)
            
            ttInd = [];
            try
                for i = 1:length(ttObjs)
                    ttObj = ttObjs(i);
                    if (isa(ttObj, 'SSA_Properties'))
                        ind = findType(species,char(ttObj.Names));
                        if (ind ~= -1)
                            ttInd = [ttInd; i, ind];
                        end
                    end
                end
            catch ME
               outME = addCause( ...
                    MException( ...
                        'MATLAB:TTUpdater:findIndices', ...
                        'Did not find %s in the specified list.', ...
                         species), ME);
                throw(outME);
            end
        end

        
        %% updateState (inherited from SystemUpdater)
        function [delState, newDelT] = update(ttUpdater, org, ...
                currentTime, delT)
            
            throw(MException('TTUpdater:invalidUpdate', 'Must use updateTransport'));
 
        end        
      
        function [delState, newDelT] = updateTransport(ttUpdater, org, ...
                currentTime, delT, transporters)
                        % initialize the array of output states and timestamps
            newState = repmat(CellState, org.NCells, 1);
            delState = repmat(CellState, org.NCells, 1);
            newDelT = [];
            
            % copy current state
            for iCell = 1:org.NCells
                newState(iCell, 1) = CellState(org.OrgCells(iCell).State);
            end
            
            if (isempty(newState(iCell,1).TT_Species))
                fprint('Empty TT populations');
            end
            
            try
             
                % update the state for each cell
                for iCell = 1:org.NCells
                    
                    if (~org.OrgCells(iCell,1).Active)
                        continue;
                    end
                    
                    currSpeciesPopulations = org.OrgCells(iCell).State.TT_Species;
                    NObj = length(currSpeciesPopulations);
                    % update TT_Species from network (transport and the others)
                    % currSpeciesPopulations(iObj).update(ttUpdater.mapping,0,iObj);
                    s = size(ttUpdater.mapping);
                    for iMap = 1:s(1)
                        map = ttUpdater.mapping(iMap,:);
                        % check if its mapping from network
                        if (map{3}==0)
                            % if it's transport protein
                            if (map{7}==1)
                                currSpeciesPopulations{map{5}}(map{6}) = ...
                                    currSpeciesPopulations{map{5}}(map{6}) + ...
                                    transporters(iCell).transportProteins(currentTime, map{1});
                            else % if it's not transport protein
                                % and inactive
                                state = org.OrgCells(iCell).State. ...
                                    getState(map{2});
                                currSpeciesPopulations{map{5}}(map{6}) = ...
                                    state(map{4});
                            end
                        end
                    end
                     
                    for iObj = 1:NObj
                        % run simulation
                        V = org.OrgCells(iCell).Type.TT_Species(iObj).V;
                        rates = org.OrgCells(iCell).Type.TT_Species(iObj).ReactionRates;
                        stoichMatrix = ...
                            org.OrgCells(iCell).Type.TT_Species(iObj).StoichMatrix;
                        % Stoich coeffs of the reactants
                        stoichCoeffs = -stoichMatrix.*(stoichMatrix<0);
                        K = sum(stoichCoeffs);         % molecularity
                        kp = rates./V.^(K-1);        % 'particle' rate constant
                        stochRates = kp.*prod(factorial(stoichCoeffs)); % stochastic rate constant

                        % depending on number of LIN39 we call stochastic
                        % function or ode solver
                        [simulationTimeSteps, outputPopulation] =...
                            simulate(ttUpdater.simulatorsThreshold, ...
                                    currSpeciesPopulations{iObj}, stochRates, ...
                                    stoichMatrix, stoichCoeffs, delT);

                        s = size(simulationTimeSteps);
                        % Set output parameters
                        newDelT = [newDelT, simulationTimeSteps(s(2))];
                        
                        currSpeciesPopulations{iObj} = outputPopulation(:,s(2));
                        
                       % update the TT_Population objects
                       for iMap = 1:size(ttUpdater.mapping,1)
                            map = ttUpdater.mapping(iMap,:);
                            % check if its mapping from this object
                            if (map{3}==iObj)
                                 currSpeciesPopulations{map{5}}(map{6}) = ...
                                     currSpeciesPopulations{map{3}}(map{4});
                            end
                            if (map{5}==iObj && map{3}~= 0)
                                 currSpeciesPopulations{map{3}}(map{4}) = ...
                                     currSpeciesPopulations{map{5}}(map{6});
                            end
                       end
                   end
                    
                    % update network (no transport, only Proteins, External signals and repressors)
 
                    for iMap = 1:size(ttUpdater.mapping,1)
                        map = ttUpdater.mapping(iMap,:);
                        % check if its mapping from network and not
                        % transport
                        if (map{3}==0 && map{7}==0)
                           % and inactive
                            newState(iCell, 1).setState(...
                                map{2}, currSpeciesPopulations{map{5}}(map{6}), map{4});
                        end
                    end
 
                    newState(iCell, 1).TT_Species = ...
                          currSpeciesPopulations;
                      
                  if (isempty(newState(iCell,1).TT_Species))
                    fprint('Empty TT populations');
                  end
            

%                     % update concentration of LAG2
%                    newState(iCell, 1).ExternalSignals(ttUpdater.LAG2_PPI_index) = ...
%                         outputPopulation(ttUpdater.LAG2_TT_index, s(2));
%                    % update concentration of LIN39 in TT_LIN39Species
%                     newState(iCell, 1). ...
%                         TT_LIN39Species(ttUpdater.LIN39_TT_index_1)=...
%                         outputPopulation(ttUpdater.LIN39_TT_index_2);
% 
%                     % update concentration of LIP1
%                     newState(iCell, 1).Repressors(ttUpdater.LIP1_PPI_index,1) =  ...
%                         outputPopulation(ttUpdater.LIP1_TT_index, s(2));
%                     
%                     % add inactive UP1 into PPI
%                     delUP1 = ...
%                         newState(iCell, 1).TT_LIN39Species(ttUpdater.UP1_cyt_TT_index) - ...
%                         org.OrgCells(iCell).State.Proteins(ttUpdater.UP1_PPI_index);
%                     
%                     if (delUP1 > 0)                    
%                         newState(iCell, 1). ...
%                             Proteins(ttUpdater.UP1_PPI_index) = ...
%                         newState(iCell, 1). ...
%                             Proteins(ttUpdater.UP1_PPI_index) + delUP1;
%                     end    
%                       
                        
                end % for iCell
                
                for iCell = 1:org.NCells
                    if (~newState(iCell,1).checkPositive())
                                warning('MATLAB:TTUpdater:invalidUpdate', ...
                                    'TT causes negative cell population in cell %d', ...
                                    iCell);
                    end
                    delState(iCell,1) = newState(iCell,1) - ...
                      org.OrgCells(iCell,1).State;
                end % for iCell  
                
                if (isempty(newDelT))
                    newDelT = 0.;
                end
                
            catch ME
              outME = addCause( ...
                  MException( ...
                      'MATLAB:TTUpdater:noUpdate', ...
                      'Error in updating from transcription/translation.'), ...
                  ME);
              throw(outME);
            end
                
            % update the time that has elapsed
            newDelT = max(newDelT);
                            
        end
    end
    
end
