%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Utilities to involve time.
classdef Time
    
    methods (Static = true, Access = public)
        function [w1, w2] = interpolate(t, time1, time2)
            % determine the weights of the first and second state to the 
            % time between two specified times.
            if (t < time1 || t > time2)
                warning('Matlab:Time:Extrapolation', ...
                    'Extrapolating to time %f from times %f and %f', ...
                    t, time1, time2);
            end
            if (time1 >= time2)
                throw(MException('Matlab:Time:InvalidTime', ...
                    'Invalid time interval: (%f, %f)', ...
                    time1, time2));
            end
            w2 = (t-time1)./(time2-time1);
            w1 = 1-w2;
        end
        
        function [indexTime] = findTimeIndices(referenceTime, timeList)
        % find the indices in a list of times that corresponds to those
        % specified in a reference time list
            numTimes = length(referenceTime);
            indexTime = zeros(numTimes,1);
            
            for iTime = 1:numTimes
                currentTime = referenceTime(iTime);
                
                % find the times on the list surrounding the current time
                lessTime    = find(timeList <= currentTime, 1, 'last');
                greaterTime = find(timeList > currentTime, 1, 'first');
                
                % interpolate to find the appropriate index
                if (~any(lessTime))
                    % if before the first time, set to zero
                    indexTime(iTime,1) = 0;
                elseif (~any(greaterTime))
                    % if after the last time, set to 1 more than the number
                    % of times
                    indexTime(iTime,1) = length(timeList) + 1;
                else
                    indexTime(iTime,1) = lessTime;
                    % interpolate between adjacent indices
                    if (timeList(lessTime) < currentTime)
                        [w1, w2] = Time.interpolate(currentTime, ...
                            timeList(lessTime), timeList(greaterTime));
                        if (w2 < 0 || w2 >= 1)
                            warning('MATLAB:Time:extrapolation', ...
                                'Performing extrapolation to time %f from times %f and %f.', ...
                                currentTime, timeList(lessTime), timeList(greaterTime));
                        end
                        indexTime(iTime,1) = indexTime(iTime,1) + w2;
                    end
                end
            end
            
        end
        
        function [timeList,indexTime] = mergeTimeLists(time1, time2, ...
                isInclusive)
            
            if (nargin < 3)
                isInclusive = false;
            end
            
            % merge the time lists
            mergeTimeTemp = unique(sort([time1, time2]));
            numTimes = length(mergeTimeTemp);
            
            % find the indices corresponding to this time
            indexTimeTemp = zeros(numTimes,2);
            indexTimeTemp(:,1) = Time.findTimeIndices(mergeTimeTemp, time1);
            indexTimeTemp(:,2) = Time.findTimeIndices(mergeTimeTemp, time2);
                        
            % set the output times
            if (~isInclusive)
                timeList = mergeTimeTemp(1:numTimes);
                indexTime = indexTimeTemp(1:numTimes,:);
            else
                
                initTime = max(time1(1),time2(1));
                initTimeIdx = find(initTime == mergeTimeTemp);
                
                finalTime = min(time1(end),time2(end));
                finalTimeIdx = find(finalTime == mergeTimeTemp);
                
                timeList = mergeTimeTemp(initTimeIdx:finalTimeIdx);
                indexTime = indexTimeTemp(initTimeIdx:finalTimeIdx,:);
            end
        end
    end
end
