%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.


classdef TransportProtein < handle
    
    properties (GetAccess = public, SetAccess = private)
        transportProteinTypes
        expectedTransportTime
        proteinNames
        activationTime
        transportTime
    end
    
    methods (Access = public)
    
        function transProt = TransportProtein(cellProp)
            if (~isa(cellProp, 'CellProperties'))
                throw(MException('Matlab:TransportProtein:invalidInit', ...
                    'Must initialize TransportProtein class with a CellProperties object'))
            end
            
            if (any(cellProp.transportTime(logical(cellProp.canTransport)) < 0))
                throw(MException('Matlab:TransportProtein:invalidParam', ...
                    'Any transportable protein must have nonnegative transport time'))
            end
            
            transProt.expectedTransportTime = cellProp.transportTime(logical(cellProp.canTransport));
            transProt.transportProteinTypes = cellProp.ProteinTypes(logical(cellProp.canTransport),:);
            transProt.proteinNames   = {};
            transProt.activationTime = [];
            transProt.transportTime  = [];
        end
        
        function [numProteins, protList] = clearTransporter(transProt, ...
                protsToClear)
            
            % set the names of all the proteins if none are specified
            if (nargin < 2)
                protList = {};
                for protName = transProt.proteinNames
                    for iProt = 1:length(protsToClear)
                        addToList = true;
                        if (strcmpi(strtrim(cell2mat(protsToClear(iProt))), ...
                                strtrim(cell2mat(protName))))
                            addToList = false;
                        end
                    end
                    if (addToList)
                        protList = [protList, protName];
                    end
                end
            else
                
                if (isa(protsToClear, 'char'))
                    protList = {};
                    for iProt=1:size(protsToClear,1)
                        protList = [protList, {protsToClear(iProt,:)}];
                    end
                elseif (isa(protsToClear, 'cell'))
                    protList = protsToClear;
                else
                    thow(MException('TransportProtein:invalidClear', ...
                        'Cannot clear proteins of type %s from the transporter.', ...
                        cloass(protsToClear)));
                end
            end
            
            % delete desired proteins from the list
            numProteins = zeros(size(protList));
            actTimes = transProt.activationTime;    
            transTimes = transProt.transportTime;
            for iProt=1:length(protList)
                if (~isempty(actTimes))
                    for j=1:length(actTimes)
                        protRemoved = transProt.removeProtein(...
                            protList(iProt), actTimes(j), transTimes(j));
                        if (protRemoved)
                            numProteins(iProt) = numProteins(iProt) + 1;
                        end
                    end
                end
            end
            
            
        end
        
        function addTransportProtein(transProt, names, times)
            
            
            if (~isa(names,'cell'))
                throw(MException('TransportProtein:invalidAdd', ...
                    'Cannot add protein name of type %s.  Must use cell', ...
                    class(names)));
            end
            
            if (length(times) == 1)
                times = times.*ones(size(names));
            end
            
            for iName = 1:length(names)
                
                transProt.proteinNames   = [transProt.proteinNames, ...
                    names(iName)];
                transProt.activationTime = [transProt.activationTime, ...
                    times(iName)];
                newTime = transProt.getTransportTime(names(iName), times(iName));
                transProt.transportTime = [transProt.transportTime, newTime];
            end
            
        end
        
        function numRemoved = inactivateProt(transProt, ...
                currentTime, name, actTime, transTime)
            
            % check that the input times are valid
            if ( (length(name) ~= length(actTime)) || ...
                    (length(name) ~= length(transTime)) )
                throw(MException('TransportProtein:invalidInact', ...
                    'Need to specify the same number of protein names (%d), activation times (%d), and transport times (%d).', ...
                    length(name), length(actTime), length(transTime)));
            end
            
            numRemoved = 0;
            
            numInvalidTime = 0;
            
            for iProt = 1:length(name)
                if (currentTime >= transTime(iProt))
                    numInvalidTime = numInvalidTime + 1;
                    continue;
                end
                protRemoved = transProt.removeProtein(name(iProt), ...
                    actTime(iProt), transTime(iProt));
                if (protRemoved)
                    numRemoved = numRemoved + 1;
                end
            end
            
            if (numRemoved ~= length(name))
                fprintf('removed %d of %d proteins.  %d at an invalid time\n', ...
                    numRemoved, length(name), numInvalidTime);
            end
            
        end
        
        function numTransport = transportProteins(transProt, ...
                currentTime, name)
            
            
            [names, actTime, transTime] = transProt.getTransportList(...
                name, currentTime, -1);
            
            numTransport = 0;
            if (isempty(names))
                return;
            end
            
            numTransport = 0;
            for iProt = 1:length(names)
                if (transTime(iProt) >= currentTime)
                    throw(MException('TransportProtein:invalidTransport', ...
                        'Cannot tranport protein with transport time %f at time %f', ...
                        transTime(iProt), currentTime));
                end
                protRemoved = transProt.removeProtein(names(iProt), ...
                    actTime(iProt), transTime(iProt));
                if (protRemoved)
                    numTransport = numTransport + 1;
                end
            end
            
            if (length(names) ~= numTransport)
                warning('TransportProtein:reducedTransport', ...
                    'Only transported %d of a possible %d', ...
                    numTransport, length(names));
            end
        end
        
        function [nameList, actTime, transTime] = getTransportList(...
                transProt, name, time, logicalDirection)
            % logical direction = 0 for transport times = time, 
            %                   > 0 for transport times > time,
            %                   < 0 for tranport times  < time
            
            keepIndices = transProt.getProteinIndex(name, ...
                transProt.activationTime, time, false, logicalDirection);
            nameList  = transProt.proteinNames((keepIndices));
            actTime   = transProt.activationTime((keepIndices));
            transTime = transProt.transportTime((keepIndices));
        end
        
        % count the total number of proteins in the transport list for
        % debugging
        function [countNames, countActTimes, countTransTimes] = ...
                getNumInTransportList(transProt) 
            countNames = length(transProt.proteinNames);
            countActTimes = length(transProt.activationTime);
            countTransTimes = length(transProt.transportTime);
        end
        
    end
    
    methods (Access = private)
        
        function idx = getProteinIndex(transProt, ...
                name, actTime, transTime, first, transLogical)
            
            % transLogical indicates the directionality to search for in
            % the transport time -- if < 0, search for values < the
            % transportTime, if > 0, search for values > the transport
            % time, if = 0, search for values = the transportTime.
            if (nargin < 6)
                transLogical = 0;
            end
            
            if (isa(name,'cell'))
                name = cell2mat(name);
            end
            
            numProt = length(transProt.proteinNames);
            
            keepIdx = false(numProt,1);
            for i=1:numProt
               keepIdx(i) = ...
                   strcmpi(strtrim(cell2mat(transProt.proteinNames(i))), strtrim(name));
               
               % find any proteins which have activiation times in the
               % specified list
               keepIdx(i) = keepIdx(i) && ...
                   any(transProt.activationTime(i) == actTime);
               
               % find any proteins occurring in the right bounds of the
               % specified transport time
               if (transLogical > 0)
                   keepIdx(i) = keepIdx(i) && ...
                       (transProt.transportTime(i) > transTime);

               elseif (transLogical < 0)
                   keepIdx(i) = keepIdx(i) && ...
                       (transProt.transportTime(i) < transTime);
               else
                   keepIdx(i) = keepIdx(i) && ...
                       (transProt.transportTime(i) == transTime);
               end
               if (keepIdx(i) && first)
                   break;
               end
            end
            
            idx = 1:numProt;
            idx = idx(logical(keepIdx));
        end
        
        function protRemoved = removeProtein(transProt, ...
                name, actTime, transTime)
            protIdx = transProt.getProteinIndex(name, ...
                actTime, transTime, true, 0);
            
            
            if (length(protIdx) > 1)
                throw(MException('TransportProtein:invalidRemove', ...
                    'Cannot remove more than one protein at a time'));
            end
            
            if (isempty(protIdx))
                protRemoved = false;
                
                
                return;
            end
            
            if (any(protIdx < 1) || ...
                any(protIdx > length(transProt.proteinNames)))
                protRemoved = false;
                
                
                return;
            end
            
            keepIndices          = true(length(transProt.proteinNames),1);
            keepIndices(protIdx) = false;

            
            transProt.proteinNames   = transProt.proteinNames(logical(keepIndices));
            transProt.activationTime = transProt.activationTime(logical(keepIndices));
            transProt.transportTime  = transProt.transportTime(logical(keepIndices));
            protRemoved = true;
            
            
        end
        
        function transportTime = getTransportTime(transProt, protName, ...
                initialTime)
            
            expectedTime = -1;
            for iProt = 1:size(transProt.transportProteinTypes,1)
                if (strcmpi(strtrim(cell2mat(protName)), ...
                        strtrim(transProt.transportProteinTypes(iProt,:))))
                    expectedTime = transProt.expectedTransportTime(iProt);
                    break;
                end
            end
            
            if (expectedTime < 0)
                throw(MException('Matlab:TransportProtein:invalidProtein', ...
                    'Protein %s not listed as a transportable protein', ...
                    strtrim(cell2mat(protName))));
            end
            
            transportTime = initialTime + ...
                exprnd(expectedTime);
        end
        
    end
    
end
