%    GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%    Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.


classdef XML_Reader < handle
    % reads XML (SBML) file and gets information
    % about proteins and reactions
    
    %  Author:  Ludmila Danilova
    %  Date:    18 November 2008
    %
    % History
    % 06Jan09  LD  Add reading of reaction parameters
    
    
    properties
       ProteinsList = Protein.empty;   % list of objects of protein type 
       ReactionsList = Reaction.empty; % list of objects of reaction type
       CellList = CellInfo.empty;      % list of objects with cell information
       CellNames;                      % list of cell names
       Tag;                            % tag which says what process is described in XML file
       Volume;                         % voulume of nucleus
    end
    
    methods 
        %% Constructor
      function obj = XML_Reader(file)
          % input:
          % file: XML file with description of process
          
          % output:
          % obj: content of XML file
          % tag: what process is described in XML file
          
       if nargin > 0
         try
           % read XML file
           tree = xmlread(file);
           % get list of proteins
           getProteins(obj, tree);
           % get list of reactions
           getReactions(obj, tree);
           % specify the rest properties of proteins 
%           assingProteinProperties(obj);
           % get cell names
           getCellNames(obj, tree);
           % get tag which says what process is described in XML file
           getTag(obj, tree);
           % get cell information
           getCellInfo(obj, tree);
           % get volume of nucleus
           getVolume(obj,tree,'nucleus');
        
         catch ME
             fprintf('XML_Reader:invalid file. Invalid input file: %s \n', file);
             outME = addCause( ...
                    MException( ...
                    'MATLAB:XML_Reader:invalid file', ...
                    'Invalid input file: %s', file), ...
                     ME);
                error('Failed to read XML file %s.',file);
                throw(outME);  
         end 
       end % end IF
      end % end Constructor
       
      %% Get proteins
      %  Find all elements with tag 'species'.
      function getProteins(obj, tree)
       ListOfSpecies = tree.getElementsByTagName('species');
       % iterate elements
       for k = 0:ListOfSpecies.getLength-1
         % get one element
         SpeciesItem = ListOfSpecies.item(k);
         % get their attributes
         allAttributes = SpeciesItem.getAttributes;
         % get 'name' attribute 
         speciesName = allAttributes.getNamedItem('id');
         % get 'speciesType' attribute 
         speciesType = allAttributes.getNamedItem('name'); 
         % get 'initialAmount' attribute 
         concentrations = allAttributes.getNamedItem('initialAmount');
         % get 'compartment' attribute
         compart = allAttributes.getNamedItem('compartment');
         % get value of 'name' attribute 
         name = char(speciesName.getNodeValue);
         % check if it ends by 'i' or 'a' and removes them
         % and says what remove
         [name, status, state] = XML_Reader.checkProteinName(name);
         % create object Protein
         p1 = Protein(name);
         % assign it to the object properties
         % get value of 'speciesType' attribute 
         if (~isempty(speciesType))
            p1.type = char(speciesType.getNodeValue);
         end
         % get value of 'initialAmount' attribute 
         switch status
            case 0
                p1.amountInactive = char(concentrations.getNodeValue);
            case 1
                p1.amountActive = char(concentrations.getNodeValue);
          end
         
         % get value of 'compartment' attribute 
         p1.compartment = char(compart.getNodeValue);
         % add protein to list
         addProtein(obj, p1);            
       end % end FOR       
      end % function getProteins
      
      %% Get reactions
      %  Find all elements with tag 'reaction' 
      function getReactions(obj, tree)
        ListOfReactions = tree.getElementsByTagName('reaction');
        % iterate elements (reactions)
        for k = 0:ListOfReactions.getLength-1
          reactants = {}; % vector of reactants
          products = {}; % vector of products
          parameterValues = {}; % vector of parameters
          % get one reaction
          ReactionItem = ListOfReactions.item(k); 
          % get their attributes
          allAttributes = ReactionItem.getAttributes;
          % get 'id' attribute 
          ReactionId = allAttributes.getNamedItem('id');
          % get value of 'id' attribute 
          id = char(ReactionId.getNodeValue);
          % get 'name' attribute 
          ReactionName = allAttributes.getNamedItem('name');
          % get value of 'name' attribute 
          name = char(ReactionName.getNodeValue);
          % get 'reversible' attribute
          Reversible = allAttributes.getNamedItem('reversible');
          % get all child nodes
          if ReactionItem.hasChildNodes
            listOfchildNodes = ReactionItem.getChildNodes;
            numChildNodes = listOfchildNodes.getLength;
            % iterate child nodes
            for count = 0:numChildNodes-1
              node = listOfchildNodes.item(count);
              switch char(node.getNodeName)
              % if it's node with reactants
               case 'listOfReactants'
               % get list of reactants
               listOfReactants = ...
                   node.getElementsByTagName('speciesReference');
               % iterate list of reactants
               for i = 0:listOfReactants.getLength-1
                  % get one element
                  ReactantItem = listOfReactants.item(i); 
                  % get their attributes
                  ReactantAttributes = ReactantItem.getAttributes;
                  % get 'species' attribute 
                  reactant = ReactantAttributes.getNamedItem('species');   
                  % get value of 'id' attribute 
                  reactantName = XML_Reader.checkProteinName(char(reactant.getNodeValue));
                  % add element into reactants vector
                  if isempty(reactants)
                   reactants = {reactantName};
                  end
                  % check if reactants has this element
                  if (~strcmp(reactantName, reactants))
                    reactants = [reactants,reactantName];
                  end 
               end % FOR i
              % if it's node with products
              case 'listOfProducts'
                listOfProducts = ...
                    node.getElementsByTagName('speciesReference');
                % iterate list of products
               for i = 0:listOfProducts.getLength-1
                  % get one element
                  ProductItem = listOfProducts.item(i); 
                  % get their attributes
                  ProductAttributes = ProductItem.getAttributes;
                  % get 'species' attribute 
                  product = ProductAttributes.getNamedItem('species');   
                  % get value of 'id' attribute 
                  productName = char(product.getNodeValue);
                  % check if it ends by 'i' or 'a' and removes them
                  productName = XML_Reader.checkProteinName(productName);        
                  % add element into products vector
                  if isempty(products)
                   products = {productName};
                  end
                  % check if reactants has this element
                  if (~strcmp(productName, products))
                    products = [products,productName];
                  end 
               end % FOR i
               % if it's node with reaction parameters              
               case 'kineticLaw'
                   listOfParameters = ...
                       node.getElementsByTagName('parameter');                   
                   % iterate list of parameters
                   for i = 0:listOfParameters.getLength-1
                       % get one element
                       ParameterItem = listOfParameters.item(i); 
                       % get their attributes
                       ParameterAttributes = ParameterItem.getAttributes;
                       % get 'value' attribute 
                       value = ParameterAttributes.getNamedItem('value');                       
                       parameterValues = ...
                           [parameterValues,char(value.getNodeValue)];
                   end % FOR i
              end % switch
            end % FOR count   
          end % IF
          % create reaction
          r = Reaction;
          r.Name = id;
          r.Reactants = reactants;
          r.Products = products;  
          r.Parameter = parameterValues(1,1);
          switch char(name)
              case {'scaffold','spontaneous_receptor', ...
                    'spontaneous_unbound','spontaneous_bound'}
                % reversed reaction
                  r1 = Reaction;
                  r1.Name = id;
                  r1.Reactants = products;
                  r1.Products = reactants;
                  r1.Parameter = parameterValues(length(parameterValues));
                  switch char(name)
                      case 'scaffold'
                        % reaction of scaffold binding
                        r.Type = [name,'_binding'];
                        % reaction of scaffold unbinding
                        r1.Type = [name,'_unbinding'];
                      case 'spontaneous_receptor'
                     % reaction of spontaneuos activation of receptors
                       r.Type = [name,'_activation'];
                       r1.Type = [name,'_inactivation'];
                      case 'spontaneous_unbound'
                         % reaction of spontaneuos activation of unbound proteins
                       r.Type = [name,'_activation'];
                       r1.Type = [name,'_inactivation'];
                     case 'spontaneous_bound'
                     % reaction of spontaneuos activation of bound proteins
                        r.Type = [name,'_activation'];
                        r1.Type = [name,'_inactivation'];
                 end %switch
                  % add both reactions into reaction list
                 obj.ReactionsList = [obj.ReactionsList, r];
                 obj.ReactionsList = [obj.ReactionsList, r1];
              otherwise
              r.Type = name;
              obj.ReactionsList = [obj.ReactionsList, r];
              % if reaction is reversible
              if (isempty(Reversible)) 
                r1 = Reaction;
                r1.Name = id;
                r1.Type = name;
                r1.Reactants = products;
                r1.Products = reactants;
                r1.Parameter = parameterValues(length(parameterValues));
                obj.ReactionsList = [obj.ReactionsList, r1];
              end  % IF
          end% switch
         end % FOR k
      end % end of function getReactions
   
      %% Get cell names
      function getCellNames(obj, tree)
        ListOfCompartments = tree.getElementsByTagName('compartment');
       % iterate elements
       for k = 0:ListOfCompartments.getLength-1
         % get one element
         CompartmentItem = ListOfCompartments.item(k);
         % get their attributes
         allAttributes = CompartmentItem.getAttributes;
         % get 'name' and 'compartmentType' attributes
         compartmentName = allAttributes.getNamedItem('name');  
         compartmentType = allAttributes.getNamedItem('compartmentType');
         % get value of 'name' attribute 
         name = char(compartmentName.getNodeValue);
         % get value of 'compartmentType' attribute 
         if (~isempty(compartmentType))
           type = char(compartmentType.getNodeValue);
           % if compartment type is cell
           if (strcmp(type, 'cell'))
               % store cell name into CellNames array
             obj.CellNames = [obj.CellNames, {name}];  
           end
         end
       end % for
      end % getCellNames
      
      %% Get cell information
      function getCellInfo(obj, tree)
        ListOfCells = tree.getElementsByTagName('cell');
       % iterate elements
       for k = 0:ListOfCells.getLength-1
           % get one element
           CellItem = ListOfCells.item(k);
           % get their attributes
           allAttributes = CellItem.getAttributes;
           % get attributes
           name = allAttributes.getNamedItem('name');
           alive = allAttributes.getNamedItem('alive');
           progeny = allAttributes.getNamedItem('progenyOf');
           rightNeighbor = allAttributes.getNamedItem('rightNeighbor');
           leftNeighbor = allAttributes.getNamedItem('leftNeighbor');
           startTime = allAttributes.getNamedItem('startTime');
           divisionTime = allAttributes.getNamedItem('divisionTime');
           growthRate = allAttributes.getNamedItem('growthRate');
           signalList = CellItem.getElementsByTagName('signal');
           signals = [];
           distances = [];
           for n = 0:signalList.getLength-1
               % get one element
                signalItem = signalList.item(n);
               % get their attributes
                allAttributes = signalItem.getAttributes;
                sigName = allAttributes.getNamedItem('id');
                signals = [signals,char(sigName.getNodeValue)];
                dist = allAttributes.getNamedItem('distanceToSource');
                distances=[distances, str2double(char(dist.getNodeValue))];
           end
           cell = CellInfo.empty;
           % get values of attributes
           if (~isempty(name))
               name = char(name.getNodeValue);
               cell = CellInfo(name);
           else
               fprint('There is no cell name in element %d', k);
               return;
           end
           if (~isempty(alive))
               cell.Alive = char(alive.getNodeValue);
           end
           if (~isempty(progeny))
               cell.ProgenyOf = char(progeny.getNodeValue);
           end
           if (~isempty(rightNeighbor))
               cell.RightNeighbor = char(rightNeighbor.getNodeValue);
           end
           if (~isempty(leftNeighbor))
               cell.LeftNeighbor = char(leftNeighbor.getNodeValue);
           end
           if (~isempty(startTime))
               cell.StartTime = char(startTime.getNodeValue);
           end
           if (~isempty(divisionTime))
               cell.DivisionTime = char(divisionTime.getNodeValue);
           end
           if (~isempty(growthRate))
               cell.GrowthRate = char(growthRate.getNodeValue);
           end
           if (~isempty(signals))
               cell.SignalTypes = signals;
           end
           if (~isempty(distances))
               cell.DistanceToSource = distances;
           end
           
           obj.CellList = [obj.CellList, cell];       
        end % for
      end % getCellInfo
      
      %% add element into array of proteins
      function addProtein(obj, value)
          % if array is empty
        if isempty(obj.ProteinsList)
          obj.ProteinsList = [obj.ProteinsList, value];        
        else
        % check if array has element with the same name
        isContain = false;
        value_name = char(value.name);
        for i = 1:length(obj.ProteinsList)
          p = obj.ProteinsList(i);
          p_name = char(p.name);
          if (strcmp(p_name, value_name))
              % update concentration
              if (p.amountActive == 0)
                  p.amountActive = value.amountActive;
              end
              if (p.amountInactive == 0)
                  p.amountInactive = value.amountInactive;
              end
             isContain = true;
             break;
          end
        end % end FOR
        % if array doesn't have element with the same name
        % add it
        if (~isContain)
            obj.ProteinsList = [obj.ProteinsList,value];
        end
        end % end IF
      end % end function addProtein
      
      %% specify the rest properties of proteins
      function assingProteinProperties(obj)
          NProt = length(obj.ProteinsList);
          % iterate reactions
          for i = 1:length(obj.ReactionsList)
              r = obj.ReactionsList(i);
              % get reaction type
              r_type = char(r.Type);
              switch r_type
                  case 'scaffold'
                      % search scaffold in the list of proteins
                      for j = 1:NProt
                          p = obj.ProteinsList(j);
                          % if protein type is 'scaffold'
                          % and protein name equals to reaction product
                          if (strcmp(char(p.type),'scaffold')&& ...
                                  strcmp(char(p.name),char(r.Products(1))))
                              % assign 'scaffolds' property
                              p.scaffolds = r.Reactants;
                              % remove 'a' in the end of proteins names
                              for k = 1:length(p.scaffolds)
                                  [t, s] = XML_Reader.checkProteinName(char(p.scaffolds(k)));
                                  p.scaffolds(k) = {t};
                              end % for k
                              break
                          end % if
                      end % for j
                      
                  case {'activation', 'repression', 'activation_ES'}
                      % compare reactants and products
                      % need to find out who activates or represses whom
                      regulator = ''; % who activates or represses
                      target = '';    % is activated or repressed
                      i1_regulator = 0; % index of regulator in reactants
                      % search for regulator
                      for k1 = 1:length(r.Reactants)
                          while (i1_regulator == 0)
                              reactant = char(r.Reactants(k1));
                              for k2 = 1:length(r.Products)
                                  product = char(r.Products(k2));
                                  % if reactant = product => reactant is activator
                                  if (strcmp(reactant,product))
                                      regulator = reactant;
                                      % remove 'a' or 'i' in the end
                                      [regulator, s] = XML_Reader.checkProteinName(regulator);
                                      % store indices of regulator
                                      i1_regulator = k1;
                                      break
                                  end % if
                              end % for k2
                          end % for k1
                      end
                      % search for target
                      for k1 = 1:length(r.Reactants)
                          % if reactant is not regulator
                          if (i1_regulator ~= 0 && k1 ~= i1_regulator)
                              % store it
                              target = char(r.Reactants(k1));
                              break;
                          end
                      end % for k1
                      
                      % search regulator in the protein list
                      for j = 1:NProt
                          p = obj.ProteinsList(j);
                          if (strcmp(char(p.name),regulator))
                              % if regulator is activator
                              if (strcmp(r_type, 'activation'))
                                  % if target ends on 'i'
                                  if (target(length(target))== 'i')
                                      % remove 'i' in the end
                                      [target, s] = XML_Reader.checkProteinName(target);
                                      % add target in 'activates' the protein property
                                      p.activates = [p.activates, target];
                                      break
                                  end
                              end
                              % if regulator is repressor
                              if (strcmp(r_type, 'repression'))
                                  % and if target ends on 'a'
                                  if (target(length(target))== 'a')
                                      % remove 'a' in the end
                                      [target, s] = XML_Reader.checkProteinName(target);
                                      % add target in 'represses' the protein property
                                      p.represses = [p.represses, target];
                                      break
                                  end
                              end
                              % if target is external signal
                              % and TF activates it
                              if (strcmp(r_type, 'activation_ES'))
                                  % add target in 'activates' the protein property
                                  p.activates = [p.activates, target];
                                  break
                              end
                          end
                      end
                      
              end % switch
          end % for i
          
      end % assingProteinProperties
      
      %% Specify Tag property
      function getTag(obj, tree)
        AnnotationTag = tree.getElementsByTagName('p');
       for k = 0:AnnotationTag.getLength-1
           Item = AnnotationTag.item(k);        
           obj.Tag = char(Item.getTextContent);
        end % for k
      end % getTag
      
      %% Specify Volume property
      function getVolume(obj, tree, compartment)
        compartments = tree.getElementsByTagName('compartment');
       for k = 0:compartments.getLength-1
           Item = compartments.item(k);
           % get their attributes
           allAttributes = Item.getAttributes;
           % get attribute
           id = allAttributes.getNamedItem('id');
           
           if (~isempty(id) && strcmp(char(id.getNodeValue),compartment))
               size = allAttributes.getNamedItem('size');
               obj.Volume = char(size.getNodeValue);
           end
        end % for k
      end % getTag
      
    end % methods
    
    %% Static methods
    methods (Static)
        % Checks protein name and remove 'a', 'i', '_u', or '_b' 
        % in the end of name if necessary
        % Returns status of protein:
        % 0 is inactive
        % 1 is active
        % Returns state of protein
        % 0 is unbound
        % 1 is bound
       function [outputName, status, state] = checkProteinName(name)
            status = 0;
            state = 0;
            % check if it end of 
            if (name(length(name))== 'i')
               name = name(1:length(name)-1);
            end
            if (name(length(name))== 'a')
                name = name(1:length(name)-1);
                status = 1;
            end
            if (name(length(name))== 'u')
                name = name(1:length(name)-2);
                if (name(length(name))== 'a')
                   status = 1;   
                end
                name = name(1:length(name)-1);  
            end
            if (name(length(name))== 'b')
               state = 1;
               name = name(1:length(name)-2);
               if (name(length(name))== 'a')
                  status = 1;
               end
               name = name(1:length(name)-1); 
            end
           outputName = name;
        end % checkProteinName
    end % static methods
    
end
