function plotSpeciesHistory(orgHistory, stateType, stateNames, ...
    plotSymbols, inHours, showLegend, plotColors, plotCells)
% plotSpeciesHistory    Plots species evolution from OrganismHistory object
%
%   plotSpeciesHistory(orgHistory, stateType, stateName, plotSymbols,
%   inHours, showLegend, plotColors, plotCells) will plot the population of
%   species of name stateNames and type stateType from the OrganismHistory
%   object orgHistory.
%
%   If inHours is true, this function will scale the time axis to reflect
%   the simulation hour.  Otherwise, the time axis will have units of
%   seconds.
%
%   plotSymbols specifies the plot colors / symbols used for each species
%   specified in stateNames while plotColors specifies the plot colors /
%   symbols used for each cell.  A legend for this color scheme can be
%   placed on the figure by setting the showLegend argument to true.
%
%   The final optional argument plotCells should be set to plot only a 
%   subset of the cells stored in orgHistory.  This should be a vector of
%   logicals of size given by the number of cells in orgHistory whose ith
%   element is set to true to plot the ith cell and false to exclude it.
%
%   See the GESSA User's Manual for further description of this function.
%
%   GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%   Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.


boundsSet = false;

stateTyp  = char(stateType);
stateName = char(stateNames);
plotSymb  = char(plotSymbols);

NState = size(stateName,1);
NCell  = length(orgHistory.cellTypes);

if (NState > size(plotSymb,1))
    throw( MException(...
        'MATLAB:Plotter:invalidPlot', ...
        'Only specified %d plot symbols.  Need at least %d.', ...
        NState, ...
        size(plotSymb,1)));
end

plotCol = char(plotColors);
if (NState > size(plotCol,1))
    throw( MException(...
                     'MATLAB:Plotter:invaliPlot', ...
                     'Only specified %d plot colors.  Need at least %d', ...
                      NState, size(plotCol,1)));
end

legendString = cell.empty;

maxY = 0;
minY = 0;

madePlot = false;
colIdx = 0;

for iCell = 1:NCell
    
    if (nargin >= 8)
        if (~plotCells(iCell))
            continue;
        end
        
    end
    
    colIdx = colIdx + 1;
    
    if (~madePlot)
        madePlot = true;
    end
    
    for iState = 1:NState
        
        cellName = orgHistory.cellNames(iCell, :);
        
        try
            
            [state, time] = orgHistory.getState(cellName, ...
                stateTyp, stateName(iState,:));
            
        catch ME
            outME = addCause( ...
                MException(...
                'MATLAB:Plotter:invalidPlot', ...
                'Error recovering state %s for cell %s', ...
                stateName(iState,:), cellName), ...
                ME);
            throw(outME);
        end
        
        plotOpts = sprintf('%s%s', ...
            plotCol(colIdx), ...
            plotSymb(iState,:));
        
        if (inHours)
            plot(time./3600., state, plotOpts);
        else
            plot(time, state, plotOpts);
        end
        
        
        if (~isempty(state))
            for jState = 1:length(state)
                if (maxY < state(jState))
                    maxY = state(jState);
                end
                if (minY > state(jState) && state(jState) > -999)
                    minY = state(jState);
                end
            end
        end
        
        if (madePlot)
            hold on;
        end
        
        plotName = sprintf('%s %s', cellName, stateName(iState, :));
        legendString = [legendString, cellstr(plotName)];
    end
    
    if (~isempty(time))
        if (~boundsSet)
            minTime = min(time);
            maxTime = max(time);
            boundsSet = true;
        else
            minTime = min(minTime, min(time));
            maxTime = max(maxTime, max(time));
        end
    end
end

if (inHours)
    minTime = minTime./3600;
    maxTime = maxTime./3600;
end
if (maxY ~= 0)
    
    axis([minTime maxTime minY maxY]);
else
    axis([minTime maxTime -10 10]);
end

if (showLegend)
    legend(legendString);
end
hold off;
end
