function speciesPerOrg( orgHistory, cellName, speciesName, speciesType, ...
    startTime, endTime, timeStep, fileName)
% speciesPerOrg     Creates a table with specified species populations.
%
%   speciesPerOrg(orgHistory, cellName, speciesName, speciesType,
%   startTime, endTime, timeStep, fileName) creates a table in fileName
%   (default: 'speciesTable.txt') containing the population of speciesName
%   of type speciesType in cells specified in cellName for times between
%   startTime and endTime at the specified timeStep (default: 100s) from
%   input orgHistory simulation results.
%
%   See the GESSA User's Manual for further description of this function.
%   
%   GESSA (Graphically Extended Stochastic Simulation Algorithm)
%   
%   Copyright (C) 2010 Michael Ochs, Elana Fertig, Ludmila Danilova
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.


   
    if (nargin < 7)
        timeStep = 100;
        fileName = 'speciesTable.txt';
    end
    if (nargin < 8)
        fileName = 'speciesTable.txt';
    end
    
    NCells = length(cellName);
    NSpecies = length(speciesName);
    NTypes = length(speciesType);
    if (NSpecies ~= NTypes)
        fprintf('MATLAB:speciesPerOrg: numbers of species and types must be the same.\n');
        return;
    end
    snapshots = (startTime+timeStep):timeStep:endTime;
    NSnapshots = length(snapshots);
            
     try
         nOrg = size(orgHistory);
         fid = fopen(fileName, 'wt');
         fprintf(fid, 'snapshots at time:\n');
         fprintf(fid, '\t%3.1f', snapshots/3600);
         fprintf(fid, '\t%s', 'mean');
         fprintf(fid, '\n');   
             
         for iOrg = 1:nOrg
             fprintf(fid, '-------------------------------------\n');
             fprintf(fid, 'organism %d',iOrg);
             fprintf(fid, '\n');   
            for iSpecies = 1:NSpecies
                 NMolecules = zeros(NCells,NSnapshots);
                 fprintf(fid, char(speciesName(iSpecies)));
                 fprintf(fid, '\n');   
 
                 k = 1;
                 for i = snapshots     
                        NMolecules(:,k) = getSpeciesNumber(orgHistory(iOrg,1),...
                            speciesName(iSpecies), speciesType(iSpecies), cellName, i);
                     k = k+1;
                 end
                for iCell = 1:NCells
                     fprintf(fid,'%s\t', char(cellName(iCell)));
                     fprintf(fid, '%g\t', NMolecules(iCell,:));
                     fprintf(fid, '%g\t', mean(NMolecules(iCell,:)));
                     fprintf(fid, '\n');
                end
             end
          end
         fclose(fid);
       catch ME
              outME = addCause( ...
                  MException( ...
                      'MATLAB:speciesPerOrg', ...
                      'Error in counting number of species.'), ME);
              throw(outME);
     end
     

end
